/************************************************************
 **
 ** bsc_blck.c : BSC block processing definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/

#include <string.h>

#include "pfile.h"
#include "pic.h"
#include "bsc_parm.h"
#include "bsc_tokn.h"
#include "bsc_vdef.h"
#include "bsc_ctrl.h"
#include "bsc_proc.h"
#include "bsc_intr.h"
#include "bsc_expr.h"
#include "bsc_cmd.h"
#include "pic_bsc.h"
#include "bsc_blck.h"
#ifdef MSDOS
#include <dir.h>
#define FILENAME_MAX MAXPATH
#endif
static void bsc_parse_pragma(bsc_file_t *bsc, unsigned sub, size_t param_ct,
    const param_t *params);
static void bsc_parse_include(bsc_file_t *bsc, unsigned sub, size_t param_ct,
    const param_t *params);

static const char *port_ids[] = { "a", "b", "c", "d", "e", 0 };
static const char *pwm_ids[]  = { "ccp1", "ccp2", "all", 0 };
static const char *usart_dir_ids[] = { "rx", "tx", 0 };
static const char *usart_bit_ids[] = { "8", 0 };

static param_dcr_t param_usart[] = {
  { param_type_cstr, usart_dir_ids },
  { param_type_cstr, usart_bit_ids },
  { param_type_expr, 0             },
  { param_type_expr, 0             }
};

static param_dcr_t param_pwm[] = {
  { param_type_cstr, pwm_ids }
};

static param_dcr_t param_c2[] = { 
  { param_type_cexpr, 0 },
  { param_type_expr,  0 }
};

static param_dcr_t param_e1[] = { 
  { param_type_expr, 0 }
};

static param_dcr_t param_e2[] = { 
  { param_type_expr, 0 },
  { param_type_expr, 0 }
};

static param_dcr_t param_e4[] = { 
  { param_type_expr, 0 },
  { param_type_expr, 0 },
  { param_type_expr, 0 },
  { param_type_expr, 0 }
};

static param_dcr_t param_p1_e1[] = { 
  { param_type_cstr, port_ids },
  { param_type_expr, 0 }
};

static param_dcr_t param_p1_e1_v1[] = { 
  { param_type_cstr, port_ids },
  { param_type_expr, 0 },
  { param_type_var,  0 }
};

static param_dcr_t param_p1_v1[] = { 
  { param_type_cstr, port_ids },
  { param_type_var,  0}
};

static param_dcr_t param_a1[] = { 
  { param_type_assign, 0 }
};

static param_dcr_t param_v1[] = { 
  { param_type_var, 0 }
};

static param_dcr_t param_e1_l2_e2[] = { 
  { param_type_expr,  0 },
  { param_type_lexpr, 0 },
  { param_type_lexpr, 0 },
  { param_type_expr,  0 },
  { param_type_expr,  0 }
};

static param_dcr_t param_e1_v1[] = {
  { param_type_expr, 0 },
  { param_type_var,  0 }
};

#define PARAM_PAIR(x) COUNT(x), x
#define PARAM_NULL    0, 0

typedef struct command_ {
  const char     *token;
  unsigned        misc;        /* used for misc. stuff */
  parse_fn_t      fn;
  size_t          param_ct;
  param_dcr_t    *param_dcr;
  const char     *reg_reqd;    /* this register must exist to continue */
} command_t;

static const command_t language[] = {
  /* variable definition */
  {"byte",         bsc_vardef_byte,                 bsc_parse_vardef,
   PARAM_NULL},
  {"sbyte",        bsc_vardef_sbyte,                bsc_parse_vardef,
   PARAM_NULL},
  {"word",         bsc_vardef_word,                 bsc_parse_vardef,
   PARAM_NULL},
  {"sword",        bsc_vardef_sword,                bsc_parse_vardef,
   PARAM_NULL},
  {"dword",        bsc_vardef_dword,                bsc_parse_vardef,
   PARAM_NULL},
  {"sdword",       bsc_vardef_sdword,               bsc_parse_vardef,
   PARAM_NULL},
  {"constant",     0,                               bsc_parse_constdef,
   PARAM_NULL},
  /* flow control */
  {"for",          bsc_intrinsic_type_null,  bsc_parse_for,
   PARAM_PAIR(param_a1)},
  {"if",           bsc_intrinsic_type_null,  bsc_parse_if,
   PARAM_PAIR(param_e1)},
  {"gosub",        cmd_branchtype_call,         bsc_parse_branch,
   PARAM_NULL},
  {"goto",         cmd_branchtype_goto,         bsc_parse_branch,
   PARAM_NULL},
  {"return",       cmd_branchtype_return,       bsc_parse_branch,
   PARAM_NULL},
  {"call",         0,                           bsc_parse_call,
   PARAM_NULL},
  {"proc",         0,                           bsc_parse_proc,
   PARAM_NULL},
  /* misc. */
  {"clear",        bsc_intrinsic_type_clear,         bsc_parse_intrinsic,
   PARAM_NULL},
  {"delay",        bsc_intrinsic_type_delay,         bsc_parse_intrinsic,
   PARAM_PAIR(param_c2)},
  {"interrupt_enable", bsc_intrinsic_type_interrupt_enable,   
    bsc_parse_intrinsic, PARAM_NULL},
  {"interrupt_disable", bsc_intrinsic_type_interrupt_disable, 
    bsc_parse_intrinsic, PARAM_NULL},
  /* IR */
  {"ir_adc_begin", bsc_ir_type_adc,                  bsc_parse_ir,
   PARAM_NULL, "_adif"},
  {"ir_pwm_begin", bsc_ir_type_pwm,                  bsc_parse_ir,
   PARAM_NULL, "_tmr2if"},
  {"ir_rx_begin",  bsc_ir_type_rx,                   bsc_parse_ir,
   PARAM_NULL, "_rcif"},
  {"ir_tx_begin",  bsc_ir_type_tx,                   bsc_parse_ir,
   PARAM_NULL, "_txif"},
  {"ir_timer_begin", bsc_ir_type_timer,              bsc_parse_ir,
   PARAM_NULL, "_tmr1if"},
  {"ir_portb_begin", bsc_ir_type_rb,                 bsc_parse_ir,
    PARAM_NULL, "_rbie"},
  {"ir_user_begin", bsc_ir_type_user,                bsc_parse_ir,
    PARAM_NULL, 0},
  /* ADC */
  {"adc_config",   bsc_intrinsic_type_adc_config,    bsc_parse_intrinsic,
   PARAM_PAIR(param_e1_l2_e2), "_adcon0"},
  {"adc_on",       bsc_intrinsic_type_adc_on,        bsc_parse_intrinsic,
   PARAM_NULL, "_adcon0"},
  {"adc_off",      bsc_intrinsic_type_adc_off,       bsc_parse_intrinsic,
   PARAM_NULL, "_adcon0"},
  {"adc_start",    bsc_intrinsic_type_adc_start,    bsc_parse_intrinsic,
   PARAM_NULL, "_adcon0"},
  {"adc_store",    bsc_intrinsic_type_adc_store,    bsc_parse_intrinsic,
   PARAM_PAIR(param_v1), "_adresl"},
  {"adc_wait",     bsc_intrinsic_type_adc_wait,     bsc_parse_intrinsic,
   PARAM_NULL, "_adcon0"},
  /* PWM */
  {"pwm_config",   bsc_intrinsic_type_pwm_config,   bsc_parse_intrinsic,
   PARAM_PAIR(param_e4)},
  {"pwm_duty",     bsc_intrinsic_type_pwm_duty,     bsc_parse_intrinsic,
   PARAM_PAIR(param_e2), "_ccp1con"},
  {"pwm_off",      bsc_intrinsic_type_pwm_off,      bsc_parse_intrinsic,
   PARAM_PAIR(param_pwm), "_ccp1con"},
  {"pwm_period",   bsc_intrinsic_type_pwm_period,   bsc_parse_intrinsic,
   PARAM_PAIR(param_e1), "_ccp1con"},
  {"pwm_reg",      bsc_intrinsic_type_pwm_reg,      bsc_parse_intrinsic,
   PARAM_PAIR(param_e1), "_ccp1con"},
  /* PORT control */
  {"interrupt_portb", bsc_intrinsic_type_interrupt_portb,  bsc_parse_intrinsic,
   PARAM_PAIR(param_e2), "_rbie"},
  {"pinhigh",      bsc_intrinsic_type_pinhigh,      bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_e1)},
  {"pinlow",       bsc_intrinsic_type_pinlow,       bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_e1)},
  {"pinrd",        bsc_intrinsic_type_pinrd,        bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_e1_v1)},
  {"portout",      bsc_intrinsic_type_portout,      bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_e1)},
  {"portrd",       bsc_intrinsic_type_portrd,       bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_v1)},
  {"setport",      bsc_intrinsic_type_setport,      bsc_parse_intrinsic,
   PARAM_PAIR(param_p1_e1)},
  /* USART */
  {"rx_err_check", bsc_intrinsic_type_rx_err_check, bsc_parse_intrinsic,
   PARAM_PAIR(param_v1), "_rcsta"},
  {"rx_store",     bsc_intrinsic_type_rx_store,     bsc_parse_intrinsic,
   PARAM_PAIR(param_v1), "_rcreg"},
  {"tx_load",      bsc_intrinsic_type_tx_load,      bsc_parse_intrinsic,
   PARAM_PAIR(param_e1), "_txreg"},
  {"usart_config", bsc_intrinsic_type_usart_config, bsc_parse_intrinsic,
   PARAM_PAIR(param_usart), "_spbrg"},
  {"wait_rx",      bsc_intrinsic_type_wait_rx,      bsc_parse_intrinsic,
   PARAM_NULL, "_rcif"},
  {"wait_tx",      bsc_intrinsic_type_wait_tx,      bsc_parse_intrinsic,
   PARAM_NULL, "_txif"},
  /* timer */
  {"timer_config", bsc_intrinsic_type_timer_config, bsc_parse_intrinsic,
   PARAM_PAIR(param_e2), "_t1con"},
  {"timer_start",  bsc_intrinsic_type_timer_start, bsc_parse_intrinsic,
   PARAM_NULL, "_t1con"},
  {"timer_stop", bsc_intrinsic_type_timer_stop, bsc_parse_intrinsic,
   PARAM_NULL, "_t1con"},
  {"timer_countdown", bsc_intrinsic_type_timer_countdown, bsc_parse_intrinsic,
   PARAM_NULL, "_tmr1if"},
  {"timer_read", bsc_intrinsic_type_timer_read, bsc_parse_intrinsic,
    PARAM_PAIR(param_v1), "_tmr1"},
  {"timer_write", bsc_intrinsic_type_timer_write, bsc_parse_intrinsic,
    PARAM_PAIR(param_e1), "_tmr1"},
  {"print",    0, bsc_parse_print, PARAM_NULL},
  {"while",    0, bsc_parse_while, PARAM_NULL},
  {"do",       0, bsc_parse_do,    PARAM_NULL},
  {"sleep",    bsc_intrinsic_type_sleep, bsc_parse_intrinsic, PARAM_NULL},
  {"clear",    bsc_intrinsic_type_clear, bsc_parse_intrinsic, PARAM_NULL},
  {"pragma",   0, bsc_parse_pragma, PARAM_NULL},
  {"include",  0, bsc_parse_include, PARAM_NULL},
  {"eeprom_write", bsc_intrinsic_type_eeprom_write, bsc_parse_intrinsic,
    PARAM_PAIR(param_e2), "_eedata"},
  {"eeprom_read", bsc_intrinsic_type_eeprom_read, bsc_parse_intrinsic,
    PARAM_PAIR(param_e1_v1), "_eedata"},
  {"eeprom_wait", bsc_intrinsic_type_eeprom_wait, bsc_parse_intrinsic,
    PARAM_NULL, "_eedata"}
};

/*
 * NAME
 *   statement_skip
 *
 * DESCRIPTION
 *   skip to the end of this statement
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   used by variable & procedure definition after the first
 *   pass
 */
void bsc_statement_skip(bsc_file_t *bsc)
{
  while (!bsc_token_is_eos(bsc)) {
    bsc_token_get(bsc, bsc_token_next);
  }
}

void bsc_statement_next(bsc_file_t *bsc)
{
  do {
    if (!bsc_token_is_eos(bsc)) {
      if (!bsc_token_is(bsc, bsc_token_current, pfile_log_none,
        "'")) {
        pfile_log(bsc_file_pf_get(bsc), pfile_log_warn,
            BSC_MSG_EXTRA_CHAR_ON_LINE);
      }
      bsc_statement_skip(bsc);
    }
    bsc_token_get(bsc, bsc_token_next);
  } while (!bsc_token_is_eof(bsc) && bsc_token_is_eos(bsc));
}



/*
 * NAME
 *   command_process
 *
 * DESCRIPTION
 *   process a command
 *
 * PARAMETERS
 *   cmd : command to process
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this parses any necessary parameters, executing the command on success
 */
static void command_process(bsc_file_t *bsc, const command_t *cmd)
{
  param_t    *param;
  size_t      param_ct;
  pfile_t    *pf;
  boolean_t   err;

  err = boolean_false;
  pf = bsc_file_pf_get(bsc);
  if (!cmd->param_ct) {
    param_ct = 0;
    param    = 0;
  } else {
    param_ct = 0;
    param    = CALLOC(sizeof(*param), cmd->param_ct);
    if (0 == param) {
      pfile_log_syserr(pf, result_memory);
    } else {
      param_t *param_ptr;

      for (param_ct = 0, param_ptr = param; 
           (param_ct < cmd->param_ct)
           && (!param_ct
             || bsc_token_is(bsc, bsc_token_current, pfile_log_err, ","));
           param_ct++, param_ptr++) {
        if (param_ct) {
          bsc_token_get(bsc, bsc_token_next);
        }
        param_ptr->type = cmd->param_dcr[param_ct].type;
        switch (cmd->param_dcr[param_ct].type) {
          case param_type_null:
            break;
          case param_type_expr:
          case param_type_lexpr:
          case param_type_cexpr:
            if (result_ok == bsc_expr_parse(bsc, 0, 
                  &param_ptr->u.var)) {
              if (param_type_lexpr == cmd->param_dcr[param_ct].type) {
                /* need one more -- !! var */
                pfile_cmd_op_add(pf, 0, operator_logical, 
                    &param_ptr->u.var, 0, 0);
              } else if ((param_type_cexpr == cmd->param_dcr[param_ct].type)
                && !value_is_const(param_ptr->u.var)) {
                pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
                err = boolean_true;
              }
            } else {
              err = boolean_true;
            }
            break;
          case param_type_var:
            if (result_ok == bsc_token_to_value(bsc, pfile_log_err, 
                  &param_ptr->u.var)) {
              bsc_token_get(bsc, bsc_token_next);
            } else {
              err = boolean_true;
            }
            break;
          case param_type_cstr:
            {
              size_t       ct;
              const char **strs;

              strs = cmd->param_dcr[param_ct].param;

              for (ct = 0; 
                   strs[ct] 
                   && !bsc_token_is(bsc, bsc_token_current, 
                     pfile_log_none, strs[ct]);
                   ct++)
                ;
              if (strs[ct]) {
                param_ptr->u.sval = ct;
                bsc_token_get(bsc, bsc_token_next);
              } else {
                pfile_log(pf, pfile_log_err, BSC_MSG_SYNTAX_ERROR);
                err = boolean_true;
              }
            }
            break;
          case param_type_assign:
            {
              value_t    *var;

              if (result_ok == bsc_token_to_value(bsc, pfile_log_err, &var)) {
                if (bsc_token_is(bsc, bsc_token_next, pfile_log_err,
                   "=")) {
                  value_t *dst;

                  bsc_token_get(bsc, bsc_token_next);
                  if (result_ok == bsc_expr_parse(bsc, 0, &dst)) {
                    pfile_cmd_op_add(pf, 0, operator_assign, &var, dst,
                      0);
                    value_lock(var);
                    value_release(dst);
                    param_ptr->u.var = var;
                  } else {
                    err = boolean_true;
                  }
                } else {
                  err = boolean_true;
                }
                value_release(var);
              } else {
                err = boolean_true;
              }
            }
            break;
          case param_type_label:
            if (result_ok == bsc_token_to_label(bsc, pfile_log_err, 
              boolean_false, &param_ptr->u.lbl)) {
              bsc_token_get(bsc, bsc_token_next);
            }
            break;
        }
      }
    }
  }
  if (!err && (param_ct == cmd->param_ct)) {
    if (cmd->reg_reqd && !pfile_variable_exists(pf, 
              PFILE_LENGTH_DEF, cmd->reg_reqd)) {
      bsc_file_log_unavail(bsc, cmd->token);
    } else {
      cmd->fn(bsc, cmd->misc, param_ct, param);
    }
  }
  while (param_ct--) {
    switch (param[param_ct].type) {
      case param_type_null:
      case param_type_cstr:
        break;
      case param_type_expr:
      case param_type_cexpr:
      case param_type_lexpr:
      case param_type_assign:
      case param_type_var:
        value_release(param[param_ct].u.var);
        break;
      case param_type_label:
        label_release(param[param_ct].u.lbl);
        break;
    }
  }
  if (param) {
    FREE(param);
  }
}


/*
 * NAME
 *   bsc_block_process
 *
 * DESCRIPTION
 *   process until a block end occurs
 *
 * PARAMETERS
 *   pf       : returned by pfile_open()
 *   endct    : # of endwords
 *   endwords : words that end the  block
 *   ew       : [out] word # that ended the block, or EOF
 *
 * RETURN
 *   none
 *
 * NOTES
 *   return on EOF or when cb returns TRUE
 */
static int token_is_end(bsc_file_t *bsc, size_t endct, const char **endwords,
  size_t *ew)
{
  int rc;

  if (bsc_token_is_eof(bsc)) {
    if (ew) {
      *ew = EOF;
    }
    rc = 1;
  } else {
    size_t ii;

    for (ii = 0;
         (ii < endct)
         && !bsc_token_is(bsc, bsc_token_current, pfile_log_none,
           endwords[ii]);
         ii++)
      ;
    rc = (ii < endct);
    if (rc) {
      if (ew) {
        *ew = ii;
      }
      bsc_token_get(bsc, bsc_token_next);
    }
  }
  return rc;
}

void bsc_block_process(bsc_file_t *bsc, boolean_t skipstmt, size_t endct, 
    const char **endwords, size_t *ew)
{
  boolean_t  allow_label;
  pfile_t   *pf;
  unsigned   line_start;

  pf = bsc_file_pf_get(bsc);
  if (skipstmt) {
    bsc_statement_next(bsc);
  }

  line_start = pfile_line_get(pf);
  allow_label = boolean_true;
  while (!token_is_end(bsc, endct, endwords, ew)) {
    char        keyword[256];
    const char *ptr;
    size_t      ii;

    ptr = bsc_token_get(bsc, bsc_token_current);
    strncpy(keyword, ptr, sizeof(keyword) - 1);
    keyword[sizeof(keyword) - 1] = 0;
    bsc_token_get(bsc, bsc_token_next);

    if (allow_label
      && bsc_token_is(bsc, bsc_token_current, pfile_log_none,
        ":")) {
      /* keyword is a label */
      label_t *label;

      label = pfile_label_find(pf, pfile_log_none, keyword);
      if (label) {
        if ((1 == bsc_file_pass_get(bsc))
          && label_flag_test(label, LABEL_FLAG_DEFINED)) {
          label_release(label);
          label = 0;
          pfile_log(pf, pfile_log_err, BSC_MSG_MULTI_DEF, keyword);
        } else if ((2 == bsc_file_pass_get(bsc))
            && !label_flag_test(label, LABEL_FLAG_USED)) {
          pfile_log(pf, pfile_log_warn, BSC_MSG_NEVER_USED, 
              label_name_get(label));
        }
      } else {
        pfile_label_alloc(pf, keyword, 0, &label);
      }
      if (label) {
        label_flag_set(label, LABEL_FLAG_DEFINED);
        pfile_cmd_label_add(pf, label);
        label_release(label);
      }
      ptr = bsc_token_get(bsc, bsc_token_next);
      if (bsc_token_is_eos(bsc)) {
        bsc_statement_next(bsc);
      } else {
        allow_label = boolean_false;
      }
    } else {
      /* end must be special cased becuase it can be used
         as a two word keyword */
      if (0 == strcmp(keyword, "end")) {
        for (ii = 0;
             (ii < endct)
             && !((0 == memcmp(endwords[ii], "end", 3))
               && (0 == strcmp(endwords[ii]+3, ptr)));
             ii++)
          ;
        if (ii < endct) {
          if (ew) {
            *ew = ii;
          }
          bsc_token_get(bsc, bsc_token_next);
          break; /* <--- */
        }
        bsc_parse_intrinsic(bsc, bsc_intrinsic_type_end, 0, 0);
        bsc_statement_next(bsc);
      } else {
        for (ii = 0;
            (ii < COUNT(language))
            && strcmp(keyword, language[ii].token);
            ii++)
          ;
        if (COUNT(language) == ii) {
          if (strcmp(keyword, "\n")) {
            /* this is an assignment */
            value_t *var;

            var = 0;
            bsc_identifier_to_value(bsc, pfile_log_err, keyword, &var);
            if (var && (value_ct_get(var) > 1)) {
              bsc_subscript_parse(bsc, var);
              if (value_subscript_get(var)) {
                variable_expr_ct_bump(value_subscript_get(var), ctr_bump_incr);
              }
            }
            if (bsc_token_is(bsc, bsc_token_current, pfile_log_err, "=")) {
              value_t *rval;
              cmd_t   *cmd_start;

              cmd_start = pfile_cmdlist_get(pf);
              bsc_token_get(bsc, bsc_token_next);
              if (result_ok == bsc_expr_parse(bsc, 0, &rval)) {
                if (var) {

                  if ((2 == bsc_file_pass_get(bsc))
                    && !value_flag_test(var, VARIABLE_FLAG_USED)
                    && !value_flag_test(var, VARIABLE_FLAG_VOLATILE)) {
                    /* assign to unused variable, simply remove the entire
                     * assignment statement */
#if 1
                    cmd_t *cmd_end;
                    label_t *lbl;

                    lbl = 0;
                    cmd_end = pfile_cmdlist_get(pf);
                    while (cmd_end != cmd_start) {
                      cmd_t *link;

                      link = cmd_link_get(cmd_end);
                      if (link == cmd_start) {
                        lbl = cmd_label_get(cmd_end);
                      }
                      cmd_free(cmd_end);
                      cmd_end = link;
                    }
                    pfile_cmdlist_set(pf, cmd_start);
#endif
                    if (lbl) {
                      pfile_cmd_label_add(pf, lbl);
                    }
                  } else {
                    pfile_cmd_op_add(pf, 0, operator_assign, &var, rval, 0);
                  }
                }
                value_release(rval);
              }
            }
            if (var && value_subscript_get(var)) {
              variable_expr_ct_bump(value_subscript_get(var), ctr_bump_decr);
            }
            value_release(var);
          }
        } else {
          command_process(bsc, language + ii);
        }
        bsc_statement_next(bsc);
      }
      allow_label = boolean_true;
    }
  }
  if (endct && bsc_token_is_eof(bsc)) {
    pfile_log(pf, pfile_log_err, BSC_MSG_END_EXPECTED,
      endwords[0], line_start);
  }
}

/*
 * NAME
 *   bsc_include_process
 *
 * DESCRIPTION
 *   process an include file
 *
 * PARAMETERS
 *   bsc   : bsc file
 *   fname : include file name
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void bsc_include_process(bsc_file_t *bsc, size_t fname_sz,
  const char *fname)
{
  const char *ptr;
  char       *token;
  pfile_t    *pf;

  pf = bsc_file_pf_get(bsc);

  if (!fname_sz) {
    fname_sz = strlen(fname);
  }
  pfile_log(pf, pfile_log_debug, BSC_MSG_DBG_INCLUDE, (int) fname_sz, fname);
  ptr = bsc_token_get(bsc, bsc_token_current);
  token = MALLOC(1 + strlen(ptr));
  if (!token) {
    pfile_log_syserr(pf, result_memory);
  } else {
    pfile_source_t *psrc;
    char           *fpath;
    result_t        rc;

    strcpy(token, ptr);
    psrc = pfile_source_get(pf);

    fpath = MALLOC(1 + FILENAME_MAX);
    if (!fpath) {
      pfile_log_syserr(pf, result_memory);
    } else {
      ptr = 0;
      do {
        size_t      sz;
        const char *ptr_nxt; /* points to the next element */

        if (!ptr) {
          sz = 0;
          rc = result_ok;
          ptr_nxt = bsc_file_include_path_get(bsc);
        } else {
          ptr_nxt = strchr(ptr, ';');
          if (ptr_nxt) {
            sz = (size_t) (ptr_nxt - ptr);
            ptr_nxt++;
            if (!*ptr_nxt) {
              ptr_nxt = 0;
            }
          } else {
            sz = strlen(ptr);
          }
        }
        /* note: i'm not going to deal with the edge case of
                 the requested file being exactly the maximum
                 path length */
        if (sz + 1 + fname_sz >= FILENAME_MAX) {
          pfile_log(pf, pfile_log_err, BSC_MSG_PATH_TOO_LONG);
          rc = result_range;
        } else {
          if (sz) {
            /* copy over the path */
            memcpy(fpath, ptr, sz);
            /* add a path seperator if necessary */
            if ((fpath[sz-1] != '/') && (fpath[sz-1] != '\\')) {
              fpath[sz++] = '/';
            }
          } else {
            /* no path, hence no seperator */
            fpath[0] = 0;
          }
          memcpy(fpath + sz, fname, fname_sz);
          fpath[fname_sz + sz] = 0;
        }
        ptr = ptr_nxt;
        rc = pfile_source_set(pf, fpath);
        if (result_ok == rc) {
          pfile_log(pf, pfile_log_debug, BSC_MSG_DBG_INCLUDE, 
            (int) strlen(fpath), fpath);
        }
      } while (ptr && (result_ok != rc));
      if (result_ok != rc) {
        pfile_log(pf, pfile_log_err, BSC_MSG_CANNOT_OPEN, fname);
      } else {
        bsc_token_get(bsc, bsc_token_first);
        bsc_block_process(bsc, 0, 0, 0, 0);
        pfile_source_restore(pf, psrc);
      }
      FREE(fpath);
    }
    bsc_token_set(bsc, token);
    FREE(token);
  }
}


/*
 * NAME
 *   source_process
 *
 * DESCRIPTION
 *   process the source file
 *
 * PARAMETERS
 *   pf : file
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this parses out the PIC, FREQ and OSCILLATOR keywords
 *   before handing things off to block_process
 */
void bsc_source_process(bsc_file_t *bsc)
{
  pfile_t *pf;
  label_t *lbl;
  char     pic_include[32];

  strcpy(pic_include, "p16f877.inc"); /* the default */

  pf = bsc_file_pf_get(bsc);
  bsc_file_pass_reset(bsc); /* bump the pass counter */
  /* get the PIC if present, otherwise assume 16f877 */
  pfile_log(pf, pfile_log_info, BSC_MSG_COMPILING_PASS, 
      bsc_file_pass_get(bsc));
  if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, 
        "pic")) {
    const char *ptr;

    ptr = bsc_token_hack(bsc);
    if (strlen(ptr) > sizeof(pic_include) - 6) {
      pfile_log(pf, pfile_log_err, BSC_MSG_SYNTAX_ERROR);
    } else {
      sprintf(pic_include, "p%s.inc", ptr);
    }
    bsc_token_get(bsc, bsc_token_next);
    bsc_statement_next(bsc);
  }
  /* get the frequency if present, otherwise assume 20000 */
  if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, 
        "freq")) {
    variable_t *freq;

    freq = pfile_variable_find(pf, pfile_log_err, PFILE_LENGTH_DEF, 
      "_freq");
    if (freq) {
      value_t *tmp;

      bsc_token_get(bsc, bsc_token_next);
      if (result_ok == bsc_expr_parse(bsc, 0, &tmp)) {
        if (!(value_flag_test(tmp, VARIABLE_FLAG_CONST))) {
          pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
        } else {
          variable_const_set(freq, 0, value_const_get(tmp));
        }
        value_release(tmp);
      }
      variable_release(freq);
      bsc_statement_next(bsc);
    }
  }
  if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, 
        "oscillator")) {
    if (bsc_file_pass_get(bsc) == 1) {
      if (bsc_token_is(bsc, bsc_token_next, pfile_log_none, 
            "crystal")) {
        bsc_token_get(bsc, bsc_token_next);
      } else if (bsc_token_is(bsc, bsc_token_current, pfile_log_none,
            "rc")) {
        bsc_token_get(bsc, bsc_token_next);
      } else {
        pfile_log(pf, pfile_log_err, BSC_MSG_OSCILLATOR_BAD);
      }
    } else {
      bsc_statement_skip(bsc);
    }
    bsc_statement_next(bsc);
  }

  /* the entry point is named, ``_main'' */
  lbl = 0;
  if (1 == bsc_file_pass_get(bsc)) {
    if (result_ok == pfile_label_alloc(pf, "_main", 0, &lbl)) {
      pfile_user_entry_set(pf, lbl);
    }
  } else {
    lbl = pfile_label_find(pf, pfile_log_err, "_main");
  }
  if (lbl) {
    pfile_cmd_label_add(pf, lbl);
    label_release(lbl);
  }
  bsc_include_process(bsc, 0, pic_include);
  bsc_block_process(bsc, 0, 0, 0, 0);
  pfile_cmd_special_add(pf, 0, cmd_type_end);
}

void bsc_file_process(bsc_file_t *bsc)
{
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);
  pic_init(pf); /* intialize the pic-specific bits */
  bsc_source_process(bsc);
  if (!pfile_errct_get(pf)) {
    bsc_isr_type_t isr_type;

    pfile_label_fixup(pf);
    pfile_variable_fixup(pf);
    for (isr_type = bsc_isr_type_first;
         isr_type < bsc_isr_type_ct;
         isr_type++) {
      label_t *isr;

      if (result_ok == bsc_file_isr_get(bsc, isr_type, boolean_false, &isr)) {
        if (label_flag_test(isr, LABEL_FLAG_USED)) {
          pfile_isr_entry_set(pf, isr);
          label_release(isr);
          break;
        } else {
          label_release(isr);
        }
      }
    }
  }
  if (!pfile_errct_get(pf)) {
    bsc_source_process(bsc);
#if 0
    bsc_printfn_create(bsc);
#endif
    if (!pfile_errct_get(pf)) {
      pfile_cmd_dump(pf);
    }
  }
  if (pfile_flag_test(pf, PFILE_FLAG_DEBUG)) {
    pfile_label_dump(pf);
    pfile_variable_dump(pf);
  }
}

static void bsc_parse_pragma(bsc_file_t *bsc, unsigned sub, size_t param_ct,
    const param_t *params)
{
  pfile_t *pf;

  pf = bsc_file_pf_get(bsc);
  if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "warn")) {
    if (bsc_token_is(bsc, bsc_token_next, pfile_log_none, "on")) {
      pfile_flag_clr(pf, PFILE_FLAG_NOWARN);
      bsc_token_get(bsc, bsc_token_next);
    } else if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "off")) {
      pfile_flag_set(pf, PFILE_FLAG_NOWARN);
      bsc_token_get(bsc, bsc_token_next);
    } else {
      pfile_log(pf, pfile_log_err, BSC_MSG_ON_OR_OFF);
    }
  } else {
    /* pass it off to the PIC for further processing */
    pic_parse_pragma(bsc);
  }
}

static void bsc_parse_include(bsc_file_t *bsc, unsigned sub, size_t param_ct,
    const param_t *params)
{
  const char *ptr;
  pfile_t    *pf;

  pf = bsc_file_pf_get(bsc);

  if (bsc_token_is_eos(bsc)) {
    pfile_log(pf, pfile_log_err, BSC_MSG_FILENAME_BAD);
  } else {
    size_t sz;

    ptr = bsc_token_get(bsc, bsc_token_current);
    if (*ptr != '"') {
      pfile_log(pf, pfile_log_err, BSC_MSG_EXPECTED, "\"");
    } else {
      ptr++;
      sz = strlen(ptr);

      if (sz && ('\"' == ptr[sz-1])) {
        sz--;
      }
      if (!sz) {
        pfile_log(pf, pfile_log_err, BSC_MSG_FILENAME_BAD);
      } else {
        bsc_include_process(bsc, sz, ptr);
      }
    }
    bsc_token_get(bsc, bsc_token_next);
  }
}

