/**********************************************************
 **
 ** bsc_vdef.c : BSC variable defintion parsing
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>

#include "variable.h"
#include "ptype.h"
#include "pfile.h"

#include "bsc_parm.h"
#include "bsc_blck.h"
#include "bsc_tokn.h"
#include "bsc_expr.h"
#include "bsc_vdef.h"

/*
 * NAME
 *   bsc_parse_def
 *
 * DESCRIPTION
 *   variable definition
 *
 * PARAMETERS
 *   pf       : parser file
 *   param_ct : 0
 *   params   : 0
 *   sz       : variable size (1, 2, 4)
 *   flags    : variable flags
 *
 * RETURN
 *   none
 *
 * NOTES
 *   create variables, format : {type} id[(positive constant expr > 0)][,...]
 */
static void bsc_parse_def(bsc_file_t *bsc, flag_t oflags,
  variable_sz_t sz)
{
  size_t            ct;
  pfile_t          *pf;

  /* only execute this on the first pass. after that just skip it */
  pf = bsc_file_pf_get(bsc);
  for (ct = 0;
       ((0 == ct)
         || (bsc_token_is(bsc, bsc_token_current, pfile_log_none, ",")));
       ct++) {
    const char *ptr;
    variable_t *var;
    flag_t      flags;

    flags = oflags;

    ptr = bsc_token_get(bsc, (0 == ct) ? bsc_token_current : bsc_token_next);
    if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "volatile")) {
      flags |= VARIABLE_FLAG_VOLATILE;
      ptr = bsc_token_get(bsc, bsc_token_next);
    }

    if (('_' != *ptr) && !ISALPHA(*ptr)) {
      pfile_log(pf, pfile_log_err, BSC_MSG_VARIABLE_EXPECTED);
    } else {
      value_t *dim;
      char     varname[64];
      size_t   varname_sz;

      varname_sz = 1 + strlen(ptr);
      if (varname_sz >= sizeof(varname)) {
        varname_sz = sizeof(varname) - 1;
      }
      memcpy(varname, ptr, varname_sz);
      varname[varname_sz] = 0;

      var = pfile_variable_find(pf, 
        (1 == bsc_file_pass_get(bsc)) ? pfile_log_none : pfile_log_err, 
        PFILE_LENGTH_DEF, varname);
      dim = 0;
      if (var) {
        if (1 == bsc_file_pass_get(bsc)) {
          pfile_log(pf, pfile_log_err, BSC_MSG_MULTI_DEF, ptr);
        } else if (!variable_flag_test(var, VARIABLE_FLAG_ASSIGNED)) {
          if (!variable_flag_test(var, VARIABLE_FLAG_USED)) {
            pfile_log(pf, pfile_log_warn, BSC_MSG_NEVER_USED,
                variable_name_get(var));
          } else if (!variable_flag_test(var, VARIABLE_FLAG_VOLATILE)
              && !variable_flag_test(var, VARIABLE_FLAG_CONST)) {
            pfile_log(pf, pfile_log_warn, BSC_MSG_NEVER_ASSIGNED,
                variable_name_get(var));
          }
        } else if (!variable_flag_test(var, VARIABLE_FLAG_USED)) {
          pfile_log(pf, pfile_log_warn, BSC_MSG_ASSIGN_NOT_USED,
              variable_name_get(var));
        }
      }
      if (bsc_token_is(bsc, bsc_token_next, pfile_log_none, "(")) {
        bsc_token_get(bsc, bsc_token_next);
        if (result_ok == bsc_expr_parse(bsc, 0, &dim)) {
          if (!value_is_const(dim)) {
            pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
            value_release(dim);
            dim = 0;
          } else if (value_const_get(dim) == 0) {
            pfile_log(pf, pfile_log_err, BSC_MSG_ARRAY_SZ_ZERO);
            value_release(dim);
            dim = 0;
          } else if ((value_const_get(dim) > 255) 
              && (flags & VARIABLE_FLAG_CONST)) {
            pfile_log(pf, pfile_log_err, 
              BSC_MSG_LOOKUP_TOO_BIG);
          }
          /* get a constant expression */
          if (bsc_token_is(bsc, bsc_token_current, pfile_log_err, ")")) {
            bsc_token_get(bsc, bsc_token_next);
          }
        }
      }
      if (var) {
#if 0
        variable_release(var);
        var = 0;
#endif
      } else {
        pfile_variable_alloc(bsc_file_pf_get(bsc),
          varname, flags, sz, (dim) ? value_const_get(dim) : 1,
          0, &var);
      }
      if (dim) {
        value_release(dim);
      }

      if (bsc_token_is(bsc, bsc_token_current,
        pfile_log_none, "@")) {
        value_t *pos;

        bsc_token_get(bsc, bsc_token_next);
        pos = 0;
        bsc_expr_parse(bsc, 0, &pos);
        if (pos) {
          if (!value_is_const(pos)) {
            pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
          } else if (var) {
            variable_base_set(var, value_const_get(pos));
          }
          value_release(pos);
        }
        if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "[")) {
          if (bsc_token_is(bsc, bsc_token_next, pfile_log_none, "all")) {
            if (var) {
              variable_region_set_all(var, (variable_region_t) -1);
            }
            bsc_token_get(bsc, bsc_token_next);
          } else {
            /* [region1[,region2...]] */
            bsc_token_get_t which;

            which = bsc_token_current;

            do {
              value_t *region;

              bsc_token_get(bsc, which); /* skip '[' or ',' */
              region = 0;
              bsc_expr_parse(bsc, 0, &region);
              if (region) {
                if (!value_is_const(region)) {
                  pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
                } else if (var) {
                  variable_region_set(var, value_const_get(region));
                }
                value_release(region);
              }
              which = bsc_token_next;
            } while (bsc_token_is(bsc, bsc_token_current, pfile_log_none, ","));
          }
          if (bsc_token_is(bsc, bsc_token_current, pfile_log_err, "]")) {
            bsc_token_get(bsc, bsc_token_next);
          }
        }
      }
      if (bsc_token_is(bsc, bsc_token_current, pfile_log_none, "=")) {
        boolean_t     is_multiple;
        variable_ct_t ii;
        variable_ct_t need;

        is_multiple = bsc_token_is(bsc, bsc_token_next, pfile_log_none, "{");
        need = (is_multiple) ? variable_ct_get(var) : 1;
        for (ii = 0;
             (ii < need)
             && (!ii 
               || bsc_token_is(bsc, bsc_token_current, pfile_log_none, ","));
             ii++) {
          value_t *n;

          if (ii || is_multiple) {
            /* skip either '{' or ',' */
            bsc_token_get(bsc, bsc_token_next);
          }
          if (result_ok == bsc_expr_parse(bsc, 0, &n)) {
            if (flags & VARIABLE_FLAG_CONST) {
              if (!value_is_const(n)) {
                pfile_log(pf, pfile_log_err, BSC_MSG_CONSTANT_EXPECTED);
              } else {
                if (var) {
                  variable_const_set(var, ii, value_const_get(n));
                }
              }
            } else if (var) {
              value_t *vval;
              int      rc;

              rc = value_alloc(&vval, var, 0, 
                  ii * variable_sz_get(var), variable_sz_get(var));
              if (0 != rc) {
                pfile_log_syserr(pf, rc);
              } else {
                pfile_cmd_op_add(pf, 0, operator_assign, &vval, n, 0);
                value_release(vval);
              }
            }
            value_release(n);
          }
        } 
        if (is_multiple
          && bsc_token_is(bsc, bsc_token_current, pfile_log_err, "}")) {
          bsc_token_get(bsc, bsc_token_next);
        }
        if (var && (ii < variable_ct_get(var))) {
          pfile_log(pf, 
            (flags & VARIABLE_FLAG_CONST) ? pfile_log_err : pfile_log_warn, 
            BSC_MSG_ARRAY_INIT,
            ii, variable_ct_get(var), variable_name_get(var));
        }
      }
      variable_release(var);
    }
  }
}

void bsc_parse_vardef(bsc_file_t *bsc, unsigned sub, size_t param_ct, 
  const param_t *params)
{
  variable_sz_t     sz;
  flag_t            flags;

  sz    = 0;
  flags = 0;
  switch ((bsc_vardef_t) sub) {
    case bsc_vardef_char:
    case bsc_vardef_byte:   flags = 0;                    sz = 1; break;
    case bsc_vardef_sbyte:  flags = VARIABLE_FLAG_SIGNED; sz = 1; break;
    case bsc_vardef_word:   flags = 0;                    sz = 2; break;
    case bsc_vardef_sword:  flags = VARIABLE_FLAG_SIGNED; sz = 2; break;
    case bsc_vardef_dword:  flags = 0;                    sz = 4; break;
    case bsc_vardef_sdword: flags = VARIABLE_FLAG_SIGNED; sz = 4; break;
  }
  bsc_parse_def(bsc, flags, sz);
}


void bsc_parse_constdef(bsc_file_t *bsc, unsigned sub, size_t param_ct,
  const param_t *params)
{
  size_t ii;
  static const struct {
    const char   *key;
    flag_t        flag;
    variable_sz_t sz;
  } types[] = {
    { "byte",                      0, 1 },
    { "sbyte",  VARIABLE_FLAG_SIGNED, 1 },
    { "word",                      0, 2 },
    { "sword",  VARIABLE_FLAG_SIGNED, 2 },
    { "dword",                     0, 4 },
    {" sdword", VARIABLE_FLAG_SIGNED, 4 }
  };

  for (ii = 0; 
       (ii < COUNT(types))
       && !bsc_token_is(bsc, bsc_token_current, pfile_log_none,
         types[ii].key);
       ii++)
    ;
  if (ii == COUNT(types)) {
    pfile_log(bsc_file_pf_get(bsc), pfile_log_err, BSC_MSG_TYPE_BAD);
  } else {
    bsc_token_get(bsc, bsc_token_next);
    bsc_parse_def(bsc, types[ii].flag | VARIABLE_FLAG_CONST, 
      types[ii].sz);
  }
}

