/************************************************************
 **
 ** cmd.h : p-code structure handling declarations
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#ifndef cmd_h__
#define cmd_h__

#include <stdio.h>
#include "ptype.h"
#include "operator.h"
#include "value.h"

typedef enum cmd_type_ {
  cmd_type_nop,         /* no operation   */
  cmd_type_operator,    /* operator       */
  cmd_type_branch,      /* control        */
  cmd_type_intrinsic,   /* intrinsic      */
  cmd_type_end,         /* end of program */
  cmd_type_sleep,       /* sleep          */
  cmd_type_clear,       /* clear memory   */
  cmd_type_isr_cleanup
} cmd_type_t;

typedef enum {
  cmd_branchtype_none,
  cmd_branchtype_goto,
  cmd_branchtype_call,
  cmd_branchtype_return
} cmd_branchtype_t;

typedef enum {
  cmd_branchcond_none, /* unconditional */
  cmd_branchcond_false,
  cmd_branchcond_true
} cmd_branchcond_t;

typedef struct cmd_ cmd_t;
typedef void (*cmd_dump_t)(cmd_t *cmd, FILE *dst);
typedef void (*cmd_free_t)(cmd_t *cmd);

typedef struct {
  cmd_free_t free_fn;
  cmd_dump_t dump_fn;
} cmd_vtbl_t;

#define CMD_FLAG_REACHABLE 0x0001 /* code is reachable       */
#define CMD_FLAG_INTERRUPT 0x0002 /* used at interrupt level */
#define CMD_FLAG_USER      0x0004 /* used at user level      */
struct cmd_ {
  struct cmd_       *link;
  flag_t             flags;
  pfile_source_t    *src;
  unsigned           line;  /* line # (from file) */
  label_t           *label; /* line label         */
  cmd_type_t         type;  /* type of command    */
  const cmd_vtbl_t  *vtbl;
};

void *cmd_alloc(label_t *label, cmd_type_t type, 
  const cmd_vtbl_t *vtbl, size_t sz);

cmd_t *cmd_op_alloc(label_t *label, operator_t op,
  value_t *dst, value_t *val1, value_t *val2);
cmd_t *cmd_branch_alloc(label_t *label, 
  cmd_branchtype_t type, cmd_branchcond_t cond, label_t *dst, 
  value_t *var);

void      cmd_flag_set(cmd_t *cmd, flag_t flag);
boolean_t cmd_flag_test(const cmd_t *cmd, flag_t flag);
void      cmd_flag_clr(cmd_t *cmd, flag_t flag);

void cmd_flag_set_all(cmd_t *cmd, flag_t flags);

void cmd_free(cmd_t *cmd);
void cmd_list_free(cmd_t *cmd);
void cmd_list_reverse(cmd_t **cmd);

/* common operations */
void         cmd_dump(cmd_t *cmd, FILE *dst);
void         cmd_free(cmd_t *cmd);

cmd_t       *cmd_link_get(const cmd_t *cmd);
void         cmd_link_set(cmd_t *cmd, cmd_t *lnk);

cmd_type_t cmd_type_get(const cmd_t *cmd);
void       cmd_type_set(cmd_t *cmd, cmd_type_t type);

unsigned   cmd_line_get(const cmd_t *cmd);
void       cmd_line_set(cmd_t *cmd, unsigned line);

pfile_source_t *cmd_source_get(const cmd_t *cmd);
void cmd_source_set(cmd_t *cmd, pfile_source_t *src);

label_t *cmd_label_get(const cmd_t *cmd);
void              cmd_label_set(cmd_t *cmd, label_t *var);

/* branch operations */
cmd_branchtype_t cmd_brtype_get(const cmd_t *cmd);
void             cmd_brtype_set(cmd_t *cmd, cmd_branchtype_t brtype);

cmd_branchcond_t cmd_brcond_get(const cmd_t *cmd);
void             cmd_brcond_set(cmd_t *cmd, cmd_branchcond_t brcond);

label_t *cmd_brdst_get(cmd_t *cmd);
void           cmd_brdst_set(cmd_t *cmd, label_t *dst);

value_t *cmd_brval_get(const cmd_t *cmd);
void              cmd_brval_set(cmd_t *cmd, value_t *val);

/* operator operations */
operator_t        cmd_optype_get(const cmd_t *cmd);
void              cmd_optype_set(cmd_t *cmd, operator_t op);

value_t *cmd_opdst_get(const cmd_t *cmd);
void              cmd_opdst_set(cmd_t *cmd, value_t *dst);

value_t *cmd_opval1_get(const cmd_t *cmd);
void              cmd_opval1_set(cmd_t *cmd, value_t *val1);

value_t *cmd_opval2_get(const cmd_t *cmd);
void              cmd_opval2_set(cmd_t *cmd, value_t *val2);

#ifdef MSDOS
#include "cmdopdc.h"
#include "cmdbrdc.h"
#endif

#endif /* cmd_h__ */

