/**********************************************************
 **
 ** cmd_op.c : the cmd_type_operator functions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/

#include "value.h"
#include "cmd.h"
#include "cmdopdc.h"
#include "label.h"

/* expression, dst = val1 op val2 */
static void cmd_op_dump(cmd_t *cmd, FILE *dst)
{
  const char     *ostr;

  ostr = "{unknown}";
  switch (cmd_optype_get(cmd)) {
    case operator_null:         ostr = "{null}"; break;
    case operator_add:          ostr = "+"; break;
    case operator_sub:          ostr = "-"; break;
    case operator_mul:          ostr = "*"; break;
    case operator_div:          ostr = "/"; break;
    case operator_mod:          ostr = "%"; break;
    case operator_lt:           ostr = "<"; break;
    case operator_le:           ostr = "<="; break;
    case operator_eq:           ostr = "=="; break;
    case operator_ne:           ostr = "<>"; break;
    case operator_ge:           ostr = ">="; break;
    case operator_gt:           ostr = ">"; break;
    case operator_andl:         ostr = "&&"; break;
    case operator_orl:          ostr = "||"; break;
    case operator_notl:         ostr = "!";  break;
    case operator_andb:         ostr = "&";  break;
    case operator_orb:          ostr = "|"; break;
    case operator_xorb:         ostr = "^"; break;
    case operator_cmpb:         ostr = "~"; break;
    case operator_assign:       ostr = "="; break;
    case operator_neg:          ostr = "neg"; break;
    case operator_bitset:       ostr = "bitset"; break;
    case operator_bitclr:       ostr = "bitclr"; break;
    case operator_bittst:       ostr = "bittst"; break;
    case operator_shift_left:   ostr = "<<"; break;
    case operator_shift_right:  ostr = ">>"; break;
    case operator_logical:      ostr = "!!"; break;
    case operator_incr:         ostr = "++"; break;
    case operator_decr:         ostr = "--"; break;
  }
  fprintf(dst, "%s ", ostr);
  if (cmd_opdst_get(cmd)) {
    value_dump(cmd_opdst_get(cmd), dst);
    if (cmd_opval1_get(cmd)) {
      fprintf(dst, ", ");
      value_dump(cmd_opval1_get(cmd), dst);
      if (cmd_opval2_get(cmd)) {
        fprintf(dst, ", ");
        value_dump(cmd_opval2_get(cmd), dst);
      }
    }
  }
}

static void cmd_op_free(cmd_t *cmd)
{
  cmd_opdst_set(cmd, 0);
  cmd_opval1_set(cmd, 0);
  cmd_opval2_set(cmd, 0);
}

static const cmd_vtbl_t cmd_op_vtbl = {
  cmd_op_free,
  cmd_op_dump
};

cmd_t *cmd_op_alloc(label_t *label, operator_t op,
  value_t *dst, value_t *val1, value_t *val2)
{
  cmd_op_t *cop;

  cop = cmd_alloc(label, cmd_type_operator, &cmd_op_vtbl, sizeof(*cop));
  if (cop) {
    cop->op   = op;
    cop->dst  = 0;
    cop->val1 = 0;
    cop->val2 = 0;

    cmd_opdst_set(&cop->common, dst);
    cmd_opval1_set(&cop->common, val1);
    cmd_opval2_set(&cop->common, val2);
  }
  return &cop->common;
}


/* operator operations */
operator_t        cmd_optype_get(const cmd_t *cmd)
{
  return (cmd_type_operator == cmd_type_get(cmd))
    ? ((const cmd_op_t *) cmd)->op
    : operator_null;
}

void              cmd_optype_set(cmd_t *cmd, operator_t op)
{
  if (cmd_type_operator == cmd_type_get(cmd)) {
    ((cmd_op_t *) cmd)->op = op;
  }
}


value_t *cmd_opdst_get(const cmd_t *cmd)
{
  return (cmd_type_operator == cmd_type_get(cmd))
    ? ((const cmd_op_t *) cmd)->dst
    : 0;
}

void cmd_opdst_set(cmd_t *cmd, value_t *dst)
{
  if (cmd_type_operator == cmd_type_get(cmd)) {
    cmd_op_t *cop;

    cop = (cmd_op_t *) cmd;
    if (dst) {
      value_assign_ct_bump(dst, ctr_bump_incr);
      value_lock(dst);
    }
    if (cop->dst) {
      value_assign_ct_bump(cop->dst, ctr_bump_decr);
      value_release(cop->dst);
    }
    cop->dst = dst;
  }
}

value_t *cmd_opval1_get(const cmd_t *cmd)
{
  return (cmd_type_operator == cmd_type_get(cmd))
    ? ((const cmd_op_t *) cmd)->val1
    : 0;
}

void cmd_opval1_set(cmd_t *cmd, value_t *val1)
{
  if (cmd_type_operator == cmd_type_get(cmd)) {
    cmd_op_t *cop;

    cop = (cmd_op_t *) cmd;
    if (val1) {
      value_lock(val1);
      /* don't bump the use count of constants because
       * otherwise lookup tables might unnecessarily be
       * added */
      /*if (!value_is_const(val1)) */{
        value_use_ct_bump(val1, ctr_bump_incr);
      }
    }
    if (cop->val1) {
      if (!value_is_const(cop->val1)) {
        value_use_ct_bump(cop->val1, ctr_bump_decr);
      }
      value_release(cop->val1);
    }
    cop->val1 = val1;
  }
}

value_t *cmd_opval2_get(const cmd_t *cmd)
{
  return (cmd_type_operator == cmd_type_get(cmd)) 
    ? ((const cmd_op_t *) cmd)->val2
    : 0;
}

void cmd_opval2_set(cmd_t *cmd, value_t *val2)
{
  if (cmd_type_operator == cmd_type_get(cmd)) {
    cmd_op_t *cop;

    cop = (cmd_op_t *) cmd;
    if (val2) {
      value_lock(val2);
      /*if (!value_is_const(val2)) */{
        value_use_ct_bump(val2, ctr_bump_incr);
      }
    }
    if (cop->val2) {
      if (!value_is_const(cop->val2)) {
        value_use_ct_bump(cop->val2, ctr_bump_decr);
      }
      value_release(cop->val2);
    }
    cop->val2 = val2;
  }
}




