/**********************************************************
 **
 ** cmdoptmz.c : command optimizier functions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <assert.h>

#include "pfile.h"
#include "cmd.h"
#include "label.h"
#include "cmdoptmz.h"

/*
 * NAME
 *   cmd_opt_branch_fixup
 *
 * DESCRIPTION
 *   make sure all branches point to the branch master
 *
 * PARAMETERS
 *   cmd : head of cmd list
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void cmd_opt_branch_fixup(cmd_t *cmd)
{
  while (cmd) {
    label_t *lbl;

    lbl = cmd_brdst_get(cmd);
    if (lbl && (label_master_get(lbl) != lbl)) {
      lbl = label_master_get(lbl);
      cmd_brdst_set(cmd, lbl);
    }
    cmd = cmd_link_get(cmd);
  }
}

/*
 * NAME
 *   cmd_opt_call_return
 *
 * DESCRIPTION
 *   call/return optimization
 *
 * PARAMETERS
 *   cmd : head of the cmd list
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this optimizes the following:
 *      call followed by return --> goto
 */
void cmd_opt_call_return(cmd_t *cmd)
{
  cmd_t *cmd_nxt;

  cmd_nxt = cmd_link_get(cmd);
  while (cmd_nxt) {
    if ((cmd_branchtype_call == cmd_brtype_get(cmd))
      && (cmd_branchtype_return == cmd_brtype_get(cmd_nxt))) {
      /* this is a call/return pair; change cmd to a goto */
      cmd_brtype_set(cmd, cmd_branchtype_goto);
    }
    cmd = cmd_nxt;
    cmd_nxt = cmd_link_get(cmd);
  }
}

/*
 * NAME
 *   cmd_opt_dst
 *
 * DESCRIPTION
 *   destination optimization
 *
 * PARAMETERS
 *   cmd : head of the cmd list
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this optimizes branches to other branches. for example:
 *          goto l1:
 *      l1: goto l2
 *   is changed to ``goto l2:''
 *   if the goto destination is a return, the goto is replaced
 *     with the return
 *   if a call destination is a return, the line is removed
 */
static label_t *dptr_bump(label_t *dptr)
{
  if (dptr && (cmd_branchtype_goto == label_brtype_get(dptr))) {
    dptr = label_brdest_get(dptr);
  } else {
    dptr = 0;
  }
  if (dptr && (cmd_branchtype_goto == label_brtype_get(dptr))) {
    dptr = label_brdest_get(dptr);
  } else {
    dptr = 0;
  }
  return dptr;
}


void cmd_opt_dst(pfile_t *pf, cmd_t *cmd_top)
{
  cmd_t *cmd;

  cmd = cmd_top;
  while (cmd) {
    if ((cmd_branchtype_goto == cmd_brtype_get(cmd))
      || (cmd_branchtype_call == cmd_brtype_get(cmd))) {
      /* what's the destination look like? */
      label_t *dst;
      label_t *dptr;

      dst = cmd_brdst_get(cmd);
      dptr = dptr_bump(dst);

      while ((dptr != dst) 
        && (cmd_branchtype_goto == label_brtype_get(dst))) {
        dst = label_brdest_get(dst);
        dptr = dptr_bump(dptr);
      }
      if (dptr == dst) {
        pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_LOOP);
        /* just for fun, dump the loop here */
        do {
          dst = label_brdest_get(dst);
          pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_LOOP_DST, 
            label_name_get(dst));
        } while (dptr != dst);
      }
      if (cmd_branchtype_return == label_brtype_get(dst)) {
        if (cmd_branchtype_goto == cmd_brtype_get(cmd)) {
          /* goto destination is a return, remove the goto
             and make it a simple return */
          cmd_brtype_set(cmd, cmd_branchtype_return);
          cmd_brdst_set(cmd, 0);
        } else {
          cmd_brdst_set(cmd, 0);
          cmd_brtype_set(cmd, cmd_branchtype_none);
          cmd_brcond_set(cmd, cmd_branchcond_none);
        }
      } else {
        cmd_brdst_set(cmd, label_master_get(dst));
      }
    }
    cmd = cmd_link_get(cmd);
  }
}

/*
 * NAME
 *   cmd_label_find
 *
 * DESCRIPTION
 *   find a label in the command list
 *
 * PARAMETERS
 *   cmd : head of list
 *
 * RETURN
 *   label, or 0
 *
 * NOTES
 */
cmd_t *cmd_label_find(cmd_t *cmd, const label_t *lbl)
{
  if (!lbl) {
    cmd = 0;
  } else {
    while (cmd && (cmd_label_get(cmd) != lbl)) {
      cmd = cmd_link_get(cmd);
    }
  }
  return cmd;
}

/*
 * NAME
 *   cmd_analyze
 *
 * DESCRIPTION
 *   analyze the command list for dead code
 *
 * PARAMETERS
 *   cmd_head : head of the cmd list
 *   cmd      : current entry
 *   flag     : CMD_FLAG_USER      : analyzing user area
 *              CMD_FLAG_INTERRUPT : analyzing an ISR
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this doesn't solve the NP problem, it simply takes each
 *   path to determine if it's possible to get there from here
 */
static size_t depth_max;
void cmd_analyze(cmd_t *cmd_top, cmd_t *cmd, flag_t flag, size_t depth)
{
  if (depth > depth_max) {
    depth_max = depth;
  }
  while (cmd && !cmd_flag_test(cmd, flag)) {
    cmd_flag_set(cmd, CMD_FLAG_REACHABLE);
    cmd_flag_set(cmd, flag);
    if (cmd_type_branch == cmd_type_get(cmd)) {
      switch (cmd_brtype_get(cmd)) {
        case cmd_branchtype_goto:
        case cmd_branchtype_call:
          {
            /* look for the label */
            cmd_t   *ptr;

            ptr = cmd_label_find(cmd_top, cmd_brdst_get(cmd));
            if ((cmd_branchtype_goto == cmd_brtype_get(cmd))
              && (cmd_brdst_get(cmd) == cmd_label_get(cmd_link_get(ptr)))) {
              /* this is unreachable! */
              cmd_flag_clr(ptr, CMD_FLAG_REACHABLE);
              cmd_flag_clr(ptr, flag);
              cmd = cmd_link_get(cmd);
            } else if ((cmd_branchtype_goto == cmd_brtype_get(cmd))
                && (cmd_branchcond_none == cmd_brcond_get(cmd))) {
              /* unconditional goto, simply change cmd */
              cmd = ptr;
            } else {
              /* recurse using ptr, then continue */
              cmd_analyze(cmd_top, ptr, flag, depth + 1);
              cmd = cmd_link_get(cmd);
            }
          }
          break;
        case cmd_branchtype_return:
          if (cmd_branchcond_none == cmd_brcond_get(cmd)) {
            /* at the end of the chain, so exit */
            cmd = 0;
            break;
          }
          /* fall through */
        case cmd_branchtype_none: 
          cmd = cmd_link_get(cmd);
          break;
      }
    } else if (cmd_type_end == cmd_type_get(cmd)) {
      cmd = 0; /* we're done */
    } else {
      cmd = cmd_link_get(cmd);
    }
  }
}

/*
 * NAME
 *   cmd_opt_branch2
 *
 * DESCRIPTION
 *   Once the code analyzer has run, look for any lines that simply
 *   branch to the next line & either remove them are mark them
 *   ``unreachable''
 *   Also, look for the sequence
 *              goto cond label1
 *              goto label2
 *      label1: ...
 *   and replace with:
 *      goto !cond label2
 *
 * PARAMETERS
 *   cmd : top of command chain
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void cmd_opt_branch2(pfile_t *pf, cmd_t *cmd_top)
{
  cmd_t *cmd;

  cmd = cmd_top;
  while (cmd) {
    cmd_t *cmd_nxt;
    cmd_t *cmd_nxt_nxt;

    for (cmd_nxt = cmd_link_get(cmd);
         cmd_nxt && !cmd_flag_test(cmd_nxt, CMD_FLAG_REACHABLE);
         cmd_nxt = cmd_link_get(cmd_nxt))
      ;
    for (cmd_nxt_nxt = cmd_link_get(cmd_nxt);
         cmd_nxt_nxt && !cmd_flag_test(cmd_nxt_nxt, CMD_FLAG_REACHABLE);
         cmd_nxt_nxt = cmd_link_get(cmd_nxt_nxt))
      ;
    if (cmd_nxt 
        && cmd_nxt_nxt 
        && !cmd_label_get(cmd_nxt)
        && cmd_label_get(cmd_nxt_nxt)
        && (cmd_branchtype_goto == cmd_brtype_get(cmd))
        && (cmd_branchtype_goto == cmd_brtype_get(cmd_nxt))
        && (cmd_branchcond_none != cmd_brcond_get(cmd))
        && (cmd_branchcond_none == cmd_brcond_get(cmd_nxt))
        && (cmd_brdst_get(cmd) == cmd_label_get(cmd_nxt_nxt))) {
      switch (cmd_brcond_get(cmd)) {
        case cmd_branchcond_none: break; /* cannot happen */
        case cmd_branchcond_true: cmd_brcond_set(cmd, cmd_branchcond_false); break;
        case cmd_branchcond_false: cmd_brcond_set(cmd, cmd_branchcond_true); break;
      }
      cmd_brdst_set(cmd, cmd_brdst_get(cmd_nxt));
      cmd_link_set(cmd, cmd_nxt_nxt);
      cmd_free(cmd_nxt);
      cmd_nxt = cmd_nxt_nxt;
    } else if (cmd_nxt 
        && (cmd_branchtype_goto == cmd_brtype_get(cmd))
        && (cmd_brdst_get(cmd) == cmd_label_get(cmd_nxt))
#if 0
        && ((cmd_label_get(cmd) != pfile_user_entry_get(pf))
          && (cmd_label_get(cmd) != pfile_isr_entry_get(pf)))) {
#else
      ) {
#endif
      /* cmd is a goto the following command, so get rid of it */
      /* note: cmd is NOT removed if it's label is the user
       *       or isr entry */
      label_t *lbl;
      cmd_t   *cmd_ptr;

      cmd_flag_set_all(cmd, 0);
      lbl = cmd_label_get(cmd);
      if (lbl) {
        /* must fixup anything pointing to this */
        for (cmd_ptr = cmd_top; cmd_ptr; cmd_ptr = cmd_link_get(cmd_ptr)) {
          if (lbl == cmd_brdst_get(cmd_ptr)) {
            cmd_brdst_set(cmd_ptr, 
              label_master_get(cmd_label_get(cmd_nxt)));
          }
        }
        if (lbl == pfile_user_entry_get(pf)) {
          pfile_user_entry_set(pf, cmd_label_get(cmd_nxt));
        }
        if (lbl == pfile_isr_entry_get(pf)) {
          pfile_isr_entry_set(pf, cmd_label_get(cmd_nxt));
        }
      }
    }
    cmd = cmd_nxt;
  }
}

/*
 * NAME
 *   cmd_opt_counters_fix
 *
 * DESCRIPTION
 *   after analyzing, reduce the assign/use counters on anything
 *   that is not reachable
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static void val_fix(value_t *val, int assign)
{
  if (val) {
    if (assign) {
      value_assign_ct_bump(val, ctr_bump_decr);
    } else {
      value_use_ct_bump(val, ctr_bump_decr);
    }
  }
}

void cmd_opt_counters_fix(cmd_t *cmd)
{
  while (cmd) {
    if (!cmd_flag_test(cmd, CMD_FLAG_REACHABLE)) {
      val_fix(cmd_opdst_get(cmd), 1);
      val_fix(cmd_opval1_get(cmd), 0);
      val_fix(cmd_opval2_get(cmd), 0);
      val_fix(cmd_brval_get(cmd), 0);
    }
    cmd = cmd_link_get(cmd);
  }

}

