/************************************************************
 **
 ** pfile.c : p-code file definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>

#include <assert.h>

#include "slist.h"
#include "label.h"
#include "pfilesrc.h"
#include "pfile.h"

#include "cmd.h"
#include "cmdoptmz.h"
#include "pic.h"
#include "pfiledcl.h"

/*
 * NAME
 *   pfile_open
 *
 * DESCRIPTION
 *   create the pfile structure
 *
 * PARAMETERS
 *   dst : [out] holds result
 *   name  : source name
 *   f_src : source file    (input)
 *   f_asm : assembly file  (output)
 *   f_lst : listinf file   (output)
 *   f_hex : intel hex file (output)
 *   f_log : log file       (output)
 *
 * RETURN
 *   0 : no error
 *
 * NOTES
 */
result_t pfile_open(pfile_t **dst, 
    FILE *f_asm, FILE *f_lst, FILE *f_hex, FILE *f_log, flag_t flags)
{
  pfile_t *pf;
  int      rc;

  pf = MALLOC(sizeof(*pf));
  if (0 == pf) {
    rc = result_memory;
  } else {
    rc = result_ok;
    pf->src   = 0;
    pf->f_asm = f_asm;
    pf->f_lst = f_lst;
    pf->f_hex = f_hex;
    pf->f_log = f_log;

    /* the following are set by token_get: 
       ch_unget
       line_no
       token_len
       token
    */   
    pf->errct = 0;
    pf->cmds = 0;
    variable_list_init(&pf->var_active);
    variable_list_init(&pf->var_const);

    variable_list_init(&pf->var_temp);
    variable_list_init(&pf->var_isr);

    label_list_init(&pf->label_user);
    pf->label_temp_ct = 0;
    label_list_init(&pf->label_temp);
    pf->label_temp_ptr  = 0;
    pf->label_next = 0;

    pf->label_main = 0;
    pf->label_isr  = 0;
    pf->label_isr_preamble = 0;

    /* allocate the predefined labels and variables */
    /* _isr_cleanup --> user jumps to this when isr processing
                        is finished
       _freq        --> clock frequency of the PIC */ 
    pfile_label_alloc(pf, "_isr_cleanup", 0, 0);
    pfile_label_alloc(pf, "_isr_preamble", 0, &pf->label_isr_preamble);
    {
      variable_const_t n;
      
      n = 20000;
      pfile_variable_alloc(pf, "_freq", VARIABLE_FLAG_CONST, 4, 1, &n, 0);
    }

    pf->hex.pc = 0;
    pf->hex.ct = 0;
    pf->flags  = flags;

    pfile_oscillator_set(pf, oscillator_crystal);
    *dst = pf;
  }
  return rc;
}

/*
 * NAME
 *   pfile_hex_line_flush
 *
 * DESCRIPTION
 *   flush the current line in the hex buffer
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pfile_hex_line_flush(pfile_t *pf)
{
  if (pf->f_hex && pf->hex.ct) {
    unsigned ii;
    unsigned chk;

    fprintf(pf->f_hex, ":%02X%02X%02X00", 
        pf->hex.ct,
        (pf->hex.pc >> 8) & 0xff,
        (pf->hex.pc     ) & 0xff);
    chk = pf->hex.ct
        + ((pf->hex.pc >> 8) & 0xff)
        + ((pf->hex.pc     ) & 0xff);
    for (ii = 0; ii < pf->hex.ct; ii++) {
      chk += pf->hex.buf[ii];
      fprintf(pf->f_hex, "%02X", pf->hex.buf[ii]);
    }
    fprintf(pf->f_hex, "%02X\n", (unsigned char) (-chk));
  }
  pf->hex.ct = 0;
}

/*
 * NAME
 *   pfile_close
 *
 * DESCRIPTION
 *   close an open pfile
 *
 * PARAMETERS
 *   pf : returned from pfile_open
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void pfile_close(pfile_t *pf)
{
  cmd_list_free(pf->cmds);

  pfile_user_entry_set(pf, 0);
  pfile_isr_entry_set(pf, 0);
  label_release(pf->label_isr_preamble);

  label_list_reset(&pf->label_user, pf);
  label_list_reset(&pf->label_temp, pf);

  variable_list_reset(&pf->var_active, pf);
  variable_list_reset(&pf->var_const, pf);
  variable_list_reset(&pf->var_temp, pf);
  variable_list_reset(&pf->var_isr, pf);
  pfile_hex_line_flush(pf);
  if (pf->f_hex) {
    fprintf(pf->f_hex, ":00000001FF\n");
  }
  pfile_source_restore(pf, 0);
  FREE(pf);
}

/*
 * NAME
 *   pfile_reset
 *
 * DESCRIPTION
 *   reset the pfile descriptor
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this:
 *      1. free all commands
 *      2. reset the temp var ptr
 *      3. reset the temp label ptr
 *      4. rewind the source file
 */
void pfile_reset(pfile_t *pf)
{
  cmd_list_free(pf->cmds);
  pf->cmds = 0;
  pf->label_temp_ptr = label_list_head(&pf->label_temp);
  pf->label_next = 0;
  pfile_rewind(pf);
}

/*
 * NAME
 *   pfile_ch_get
 *
 * DESCRIPTION
 *   get the next character from the file
 *
 * PARAMETERS
 *   pf : returned from pfile_open
 *
 * RETURN
 *   character read, or EOF
 *
 * NOTES
 *   needed due to problems in some versions of ungetc()
 */
int pfile_ch_get(pfile_t *pf)
{
  return pfile_source_ch_get(pf->src);
}

/*
 * NAME
 *   pfile_ch_unget
 *
 * DESCRIPTION
 *   return a character to the file
 *
 * PARAMETERS
 *   pf : returned from pfile_open
 *   ch : last character read
 *
 * RETURN
 *   none
 *  
 * NOTES
 *   only one character can be unget()
 */
void pfile_ch_unget(pfile_t *pf, int ch)
{
  pfile_source_ch_unget(pf->src, ch);
}

/*
 * NAME
 *   pfile_constant_get
 *
 * DESCRIPTION
 *   return an unnamed constant whose value matches val
 *
 * PARAMETERS
 *   pf  :
 *   val : value to match
 *   dst : holds result on success
 *
 * RETURN
 *   0 : no error
 *
 * NOTES
 */

/* unnamed constant find */
typedef struct uncfind_data_ {
  variable_const_t n;
} uncfind_data_t;

static boolean_t uncfind_cb(void *arg, const variable_t *var, 
    const void *unused)
{
  uncfind_data_t *data;

  data = arg;

  return (variable_flag_test(var, VARIABLE_FLAG_CONST))
    && (data->n == variable_const_get(var, 0));
}

result_t pfile_constant_get(pfile_t *pf, variable_const_t n, value_t **dst)
{
  /* let's find a suitable constant or create one */
  result_t       rc;
  uncfind_data_t data;
  variable_t    *var;
  value_t       *val;

  data.n = n;

  /* now that we know what to look for, find it! */
  rc = result_ok;
  var = variable_list_find(&pf->var_const, uncfind_cb, &data, 0);
  if (!var) {
    /* need to create one! */
    rc = variable_alloc(&var, 0, VARIABLE_FLAG_CONST, 0, 1, &data.n);
    if (result_ok == rc) {
      variable_list_append(&pf->var_const, var);
    }
  }
  if (result_ok == rc) {
    rc = value_alloc(&val, var, 0, 0, variable_sz_get(var));
    variable_release(var);
  }
  if (result_ok != rc) {
    pfile_log_syserr(pf, rc);
  } else {
    if (dst) {
      *dst = val;
    } else {
      value_release(val);
    }
  }
  return rc;
}

/*
 * NAME
 *   pfile_value_alloc
 *
 * DESCRIPTION
 *   allocate a variable and a value to contain it
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
result_t pfile_value_alloc(pfile_t *pf, const char *name, flag_t flags,
    variable_sz_t sz, variable_ct_t ct, const variable_const_t *n, value_t **dst)
{
  result_t    rc;
  variable_t *var;

  rc = pfile_variable_alloc(pf, name, flags, sz, ct, n, &var);
  if (result_ok == rc) {
    value_t *val;
    variable_release(var);

    val = pfile_value_find(pf, pfile_log_err, strlen(name), name);
    if (val) {
      if (dst) {
        *dst = val;
      } else {
        value_release(val);
      }
    } else {
      rc = result_not_found;
    }
  }
  return rc;
}

/*
 * NAME
 *   pfile_variable_temp_get
 *
 * DESCRIPTION
 *   return a temporary variable of the correct size
 *
 * PARAMETERS
 *   pf  : returned by pfile_open()
 *   sz  : size needed
 *   dst : [out] holds result on success
 *
 * RETURN
 *   0 : no error
 *
 * NOTES
 *   [dst] will be locked on success
 */

static boolean_t tmpfind_cb(void *arg, const variable_t *var, const void *data)
{
  variable_sz_t sz;
  
  sz  = *(const variable_sz_t *) data;

  return (variable_expr_ct_get(var) == 0) 
    && (variable_sz_get(var) == sz);
}


static result_t pfile_variable_temp_get(pfile_t *pf, variable_sz_t sz, 
  flag_t flags, variable_t **dst)
{
  result_t    rc;
  variable_t *var;

  var = variable_list_find(&pf->var_temp, tmpfind_cb, 0, &sz);
  if (var) {
    rc = result_ok;
  } else {
    char tmp_name[32];

    sprintf(tmp_name, "_t%u", variable_list_ct_get(&pf->var_temp));
    rc = variable_alloc(&var, tmp_name, 0, sz, 1, 0);
    if (result_ok == rc) {
      variable_list_append(&pf->var_temp, var);
    }
  }
  if (result_ok == rc) {
    variable_flags_set_all(var, flags);
    *dst = var;
  } else {
    pfile_log_syserr(pf, rc);
  }
  return rc;
}

/*
 * NAME
 *   pfile_value_temp_get
 *
 * DESCRIPTION
 *   return a temporary; allocate a new one if necessary
 *
 * PARAMETERS
 *   pf    : 
 *   sz    : 
 *   flags : 
 *   dst   : [out] holds result on success
 *
 * RETURN
 *
 * NOTES
 */
result_t pfile_value_temp_get(pfile_t *pf, variable_sz_t sz, flag_t flags,
    value_t **dst)
{
  result_t    rc;
  variable_t *var;

  rc = pfile_variable_temp_get(pf, sz, flags, &var);
  if (result_ok == rc) {
    rc = value_alloc(dst, var, 0, 0, sz);
    variable_release(var);
    if (result_ok != rc) {
      pfile_log_syserr(pf, rc);
    }
  }
  return rc;
}


/*
 * NAME
 *   pfile_line
 *
 * DESCRIPTION
 *   return the current line #
 *
 * PARAMETERS
 *   pf : returned by pfile_open()
 *
 * RETURN
 *   current line #
 *
 * NOTES
 */
unsigned pfile_line_get(const pfile_t *pf)
{
  return pfile_source_line_get(pf->src);
}

/*
 * NAME
 *   pfile_log
 *
 * DESCRIPTION
 *   log a message
 *
 * PARAMETERS
 *   pf   : returned by pfile_open()
 *   plog : log level
 *   fmt  : log format
 *   ...
 *
 * RETURN
 *   none
 *
 * NOTES
 */
void pfile_log(pfile_t *pf, pfile_log_t plog, const char *fmt, ...)
{
  if ((pfile_log_none != plog) 
    && ((pfile_log_warn != plog) || !pfile_flag_test(pf, PFILE_FLAG_NOWARN))) {
    va_list     lst;
    FILE       *dst;
    const char *sev_str;

    if (pfile_log_err == plog) {
      pf->errct++;
    }

    if (plog < pfile_log_warn) {
      dst = stdout;
    } else {
      dst = stdout;
    }
    sev_str = "???";
    switch (plog) {
      case pfile_log_crit: 
        sev_str = "critical: "; 
        break;
      case pfile_log_err:  
        sev_str = pfile_line_get(pf) ? "" : "error: "; 
        break;
      case pfile_log_warn: 
        sev_str = "warning: "; 
        break;
      case pfile_log_none: 
        sev_str = ""; 
        break;
      case pfile_log_info: 
        sev_str = (pfile_flag_test(pf, PFILE_FLAG_QUIET)) ? 0 : ""; 
        break;
      case pfile_log_debug: 
        sev_str = (pfile_flag_test(pf, PFILE_FLAG_DEBUG)) ? "" : 0; 
        break;
    }
    if (sev_str) {
      if ((pfile_source_line_get(pf->src)
            && ((pfile_log_err == plog) 
              || (pfile_log_warn == plog)
        || (pfile_log_crit == plog)))) {
        fprintf(dst, "%s:%u: ", pfile_source_name_get(pf->src),
            pfile_source_line_get(pf->src));
      }
      fprintf(dst, "%s", sev_str);
      va_start(lst, fmt);
      vfprintf(dst, fmt, lst);
      va_end(lst);
      fputc('\n', dst);
      fflush(dst);
    }
  }
}

/*
 * NAME
 *   pfile_log_syserr
 *
 * DESCRIPTION
 *   log a system error
 *
 * PARAMETERS
 *   pf : pfile
 *   rc : result
 *
 * RETURN
 *   none
 *
 * NOTES
 *   used mostly for ``out of memory'' and such
 */
void pfile_log_syserr(pfile_t *pf, result_t rc)
{
  if (result_ok != rc) {
    const char *str;

    str = "{??unknown??}";
    switch (rc) {
      case result_ok:             str = "no error";       break;
      case result_memory:         str = "out of memory";  break;
      case result_syntax:         str = "syntax error";   break;
      case result_not_found :     str = "not found";      break;
      case result_divide_by_zero: str = "divide by zero"; break;
      case result_exists:         str = "exists";         break;
      case result_range:          str = "out of range";   break;
      case result_internal:       str = "internal error"; break;
      case result_invalid:        str = "invalid";        break;
    }
    pfile_log(pf, pfile_log_err, PFILE_MSG_INTERNAL_ERROR, str);
  }
}



/*
 * NAME
 *   pfile_variable_find
 *
 * DESCRIPTION
 *   find an existing variable
 *
 * PARAMETERS
 *   pf   : returned by pfile_open()
 *   log  : how to log result
 *   name : name to find
 *   dst  : holds result on success
 *
 * RETURN
 *   0      : no error
 *   ENOENT : not found
 *
 * NOTES
 */
static boolean_t varname_cb(void *arg, const variable_t *var, const void *data)
{
  const char   *name;
  size_t        sz;

  sz = *(size_t *) arg;

  name = variable_name_get(var);
  return (sz == strlen(name)) && (0 == memcmp(name, data, sz));
}

variable_t *pfile_variable_find(pfile_t *pf, pfile_log_t plog, size_t sz,
  const char *name)
{
  variable_t *var;

  if (PFILE_LENGTH_DEF == sz) {
    sz = strlen(name);
  }
  var = variable_list_find(&pf->var_active, varname_cb, &sz, name);
  if (!var) {
    pfile_log(pf, plog, PFILE_MSG_NOT_FOUND, name);
  }
  return var;
}

boolean_t pfile_variable_exists(pfile_t *pf, size_t sz, const char *name)
{
  variable_t *var;

  var = pfile_variable_find(pf, pfile_log_none, sz, name);
  if (var) {
    variable_release(var);
  }
  return (0 != var);
}

/*
 * NAME
 *   pfile_value_find
 *
 * DESCRIPTION
 *   find an existing variable & wrap it in a value
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
value_t *pfile_value_find(pfile_t *pf, pfile_log_t plog, size_t sz,
  const char *name)
{
  int         rc;
  variable_t *var;
  value_t    *val;

  val = 0;
  var = pfile_variable_find(pf, plog, sz, name);
  if (var) {
    rc = value_alloc(&val, var, 0, 0, variable_sz_get(var));
    if (0 != rc) {
      pfile_log_syserr(pf, rc);
    }
    variable_release(var);
  }
  return val;
}

/*
 * NAME
 *   pfile_variable_alloc
 *
 * DESCRIPTION
 *   allocate a variable
 *
 * PARAMETERS
 *   pf : returned by pfile_open()
 *   name : variable name (0 for unnamed)
 *   flags :
 *   sz    :
 *   ct    : array size, must be >= 1
 *   cbuf  : constant buffer
 *   dst   : [out] holds result on success
 *
 * RETURN
 *
 * NOTES
 */
result_t pfile_variable_alloc(pfile_t *pf, const char *name, flag_t flags,
    variable_sz_t sz, variable_ct_t ct, const variable_const_t *n,
    variable_t **dst)
{
  result_t    rc;
  variable_t *var;

  var = pfile_variable_find(pf, pfile_log_none, strlen(name), name);
  if (var) {
    rc = result_exists;
    variable_release(var);
  } else {
    rc = variable_alloc(&var, name, flags, sz, ct, n);
    if (result_ok == rc) {
      variable_list_append(&pf->var_active, var);
      if (dst) {
        *dst = var;
      } else {
        variable_release(var);
      }
    }
  }
  if (result_ok != rc) {
    pfile_log_syserr(pf, rc);
  }
  return rc;
}

/*
 * NAME
 *   pfile_expr_push
 *
 * DESCRIPTION
 *   push a subexpression onto the expression stack
 *
 * PARAMETERS
 *   pf   : returned by pfile_open()
 *   stk  : top of the expression stack
 *   expr : expression to append
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this creates commands as necessary
 */
result_t pfile_expr_push(pfile_t *pf, expr_t **stk, value_t *val,
    operator_t op)
{
  result_t rc;
  expr_t  *stkptr;
  expr_t  *expr;

  rc     =  expr_alloc(&expr, op, val);
  stkptr = *stk;
  if (result_ok == rc) {
    while (stkptr 
      && (operator_priority_get(expr_operator_get(stkptr))
        >= operator_priority_get(expr_operator_get(expr)))) {
      expr_t  *stkpv;
      value_t *dst;

      stkpv = expr_link_get(stkptr);
      dst   = 0;

      if (operator_is_assign(expr_operator_get(stkptr))) {
        /* stkptr->val stkptr->op expr->val */
        dst = expr_val_get(stkptr);
        if (value_is_const(dst)) {
          dst = 0;
        }
        pfile_cmd_op_add(pf, 0, expr_operator_get(stkptr),
          &dst, expr_val_get(expr), 0);
        value_lock(dst);
      } else {
        /* dst = stkptr->val stkptr->op expr->val */
        pfile_cmd_op_add(pf, 0, expr_operator_get(stkptr),
          &dst, expr_val_get(stkptr), expr_val_get(expr));
      }
      expr_val_set(expr, dst);
      expr_free(stkptr);
      value_release(dst);
      stkptr = stkpv;
    }
    expr_link_set(expr, stkptr);
    *stk = expr;
  }
  if (result_ok != rc) {
    pfile_log_syserr(pf, rc);
  }
  return rc;
}

result_t pfile_label_alloc(pfile_t *pf, const char *label, 
  label_free_cb_t free_fn, label_t **dst)
{
  lbllist_t *lst;
  char       name[32];
  label_t   *lbl;
  result_t   rc;

  rc  = result_ok;

  lst = (label) ? &pf->label_user : &pf->label_temp;
  if (!label) {
    if (pf->label_temp_ptr) {
      lbl = pf->label_temp_ptr;
      pf->label_temp_ptr = label_next_get(pf->label_temp_ptr);
      label_lock(lbl);
    } else {
      lbl = 0;
      sprintf(name, "_l%u", ++pf->label_temp_ct);
    }
  } else {
    lbl = pfile_label_find(pf, pfile_log_none, label);
    if (lbl) {
      label_release(lbl);
      rc = result_exists;
    } else if (result_not_found == rc) {
      rc = 0;
    }
  }
  if ((0 == rc) && !lbl) {
    rc = label_alloc(&lbl, (label) ? label : name, free_fn);
    if (0 == rc) {
      label_list_append(lst, lbl);
    }
  }
  if (0 != rc) {
    pfile_log_syserr(pf, rc);
  } else {
    if (dst) {
      *dst = lbl;
    } else {
      label_release(lbl);
    }
  }
  return rc;
}

/*
 * NAME
 *   pfile_label_find
 *
 * DESCRIPTION
 *   find an existing label
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static boolean_t lblname_cb(void *arg, const label_t *lbl, const void *data)
{
  const char   *name;

  name = label_name_get(lbl);
  return (0 == strcmp(name, data));
}

label_t *pfile_label_find(pfile_t *pf, pfile_log_t plog, const char *name)
{
  label_t *lbl;

  lbl = label_list_find(&pf->label_user, lblname_cb, 0, name);
  if (!lbl) {
    pfile_log(pf, plog, PFILE_MSG_NOT_FOUND, name);
  }
  return lbl;
}

/*
 * NAME
 *   pfile_cmd_add
 *
 * DESCRIPTION
 *   add a command to the command list
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
void pfile_cmd_add(pfile_t *pf, cmd_t *cmd)
{
  label_t *lbl;

  lbl = cmd_label_get(cmd);
  if (pf->label_next) {
    if (lbl) {
      label_master_set(lbl, pf->label_next);
    }
    lbl = pf->label_next;
    cmd_label_set(cmd, lbl);
  }
  pf->label_next = 0;

  if (pf->src) {
    cmd_line_set(cmd, pfile_source_line_get(pf->src));
    cmd_source_set(cmd, pf->src);
  }
  cmd_link_set(cmd, pf->cmds);
  pf->cmds = cmd;
}

void pfile_cmd_special_add(pfile_t *pf, label_t *lbl, cmd_type_t type)
{
  cmd_t *cmd;

  cmd = cmd_alloc(0, type, 0, sizeof(*cmd));
  if (lbl) {
    pfile_cmd_label_add(pf, lbl);
  }
  if (cmd) {
    pfile_cmd_add(pf, cmd);
  }
}

/*
 * NAME
 *   pfile_cmd_op_fixup
 *
 * DESCRIPTION
 *   fixup the top pfile cmd
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this changes expressions as follows:
 *     x = x + 1 --> x += 1
 *   using rules of commutativity
 */
static void pfile_cmd_op_fixup(pfile_t *pf)
{
  cmd_t *cmd;

  cmd = pf->cmds;
  if (cmd_type_operator == cmd_type_get(cmd)) {
    if (value_same(cmd_opdst_get(cmd), cmd_opval1_get(cmd))) {
      /* x = x + y --> x += y */
      cmd_opval1_set(cmd, cmd_opval2_get(cmd));
      cmd_opval2_set(cmd, 0);
    } else if (operator_is_commutative(cmd_optype_get(cmd))
        && value_same(cmd_opdst_get(cmd), cmd_opval2_get(cmd))) {
      /* x = y op x, op is commutative --> x op= y */
      cmd_opval2_set(cmd, 0);
    } else if (operator_is_relation(cmd_optype_get(cmd))
        && value_same(cmd_opdst_get(cmd), cmd_opval2_get(cmd))) {
      /* x = y op x, op is relation --> op = reverse relation, x op= y */
      operator_t op;

      switch (cmd_optype_get(cmd)) {
        case operator_lt: op = operator_gt; break;
        case operator_le: op = operator_ge; break;
        case operator_ge: op = operator_le; break;
        case operator_gt: op = operator_lt; break;
        default: op = operator_null; break;
      } 
      if (operator_null != op) {
        cmd_opval2_set(cmd, 0);
        cmd_optype_set(cmd, op);
      }
    }
  }
}

/*
 * NAME
 *   pfile_cmd_op_add
 *
 * DESCRIPTION
 *   add an operation
 *
 * PARAMETERS
 *   pf:
 *   label : line label
 *   op    : operation
 *   dst   : destination, can be 0 to allocate a temporary
 *   val1  : value 1 (can be 0 for unary operations, cannot be 0 for binary)
 *   val2  : value 2 (must be 0 for unary, can be 0 for binary)
 *
 * RETURN
 *
 * NOTES
 *   if dst is NULL on entry, it will hold a value on successful exit
 *   this takes care of the following identities:
 *     commutative & relational -- constant is the 2nd parameter
 *     various identities:
 *     x + 0 = nop
 *     x - 0 = nop
 *     x * 1 = x
 *     x / 1 = x
 *     x * 0 = 0
 *     x / 0 = err
 *     x < 0, x unsigned is always 0
 *     x >= 0, x unsigned is always 1
 *     x >  0, x unsigned --> x != 0
 *     x <= 0, x unsigned --> x == 0
 *     0 - x = -x
 *     0 / x = 0
 *     x % 0 = 0
 *     x % 1 = 0
 *     x ^ 0  = x
 *     x | 0  = x
 *     x && 0 = 0
 *     x & 0  = 0
 *     x || 0 = !!x
 *     x &  x = x 
 *     x | x  = x 
 *     x ^ x  = 0
 *     x && x = !!x
 *     x || x = !!x
 *     x >> # = 0 if # >= bits in x
 *     x << # = 0 if # >= bits in x
 *     0 >> # = 0
 *     0 << # = 0
 *     x = ?, x never used -- nop
 */
void pfile_cmd_op_add(pfile_t *pf, label_t *label,
  operator_t op, value_t **pdst, value_t *val1, value_t *val2)
{
  cmd_t        *ptr;
  int           rc;
  value_t      *dst;
  int           is_const;
  int           release_val1;
  variable_sz_t dst_sz;
  boolean_t     dst_is_signed;

  release_val1 = 0;

  rc = 0;
  dst = (pdst) ? *pdst : 0;

  if (operator_null == op) {
    is_const = 0;
  } else if (operator_is_assign(op)) {
    assert(val2 == 0);
    assert(val1);
    is_const = 0;
  } else if (operator_is_unary(op)) {
    assert(val2 == 0);
    assert(val1 || dst);
    if (!val1) {
      val1 = dst;
    }
    is_const = value_is_const(val1);
  } else {
    assert(val1 != 0);
    assert(val2 || dst);
    if (!val2) {
      val2 = val1;
      val1 = dst;
    }
    is_const = value_is_const(val1) && value_is_const(val2);
  }
  if (dst) {
    dst_sz = value_sz_get(dst);
    dst_is_signed = value_flag_test(dst, VARIABLE_FLAG_SIGNED);
  } else {
    /* determine dst characteristics:
         smaller size --> bigger size
         signed       --> unsigned
    */     
    dst_sz = value_sz_get(val1);
    if (val2 && (value_sz_get(val2) > dst_sz)) {
      dst_sz = value_sz_get(val2);
    }

    dst_is_signed = value_flag_test(val1, VARIABLE_FLAG_SIGNED);
    if (dst_is_signed && val2) {
      dst_is_signed = value_flag_test(val2, VARIABLE_FLAG_SIGNED);
    }
  }
  /* now...val1 holds a value, and value2 does also for binary ops */
  if (is_const) {
    variable_const_t n;

    rc = operator_exec(op, &n, value_const_get(val1), 
        value_const_get(val2));
    if (result_ok != rc) {
      pfile_log_syserr(pf, rc);
    }
    rc = pfile_constant_get(pf, n, &dst);
    if (result_ok == rc) {
      value_sz_set(dst, dst_sz);
    }
  } else if (dst && (value_flag_test(dst, VARIABLE_FLAG_CONST))) {
    pfile_log(pf, pfile_log_err, PFILE_MSG_ASSIGN_TO_CONST);
    rc = result_syntax;
  } else {
    /* let's work on some identiies                      */
    /* first -- constant is always on the right for +, * */
    /*          and relations                            */
    if (dst
        && (0 != memcmp(value_name_get(dst), "_t", 0))
        && !value_use_ct_get(dst)) {
      op = operator_assign;
      val2 = 0;
      val1 = dst;
    } else if (value_is_const(val1)) {
      if (operator_is_commutative(op) || operator_is_relation(op)) {
        /* swap val1 & val2 */
        value_t *tmp;

        tmp  = val1;
        val1 = val2;
        val2 = tmp;
        /* change the relation */
        if (operator_lt == op) {
          op = operator_gt;
        } else if (operator_le == op) {
          op = operator_ge;
        } else if (operator_ge == op) {
          op = operator_le;
        } else if (operator_gt == op) {
          op = operator_lt;
        }
      }
    }
    if (val2 && value_is_const(val2)) {
      variable_const_t n;

      n = value_const_get(val2);
      if ((0 == n) 
          && ((operator_add == op) 
            || (operator_sub == op)
            || (operator_orb == op)
            || (operator_xorb == op))) {
        /* x + 0 = x */
        /* x - 0 = x */
        /* x | 0 = x */
        /* x ^ 0 = x */
        val2 = 0;
        op = operator_assign; /* identity */
      } else if (((operator_mul == op) || (operator_div == op))
          && (1 == n)) {
        /* x * 1 = x */
        /* x / 1 = x */
#if 1
        val2 = 0;
        op = operator_assign; /* identity */
#endif
      } else if (((0 == n)
          && ((operator_mul == op)
            || (operator_andl == op)
            || (operator_andb == op)))
        || ((operator_mod == op) && (1 == n))) {
        /* x *  0 = 0 */
        /* x &  0 = 0 */
        /* x && 0 = 0 */
        /* x %  1 = 0 */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 0, &val1);
        release_val1 = 1;
      } else if (((operator_shift_left == op) && (n >= 8U * dst_sz))
          || ((operator_shift_right == op) && (n >= 8U * dst_sz))) {
        /* shift greater than dest size == 0 */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 0, &val1);
        release_val1 = 1;
      } else if ((0 == n) && (operator_orl == op)) {
        /* x || 0 == !!x */
        val2 = 0;
        op = operator_logical;
      } else if ((operator_lt == op) 
          && !(value_flag_test(val1, VARIABLE_FLAG_SIGNED))
          && (0 == n)) {
        /* x < 0, x is unsigned --> always false */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 0, &val1);
        release_val1 = 1;
      } else if ((operator_ge == op)
          && !(value_flag_test(val1, VARIABLE_FLAG_SIGNED))
          && (0 == n)) {
        /* x >= 0, x is unsigned --> always true */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 1, &val1);
        release_val1 = 1;
      } else if ((operator_gt == op)
          && !(value_flag_test(val1, VARIABLE_FLAG_SIGNED))
          && (0 == n)) {
        /* x > 0, x unsigned --> x != 0 */
        op = operator_ne;
      } else if ((operator_le == op)
          && !(value_flag_test(val1, VARIABLE_FLAG_SIGNED))
          && (0 == n)) {
        /* x <= 0, x unsigned --> x == 0 */
        op = operator_eq;
      } else if (((operator_div == op) || (operator_mod == op))
        && (0 == n)) {
        pfile_log(pf, pfile_log_err, PFILE_MSG_DIVIDE_BY_ZERO);
      }
    }
    if ((0 == rc) && value_is_const(val1)) {
      variable_const_t n;

      n = value_const_get(val1);
      if ((n == 0)
        && ((operator_div == op)
          || (operator_shift_left == op)
          || (operator_shift_right == op))) {
        /* 0 / x = 0 */
        /* 0 << x = 0 */
        /* 0 >> x = 0 */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 0, &val1);
        release_val1 = 1;
      } else if ((operator_sub == op) && (0 == n)) {
        /* 0 - x = -x */
        val1 = val2;
        val2 = 0;
        op = operator_neg;
      }
    }
    if (val2 && value_same(val1, val2)
        && value_sz_get(val1) == value_sz_get(val2)) {
      if ((operator_andb == op) || (operator_orb == op)) {
        /* x & x = x (bitwise) */
        /* x | x = x (bitwise) */
        op = operator_assign;
        val2 = 0;
      } else if (operator_xorb == op) {
        /* x ^ x = 0 */
        val2 = 0;
        op = operator_assign;
        rc = pfile_constant_get(pf, 0, &val1);
        release_val1 = 1;
      } else if ((operator_andl == op) || (operator_orl == op)) {
        /* x & x = !!x (logical) */
        /* x | x = !!x (logical) */
        op = operator_logical;
        val2 = 0;
      }
    }
    /* let's put out some warnings. Specifically:
     *   bittst var,const where const > 8 * var sz
     *   andb   var,const where const > var */
    if ((result_ok == rc) && !dst && (operator_null != op)) {
      rc = pfile_value_temp_get(pf, dst_sz, 
          (dst_is_signed) ? VARIABLE_FLAG_SIGNED : 0, &dst);
    }
    if ((result_ok == rc) && (operator_null == op)) {
      ptr = cmd_op_alloc(label, op, 0, 0, 0);
      if (ptr) {
        pfile_cmd_add(pf, ptr);
      } else {
        pfile_log_syserr(pf, result_memory);
      }
    } else if ((operator_assign == op) && value_same(dst, val1)) {
      /* x = x --> nop */
#if 1
    } else if ((operator_assign == op)
      && pf->cmds
      && (cmd_type_operator == cmd_type_get(pf->cmds))
      && value_same(val1, cmd_opdst_get(pf->cmds))
      && (0 == memcmp(value_name_get(cmd_opdst_get(pf->cmds)), "_t", 2))
      && (!value_flag_test(cmd_opdst_get(pf->cmds), VARIABLE_FLAG_NO_OPT))) {
      /* _t# = y   */
      /*  x  = _t# */
      cmd_t *cmd;

      cmd = pf->cmds;
      if ((operator_assign == cmd_optype_get(cmd))
          && value_same(dst, cmd_opval1_get(cmd))) {
        /* _t# = x     
         * x   = _t0   
         * ---> nop  */
        /* nb : ONLY if sizes are the same! */
        pf->cmds = cmd_link_get(cmd);
        cmd_free(cmd); 
      } else {
        /* _t0 = y
         * x   = _t0
         * --->
         * x   = y */
        cmd_opdst_set(pf->cmds, dst);
        pfile_cmd_op_fixup(pf);
        rc = 0;
      }
#endif
    } else {
      /* finally, if the form is x = x op y, change it to x op= y */
      if (value_same(dst, val1)) {
        val1 = val2;
        val2 = 0;
      } else if (operator_is_commutative(op)
          && value_same(dst, val2)) {
        val2 = 0;
      }
      ptr = cmd_op_alloc(label, op, dst, val1, val2);
      if (!ptr) {
        rc = result_memory;
        pfile_log_syserr(pf, rc);
      } else {
        pfile_cmd_add(pf, ptr);
        pfile_cmd_op_fixup(pf);
      }
    }
  }
  if (release_val1) {
    value_release(val1);
  }
  if (pdst) {
    if (dst != *pdst) {
      if (*pdst) {
        value_release(*pdst);
      }
      *pdst = dst;
    }
  } else if (dst) {
    value_release(dst);
  }
}

/*
 * NAME
 *   pfile_cmd_branch_add
 *
 * DESCRIPTION
 *   add a branching command
 *
 * PARAMETERS
 *   pf : pfile
 *   label : line label
 *   type  : branching type
 *   cond  : branching condition
 *   dst   : destination label
 *   var   : condition variable
 *
 * RETURN
 *
 * NOTES
 */
void pfile_cmd_branch_add(pfile_t *pf, label_t *label,
  cmd_branchtype_t type, cmd_branchcond_t cond, label_t *dst,
  value_t *var)
{
  cmd_t *ptr;

  if ((cmd_branchcond_none != cond) 
    && value_is_const(var)) {
    /* this is an absolute branch disguised as a conditonal one */
    variable_const_t n;

    n = value_const_get(var);
    if (((0 == n) && (cmd_branchcond_false == cond)) 
        || ((0 != n) && (cmd_branchcond_true == cond))) {
      cond = cmd_branchcond_none;
    } else {
      type = cmd_branchtype_none;
    }
  }
  if (cmd_branchtype_none == type) {
    /* do nothing */
  } else {
    ptr = cmd_branch_alloc(label, type, cond, dst, var);
    if (!ptr) {
      pfile_log_syserr(pf, result_memory);
    } else {
      pfile_cmd_add(pf, ptr);
    }
  }
}


/*
 * NAME
 *   pfile_cmd_label_add
 *
 * DESCRIPTION
 *   add a label only
 *
 * PARAMETERS
 *   pf    :
 *   label :
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this doesn't really do anything other than set the
 *   ``next command gets this label''
 */
void pfile_cmd_label_add(pfile_t *pf, label_t *label)
{
  if (!pf->label_next) {
    pf->label_next = label;
  }
  label_master_set(label, pf->label_next);
}


/*
 * NAME
 *   pfile_cmd_remove_unreachable
 *
 * DESCRIPTION
 *   remove unreachable commands
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pfile_cmd_remove_unreachable(pfile_t *pf)
{
  cmd_t *cmd;
  cmd_t *cmd_pv;
  unsigned ct;

  cmd = pf->cmds;
  cmd_pv = 0;

  /* first free unused commands */
  ct = 0;
  while (cmd) {
    if (!cmd_flag_test(cmd, CMD_FLAG_REACHABLE)) {
      if (cmd_pv) {
        cmd_link_set(cmd_pv, cmd_link_get(cmd));
      } else {
        pf->cmds = cmd_link_get(cmd);
      }
      cmd_free(cmd);
      cmd = cmd_pv;
      ct++;
    }
    cmd_pv = cmd;
    cmd = cmd_link_get(cmd);
  }
  pfile_log(pf, pfile_log_debug, "%u commands freed", ct);
}

/*
 * NAME
 *   pfile_label_remove_unused
 *
 * DESCRIPTION
 *   remove unused labels
 *
 * PARAMETERS
 *   lst  : label list
 *   name : list name (for reporting)
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pfile_label_remove_unused(pfile_t *pf, lbllist_t *lst, 
    const char *name)
{
  label_t *lbl;
  unsigned ct;

  lbl = label_list_head(lst);
  ct  = 0;
  while (lbl) {
    label_t *lbl_next;

    lbl_next = label_next_get(lbl);
    if (!label_usage_get(lbl)) {
      label_remove(lbl);
      label_release(lbl);
      ct++;
    }
    lbl = lbl_next;
  }
  pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_LABELS_REMOVED, ct, name);
}

/*
 * NAME
 *   pfile_mem_minimize
 *
 * DESCRIPTION
 *   minimize the amount of used memory
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   this release all unused temporaries, isr temporaries,
 *   labels, and cmds before calling pic_cmd_alloc() for
 *   the times when memory is at a premium.
 */

static void pfile_mem_minimize(pfile_t *pf)
{
  return;
  pfile_cmd_remove_unreachable(pf);
  pfile_label_remove_unused(pf, &pf->label_user, "user");
  pfile_label_remove_unused(pf, &pf->label_temp, "temporary");
}

/*
 * NAME
 *   dump the command list
 *
 * DESCRIPTION
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
void pfile_cmd_dump(pfile_t *pf)
{
  cmd_t          *cmd;
  unsigned        cmd_no;
  label_t        *lbl;
  int             is_dead;
  variable_t     *vars[3];
  boolean_t       first;
  pfile_source_t *src;

  first = boolean_true;

  is_dead = 0;
  cmd_list_reverse(&pf->cmds);
#if 1
  pfile_source_rewind(pf->src);
  /* first some optimizations */
  pfile_log(pf, pfile_log_info, PFILE_MSG_PCODE_OPT);
  cmd_opt_branch_fixup(pf->cmds);
  cmd_opt_call_return(pf->cmds);
  cmd_opt_dst(pf, pf->cmds);
  /* now analyze the result to get rid of dead code */
  lbl = pfile_isr_entry_get(pf);
  if (lbl) {
    cmd_analyze(pf->cmds, cmd_label_find(pf->cmds, lbl), 
      CMD_FLAG_INTERRUPT, 0);
    label_release(lbl);
  }
  lbl = pfile_user_entry_get(pf);
  if (lbl) {
    cmd_analyze(pf->cmds, cmd_label_find(pf->cmds, lbl),
      CMD_FLAG_USER, 0);
    label_release(lbl);
  }
  cmd_opt_branch2(pf, pf->cmds);
  cmd_opt_counters_fix(pf->cmds);
  /* now that its been analyzed, add in the isr preamble */

  if (pf->label_isr) {
    pfile_isr_preamble(pf);
  }
  pfile_mem_minimize(pf);
  vars[0] = variable_list_head(&pf->var_temp);
  vars[1] = variable_list_head(&pf->var_isr);
  vars[2] = variable_list_head(&pf->var_active);

  pic_cmd_alloc(pf, pf->cmds, COUNT(vars), vars);
  pfile_log(pf, pfile_log_info, PFILE_MSG_DUMPING_RESULT);
  /* dump all of the variables */
  src = cmd_source_get(pf->cmds);
  pfile_source_rewind(src);

  for (cmd = pf->cmds, cmd_no = 0; cmd; cmd = cmd_link_get(cmd), cmd_no++) {
    int ch;

    if (!is_dead && !cmd_flag_test(cmd, CMD_FLAG_REACHABLE)) {
      pfile_write(pf, pfile_write_lst, ";***dead code\n");
      is_dead = 1;
    } else if (is_dead && cmd_flag_test(cmd, CMD_FLAG_REACHABLE)) {
      pfile_write(pf, pfile_write_lst, ";************\n");
      is_dead = 0;
    }
    /* first, dump anything that isn't in the current file */
    if (src != cmd_source_get(cmd)) {
#if 0
      while (src != cmd_source_get(cmd)) {
        pfile_write(pf, pfile_write_lst, ";%5u ", 
            pfile_source_line_get(src));
        
        while (EOF != (ch = pfile_source_ch_get(src))) {
          if ('\n' == ch) {
            pfile_write(pf, pfile_write_lst, "\n");
            break;
          } else {
            pfile_write(pf, pfile_write_lst, "%c", ch);
          }
        }
      }
#endif
      src = cmd_source_get(cmd);
      pfile_source_rewind(src);
    }
#if 1
    while (cmd_line_get(cmd) != pfile_source_line_get(src)) {
      pfile_write(pf, pfile_write_lst, ";%5u ", 
          1 + pfile_source_line_get(src));
      
      while (EOF != (ch = pfile_source_ch_get(src))) {
        if ('\n' == ch) {
          pfile_write(pf, pfile_write_lst, "\n");
          break;
        } else {
          pfile_write(pf, pfile_write_lst, "%c", ch);
        }
      }
    }
#endif
    if (pfile_flag_test(pf, PFILE_FLAG_DEBUG)) {
      pfile_write(pf, pfile_write_lst, "; c(%u) l(%u)", cmd_no,
        cmd_line_get(cmd));
    } else {
      pfile_write(pf, pfile_write_lst, ";      ");
    }
    pfile_write(pf, pfile_write_lst, " %s", 
        (cmd_type_intrinsic == cmd_type_get(cmd)) ? "==" : "**");
    pfile_write(pf, pfile_write_lst, "{%c%c} ", 
      (cmd_flag_test(cmd, CMD_FLAG_USER)) ? 'u' : '-',
      (cmd_flag_test(cmd, CMD_FLAG_INTERRUPT)) ? 'i' : '-');

    if (pf->f_lst) {
      cmd_dump(cmd, pf->f_lst);
    }
    if (!is_dead) {
       pic_cmd_dump(pf, cmd, first);
       first = boolean_false;
    }
  }
  pic_cmd_dump(pf, 0, first);
  pic_code_free(pf);
#endif
  cmd_list_reverse(&pf->cmds);
}


/*
 * NAME
 *   pfile_label_list_dump
 *
 * DESCRIPTION
 *   dump the label list
 *
 * PARAMETERS
 *   pf   : 
 *   lst  :
 *   name :
 *
 * RETURN
 *
 * NOTES
 */
static void pfile_label_list_dump(pfile_t *pf, lbllist_t *lst, const char *name)
{
#if 1
  label_t *lbl;

  pfile_write(pf, pfile_write_lst, ";%s Labels\n", name);
  for (lbl = label_list_head(lst);
       lbl;
       lbl = label_next_get(lbl)) {
    const label_t *master;

    pfile_write(pf, pfile_write_lst, ";%s(%u:%u:pc=%04x)", 
        label_name_get(lbl), label_usage_get(lbl),
        label_ref_ct_get(lbl), label_pc_get(lbl));
    master = label_master_get(lbl);
    if (lbl != master) {
      pfile_write(pf, pfile_write_lst, " --> %s", label_name_get(master));
    }
    if (cmd_branchtype_none != label_brtype_get(lbl)) {
      pfile_write(pf, pfile_write_lst, " (");
      if (cmd_branchtype_return == label_brtype_get(lbl)) {
        pfile_write(pf, pfile_write_lst, "return");
      } else if (cmd_branchtype_goto == label_brtype_get(lbl)) {
        pfile_write(pf, pfile_write_lst, "goto %s", 
            label_name_get(label_brdest_get(lbl)));
      }
      pfile_write(pf, pfile_write_lst, ")");
    }
    pfile_write(pf, pfile_write_lst, "\n");
  }
#endif
}


void pfile_label_dump(pfile_t *pf)
{
  pfile_label_list_dump(pf, &pf->label_user, "User Defined");
  pfile_label_list_dump(pf, &pf->label_temp, "Temporary");
}


/*
 * NAME
 *   pfile_variable_fixup
 *
 * DESCRIPTION
 *   look for any variables that are never assigned & set them to const
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
void pfile_variable_fixup(pfile_t *pf)
{
  variable_t *var;

  for (var = variable_list_head(&pf->var_active);
       var;
       var = variable_link_get(var)) {
    if (!variable_assign_ct_get(var)
        && !(variable_flag_test(var, VARIABLE_FLAG_ASSIGNED))
        && !(variable_flag_test(var, VARIABLE_FLAG_VOLATILE))
        && !(variable_flag_test(var, VARIABLE_FLAG_CONST))) {
      variable_flag_set(var, VARIABLE_FLAG_CONST);
      variable_const_set(var, 0, 0);
    }
    if (variable_assign_ct_get(var)) {
      variable_flag_set(var, VARIABLE_FLAG_ASSIGNED);
    }
    if (variable_use_ct_get(var)) {
      variable_flag_set(var, VARIABLE_FLAG_USED);
    }
  }
}

/*
 * NAME
 *   pfile_variable_list_dump
 *
 * DESCRIPTION
 *   dump the variable list
 *
 * PARAMETERS
 *   pf    : pfile
 *   lst   :
 *   title :
 *
 * RETURN
 *
 * NOTES
 */
static void pfile_variable_list_dump(pfile_t *pf, varlist_t *lst, 
      const char *title)
{
#if 1
  variable_t *var;

  pfile_write(pf, pfile_write_lst, ";%s Variables\n"
         ";============\n", title);
  for (var = variable_list_head(lst); var; var = variable_link_get(var)) {
    pfile_write(pf, pfile_write_lst, 
        ";%s F(%c%c%c%c%c%c) sz(%u) ct(%u) n(%ld) use(%u) region(0x%x)",
        variable_name_get(var),
        (variable_flag_test(var, VARIABLE_FLAG_CONST))    ? 'C' : '-',
        (variable_flag_test(var, VARIABLE_FLAG_SIGNED))   ? 'S' : '-',
        (variable_flag_test(var, VARIABLE_FLAG_VOLATILE)) ? 'V' : '-',
        (variable_flag_test(var, VARIABLE_FLAG_NOTEMP))   ? 'T' : '-',
        (variable_assign_ct_get(var))                      ? 'A' : '-',
        (variable_use_ct_get(var))                         ? 'U' : '-',
        variable_sz_get(var),
        variable_ct_get(var),
        (unsigned long) variable_const_get(var, 0),
        variable_use_ct_get(var),
        variable_region_get_all(var));
    if (!(variable_flag_test(var, VARIABLE_FLAG_CONST))) {
      pfile_write(pf, pfile_write_lst, " assign(%u) base(%x)", 
        variable_assign_ct_get(var),
        variable_base_get(var));
    }
    pfile_write(pf, pfile_write_lst, "\n");
  }
#endif
}

void pfile_variable_dump(pfile_t *pf)
{
  pfile_variable_list_dump(pf, &pf->var_active, "Active");
  pfile_variable_list_dump(pf, &pf->var_const, "Unnamed Constant");
  pfile_variable_list_dump(pf, &pf->var_temp, "Temporary");
  pfile_variable_list_dump(pf, &pf->var_isr, "ISR Temporary");
}

unsigned pfile_errct_get(const pfile_t *pf)
{
  return pf->errct;
}

/*
 * NAME
 *   pfile_isr_preamble
 *
 * DESCRIPTION
 *   generate the generic ISR preamble code
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   isr preamble:
 *      1. save any temporaries
 *      2. jump to first defined ISR
 *   this is ignored if there are no ISRs defined
 */
static void pfile_isr_var_flag_fix(value_t *val, flag_t flags)
{
  if (value_name_get(val) && (0 == memcmp(value_name_get(val), "_t", 2))) {
    variable_t *var;

    var = value_variable_get(val);
    variable_flag_set(var, flags);
  }
}

static int pfile_isr_var_changed(pfile_t *pf, value_t *src, value_t **dst)
{
  flag_t flags;
  int    rc;

  flags = value_flags_get_all(src);
  if ((flags & (VARIABLE_FLAG_INTERRUPT | VARIABLE_FLAG_USER))
    == VARIABLE_FLAG_INTERRUPT) {
    /* change this to it's interrupt-only companion */
    variable_t *visr;
    char        visr_name[32];

    sprintf(visr_name, "_tisr%s", value_name_get(src) + 2);

    for (visr = variable_list_head(&pf->var_isr);
         visr && (0 != strcmp(variable_name_get(visr), visr_name));
         visr = variable_link_get(visr))
      ;
    if (visr) {
      rc = value_alloc(dst, visr, 0, 0, value_sz_get(src));
      if (0 != rc) {
        pfile_log_syserr(pf, rc);
      }
      rc = 1;
    } else {
      rc = 0;
    }
  } else {
    rc = 0;
  }
  return rc;
}

static void pfile_isr_tmp_save(pfile_t *pf, int save)
{
  variable_t *tmp;
  variable_t *visr;

  for (tmp = variable_list_head(&pf->var_temp),
         visr = variable_list_head(&pf->var_isr);
       tmp;
       tmp = variable_link_get(tmp)) {
    if ((variable_flags_get_all(tmp) 
      & (VARIABLE_FLAG_INTERRUPT | VARIABLE_FLAG_USER))
      == (VARIABLE_FLAG_INTERRUPT | VARIABLE_FLAG_USER)) {
      while (visr && (0 != strcmp(variable_name_get(tmp) + 2,
        variable_name_get(visr) + 5))) {
        visr = variable_link_get(visr);
      }
      if (visr) {
        value_t *val_tmp;
        value_t *val_isr;
        int      rc;

        rc = value_alloc(&val_tmp, tmp, 0, 0, variable_sz_get(tmp));
        if (0 == rc) {
          rc = value_alloc(&val_isr, visr, 0, 0, variable_sz_get(visr));
          if (0 == rc) {
            if (save) {
              pfile_cmd_op_add(pf, 0, operator_assign, &val_isr, val_tmp, 0);
            } else {
              pfile_cmd_op_add(pf, 0, operator_assign, &val_tmp, val_isr, 0);
            }
            value_release(val_isr);
          }
          value_release(val_tmp);
        }
        if (0 != rc) {
          pfile_log_syserr(pf, rc);
        }
      }
    }
  }
}

void pfile_isr_preamble(pfile_t *pf)
{
  cmd_t      *cmd;
  variable_t *tmp;
  cmd_t      *cmdlist; /* command list  */
  label_t    *lbl;
  int         rc;

  /* first, scan all commands looking for those that
     execute in the interrupt context */
  pfile_cmd_label_add(pf, pf->label_isr_preamble);
  for (cmd = pf->cmds; cmd; cmd = cmd_link_get(cmd)) {
    if (cmd_flag_test(cmd, CMD_FLAG_INTERRUPT)) {
      flag_t vflag;

      vflag = VARIABLE_FLAG_INTERRUPT;
      if (cmd_flag_test(cmd, CMD_FLAG_USER)) {
        vflag |= VARIABLE_FLAG_USER;
      }
      /* let's see if there's anything we need to do */
      switch (cmd_type_get(cmd)) {
        case cmd_type_branch:
          pfile_isr_var_flag_fix(cmd_brval_get(cmd), vflag);
          break;
        case cmd_type_operator:
          pfile_isr_var_flag_fix(cmd_opdst_get(cmd), vflag);
          pfile_isr_var_flag_fix(cmd_opval1_get(cmd), vflag);
          pfile_isr_var_flag_fix(cmd_opval2_get(cmd), vflag);
          break;
        case cmd_type_intrinsic:
        case cmd_type_nop:
        case cmd_type_end:
        case cmd_type_isr_cleanup:
        case cmd_type_sleep:
        case cmd_type_clear:
          break;
      }
    }
  }
  /* now, create any isr temporaries. one is needed for each
     temporary used in the isr context */
  rc = 0;
  for (tmp = variable_list_head(&pf->var_temp);
       (0 == rc) && tmp;
       tmp = variable_link_get(tmp)) {
    if (variable_flag_test(tmp, VARIABLE_FLAG_INTERRUPT)) {
      /* we need an _tisr# version of this */
      char        visr_name[32];
      variable_t *visr;    /* isr variables */

      sprintf(visr_name, "_tisr%s", variable_name_get(tmp) + 2);
      rc = variable_alloc(&visr, visr_name, 0, variable_sz_get(tmp), 1, 0);
      if (0 != rc) {
        pfile_log_syserr(pf, rc);
      } else {
        variable_list_append(&pf->var_isr, visr);
        variable_release(visr);
      }
    }
  }
  /* run through the code, replacing any temporaries
     that are only used in interrupt context with their interrupt
     counterparts */
  if (0 == rc) {
    for (cmd = pf->cmds; cmd; cmd = cmd_link_get(cmd)) {
      if (cmd_flag_test(cmd, CMD_FLAG_INTERRUPT)) {
        value_t *visr;

        switch (cmd_type_get(cmd)) {
          case cmd_type_branch:
            if (pfile_isr_var_changed(pf, cmd_brval_get(cmd), &visr)) {
              cmd_brval_set(cmd, visr);
              value_release(visr);
            }
            break;
          case cmd_type_operator:
            if (pfile_isr_var_changed(pf, cmd_opdst_get(cmd), &visr)) {
              cmd_opdst_set(cmd, visr);
              value_release(visr);
            }
            if (pfile_isr_var_changed(pf, cmd_opval1_get(cmd), &visr)) {
              cmd_opval1_set(cmd, visr);
            }
            if (pfile_isr_var_changed(pf, cmd_opval2_get(cmd), &visr)) {
              cmd_opval2_set(cmd, visr);
              value_release(visr);
            }
            break;
          case cmd_type_intrinsic:
          case cmd_type_nop:
          case cmd_type_end:
          case cmd_type_isr_cleanup:
          case cmd_type_sleep:
          case cmd_type_clear:
            break;
        }
      }
    }
    /* create the isr_preamble -- any temporaries used by routines common
       to both user and interrupt level need to be saved during the context
       switch */
    cmdlist = pf->cmds; /* save the command list */
    pf->cmds = 0;

    pfile_isr_tmp_save(pf, 1);
    /* jump to the first ISR */
    pfile_cmd_branch_add(pf, 0, cmd_branchtype_goto,
      cmd_branchcond_none, pf->label_isr, 0);

    /* the last isr will jump to _isr_cleanup, so we'd best create it */
    lbl = pfile_label_find(pf, pfile_log_err, "_isr_cleanup");
    if (0 == rc) {
      pfile_cmd_label_add(pf, lbl);
      label_release(lbl);
    }
    /* restore any temporaries that were swapped */
    pfile_isr_tmp_save(pf, 0);
    /* pfile_cmd_op_add(pf, 0, operator_null, 0, 0, 0); */
    pfile_cmd_special_add(pf, 0, cmd_type_isr_cleanup);
    cmd_list_reverse(&pf->cmds);
    /* prepend the new commands to the existing command list, but
       note that each of the new commands executes at interrupt context */
    for (cmd = pf->cmds; cmd; cmd = cmd_link_get(cmd)) {
      cmd_flag_set(cmd, CMD_FLAG_REACHABLE);
      cmd_flag_set(cmd, CMD_FLAG_INTERRUPT);
      if (!cmd_link_get(cmd)) {
        cmd_link_set(cmd, cmdlist);
        break;
      }
    } 
  }
}

cmd_t *pfile_cmdlist_get(pfile_t *pf)
{
  return pf->cmds;
}

void pfile_cmdlist_set(pfile_t *pf, cmd_t *cmd)
{
  pf->cmds = cmd;
}

label_t *pfile_lblptr_get(pfile_t *pf)
{
  return pf->label_temp_ptr;
}

void pfile_lblptr_set(pfile_t *pf, label_t *lbl)
{
  pf->label_temp_ptr = lbl;
}

void pfile_rewind(pfile_t *pf)
{
  pfile_source_rewind(pf->src);
}

void pfile_user_entry_set(pfile_t *pf, label_t *lbl)
{
  if (lbl) {
    label_lock(lbl);
    label_usage_bump(lbl, ctr_bump_incr);
  }
  if (pf->label_main) {
    label_usage_bump(pf->label_main, ctr_bump_decr);
    label_release(pf->label_main);
  }
  pf->label_main = lbl;
}

label_t *pfile_user_entry_get(pfile_t *pf)
{
  label_t *lbl;

  lbl = pf->label_main;
  if (lbl) {
    label_lock(lbl);
  }
  return lbl;
}

void pfile_isr_entry_set(pfile_t *pf, label_t *lbl)
{
  if (lbl) {
    label_lock(lbl);
    label_usage_bump(lbl, ctr_bump_incr);
  }
  if (pf->label_isr) {
    label_usage_bump(pf->label_isr, ctr_bump_decr);
    label_release(pf->label_isr);
  }
  pf->label_isr = lbl;
}

label_t *pfile_isr_entry_get(pfile_t *pf) 
{
  label_t *lbl;

  lbl = pf->label_isr;
  if (lbl) {
    label_lock(lbl);
  }
  return lbl;
}

int pfile_write(pfile_t *pf, pfile_write_t where, const char *fmt, ...)
{
  FILE   *out;
  int     rc;

  out = pf->f_lst;
  switch (where) {
    case pfile_write_lst: out = pf->f_lst; break;
    case pfile_write_asm: out = pf->f_asm; break;
  }
  if (out) {
    va_list vlst;

    va_start(vlst, fmt);
    rc = vfprintf(out, fmt, vlst);
    va_end(vlst);
  } else {
    rc = 0;
  }
  return rc;
}

void pfile_write_hex(pfile_t *pf, unsigned pc, unsigned char data)
{
  if (pc != pf->hex.pc + pf->hex.ct) {
    pfile_hex_line_flush(pf);
    pf->hex.pc = pc;
  }
  pf->hex.buf[pf->hex.ct] = data;
  pf->hex.ct++;
  if (!((pf->hex.pc + pf->hex.ct) & 0x0f)) {
    pfile_hex_line_flush(pf);
  }
}

/*
 * NAME
 *   pfile_string_find
 *
 * DESCRIPTION
 *   return a string if it exists
 *
 * PARAMETERS
 *   pf  : 
 *   sz  : 
 *   ptr : 
 *
 * RETURN
 *   value describing the string, or NULL if not found
 *
 * NOTES
 *   this checks all existing lookup tables to see if a substring
 *   matches the passed in string.
 */
value_t *pfile_string_find(pfile_t *pf, size_t sz, const char *ptr)
{
  variable_t *var;
  value_t    *val;
  size_t ii;

  val = 0;
  ii  = 0;
  for (var = variable_list_head(&pf->var_active); 
       var; 
       var = variable_link_get(var)) {
    size_t jj;

    jj = 0;
    if (variable_flag_test(var, VARIABLE_FLAG_CONST)
        && (variable_ct_get(var) >= sz)) {

      for (ii = 0; ii <= variable_ct_get(var) - sz; ii++) {

        for (jj = 0; 
             (jj < sz) && (ptr[sz - jj - 1] == variable_const_get(var, jj + ii)); 
             jj++)
          ; /* null body */
        if (jj == sz) {
          break;
        }
      }
    }
    if (jj == sz) {
      break;
    }
  }
  if (var) {
    result_t rc;

    rc = value_alloc(&val, var, 0, 0, variable_sz_get(var));
    if (result_ok != rc) {
      pfile_log_syserr(pf, rc);
    } else {
      value_ct_ofs_set(val, ii);
      value_ct_set(val, sz);
    }
  }
  return val;

}

boolean_t pfile_flag_test(pfile_t *pf, flag_t flag)
{
  return (pf) ? ((pf->flags & flag) != 0) : 0;
}

void pfile_flag_set(pfile_t *pf, flag_t flag)
{
  pf->flags |= flag;
}

void pfile_flag_clr(pfile_t *pf, flag_t flag)
{
  pf->flags &= ~flag;
}

label_t *pfile_isr_preamble_entry_get(pfile_t *pf)
{
  return pf->label_isr_preamble;
}

void pfile_label_fixup(pfile_t *pf)
{
  label_t *lbl;

  for (lbl = label_list_head(&pf->label_user); 
       lbl;
       lbl = label_next_get(lbl)) {
    if (label_usage_get(lbl)) {
      label_flag_set(lbl, LABEL_FLAG_USED);
    }
  }
}

oscillator_t pfile_oscillator_get(pfile_t *pf)
{
  return pf->osc;
}

void pfile_oscillator_set(pfile_t *pf, oscillator_t osc)
{
  pf->osc = osc;
}

pfile_source_t *pfile_source_get(pfile_t *pf)
{
  return pf->src;
}

void pfile_source_restore(pfile_t *pf, pfile_source_t *src)
{
  if (pf->src) {
    pfile_source_release(pf->src);
  }
  pf->src = src;
}

result_t pfile_source_set(pfile_t *pf, const char *fname)
{
  result_t rc;

  rc = pfile_source_open(&pf->src, fname, pf);
  if (result_ok == rc) {
    pfile_rewind(pf);
  }
  return rc;
}

