/************************************************************
 **
 ** pfile.h : p-code file declarations
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#ifndef pfile_h__
#define pfile_h__

#include <stdio.h>

#include "ptype.h"
#include "variable.h"
#include "label.h"
#include "expr.h"
#include "cmd.h"

#define PFILE_FLAG_NORICK 0x0001 /* not using Rick Farmer's PIC loader */
#define PFILE_FLAG_DEBUG  0x0002 /* debugging */
#define PFILE_FLAG_QUIET  0x0004 /* don't show status */
#define PFILE_FLAG_NOWARN 0x0008 /* do not emit warnings */

/* this is a *really* useful way to make sure the format
   and arguments to a function match */
#ifdef __GNUC__
#define ATTRIBUTE __attribute__ ((__format__ (__printf__, 3, 4)))
#else
#define ATTRIBUTE
#endif

/* this doesn't really belong here, but i don't know a better place */
typedef enum {
  oscillator_crystal,
  oscillator_rc
} oscillator_t;

#define PFILE_LENGTH_DEF ((size_t) -1)
result_t pfile_open(pfile_t **dst, FILE *f_asm, FILE *f_lst,
    FILE *f_hex, FILE *f_log, flag_t flags);
void pfile_close(pfile_t *pf);
void pfile_reset(pfile_t *pf);

typedef enum pfile_log_ {
  pfile_log_none,
  pfile_log_crit,
  pfile_log_err,
  pfile_log_warn,
  pfile_log_info,
  pfile_log_debug
} pfile_log_t;

int pfile_ch_get(pfile_t *pf);
void pfile_ch_unget(pfile_t *pf, int ch);

oscillator_t pfile_oscillator_get(pfile_t *pf);
void         pfile_oscillator_set(pfile_t *pf, oscillator_t osc);
/* unnamed constant manipulation */
result_t pfile_constant_get(pfile_t *pf, variable_const_t n, value_t **dst);

/* value manipulation */
result_t pfile_value_temp_get(pfile_t *pf, variable_sz_t sz, flag_t flags, 
    value_t **dst);
result_t pfile_value_alloc(pfile_t *pf, const char *name, flag_t flags,
    variable_sz_t sz, variable_ct_t ct, const variable_const_t *n, value_t **dst);
value_t *pfile_value_find(pfile_t *pf, pfile_log_t plog, size_t sz,
  const char *name);
value_t *pfile_string_find(pfile_t *pf, size_t sz, const char *str);

/* variable manipulation */
result_t pfile_variable_alloc(pfile_t *pf, const char *name, flag_t flags,
    variable_sz_t sz, variable_ct_t ct, 
    const variable_const_t *data, variable_t **dst);
variable_t *pfile_variable_find(pfile_t *pf, pfile_log_t plog, size_t sz,
  const char *name);
boolean_t pfile_variable_exists(pfile_t *pf, size_t sz, const char *name);
void pfile_variable_fixup(pfile_t *pf);
void pfile_variable_dump(pfile_t *pf);
void pfile_write_hex(pfile_t *pf, unsigned pc, unsigned char data);

/* label manipulation */
result_t pfile_label_alloc(pfile_t *pf, const char *name, 
  label_free_cb_t free_fn, label_t **dst);
label_t *pfile_label_find(pfile_t *pf, pfile_log_t plog, const char *name);
void pfile_label_dump(pfile_t *pf);
void pfile_label_fixup(pfile_t *pf);

/* expression manipulaton */
result_t pfile_expr_push(pfile_t *pf, expr_t **stk, value_t *val,
    operator_t op);
void pfile_expr_dump(pfile_t *pf);

/* command list manipulation */
void     pfile_cmd_add(pfile_t *pf, cmd_t *cmd);
void     pfile_cmd_op_add(pfile_t *pf, label_t *label,
  operator_t op, value_t **dst, value_t *val1, value_t *val2);
void     pfile_cmd_branch_add(pfile_t *pf, label_t *label,
  cmd_branchtype_t type, cmd_branchcond_t cond, label_t *dst,
  value_t *var);
void   pfile_cmd_special_add(pfile_t *pf, label_t *lbl, cmd_type_t type);
void   pfile_cmd_label_add(pfile_t *pf, label_t *label);
void   pfile_cmd_dump(pfile_t *pf);
cmd_t *pfile_cmdlist_get(pfile_t *pf);
void   pfile_cmdlist_set(pfile_t *pf, cmd_t *cmd);

/* misc bits */
label_t *pfile_lblptr_get(pfile_t *pf);
void     pfile_lblptr_set(pfile_t *pf, label_t *lbl);
void     pfile_rewind(pfile_t *pf);
void     pfile_log(pfile_t *pf, pfile_log_t sev, const char *fmt, ...)
     ATTRIBUTE;
void     pfile_log_syserr(pfile_t *pf, result_t rc);
unsigned pfile_errct_get(const pfile_t *pf);
unsigned pfile_line_get(const pfile_t *pf);

void pfile_isr_preamble(pfile_t *pf);

void     pfile_user_entry_set(pfile_t *pf, label_t *lbl);
label_t *pfile_user_entry_get(pfile_t *pf);
void     pfile_isr_entry_set(pfile_t *pf, label_t *lbl);
label_t *pfile_isr_entry_get(pfile_t *pf);

label_t *pfile_isr_preamble_entry_get(pfile_t *pf);

typedef enum pfile_write_ {
  pfile_write_asm, /* out the the `.asm' file */
  pfile_write_lst  /* out to the `.lst' file  */
} pfile_write_t;

int pfile_write(pfile_t *pf, pfile_write_t dst, const char *fmt, ...);

boolean_t pfile_flag_test(pfile_t *pf, flag_t flag);
void      pfile_flag_set(pfile_t *pf, flag_t flag);
void      pfile_flag_clr(pfile_t *pf, flag_t flag);

pfile_source_t       *pfile_source_get(pfile_t *pf);
void                  pfile_source_restore(pfile_t *pf, pfile_source_t *src);
result_t              pfile_source_set(pfile_t *pf, const char *fname);
#ifdef MSDOS
#include "pfiledcl.h"
#endif

#endif /* pfile_h__ */

