/************************************************************
 **
 ** pic.h : PIC code generation definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#ifndef _pic_h__
#define _pic_h__

#include "pic_inst.h"

/* a note about bitstate:
     unknown       : has not yet been analyzed
     set           : is known to be set
     clear         : is known to be clear
     indeterminate : sometimes set, sometimes clear, it will
                     need to be explicitly put into a real state */
typedef enum pic_bitstate_ {
  pic_bitstate_unknown,
  pic_bitstate_set,
  pic_bitstate_clr,
  pic_bitstate_indeterminate
} pic_bitstate_t;

/* general notes:
     special handling : btfsc, btfss, incfsz, decfsz
       since these skip the *next* instruction, i'll need to
       peek ahead & make any changes to irp, rp1:rp0, or pclath4:pclath3
       before this instruction
     futzing: even though each PIC instruction is only one instruction
       long, I'm going to change it to 1...6 instructions long
       & pretend that setting the various bits is part of the instruction
       instead of their own instructions

   pass 1: simply generate the instruction list, leaving all bits
           marked ``unknown''
   pass 2: analyze the real bit values as follows:
           * at a label that follows an absolute GOTO or RETURN,
             pclath<4:3> are known. to find the others we'll need to
             look at every path to this label
           * at any other label pclath<4:3> may not be known
           * when a CALL is placed, we'll need to find all RETURNs to 
             determine the bit states
           this looks like it could be a multi-pass proposition!
*/
typedef struct {
  unsigned irp:2;
  unsigned rp1:2;
  unsigned rp0:2;
#if 0
  pic_bitstate_t irp;
  pic_bitstate_t rp1;
  pic_bitstate_t rp0;
#endif
} pic_databits_t;

typedef struct {
  unsigned pclath4:2;
  unsigned pclath3:2;
#if 0
  pic_bitstate_t pclath4;
  pic_bitstate_t pclath3;
#endif
} pic_branchbits_t;

typedef unsigned short pic_pc_t;

typedef struct pic_code_ {
  struct pic_code_ *link;
  cmd_t            *cmd;
  pic_pc_t          pc;       /* program counter */

  /* databits are modified by variable reference */
  struct {
    pic_databits_t before; /* bits on entry to the instruction */
    pic_databits_t action; /* bits modified by the instruction */
  } databits;

  /* branchbits are modified by branching instructions */
  struct {
    pic_branchbits_t before; /* bits on entry to the instruction */
    pic_branchbits_t action; /* bits modified by the instruction */
  } branchbits;

  label_t       *lbl;
  pic_opcode_t       op;
  /* unsigned char  op; */
  /* pic_opcode_t       op; */
  /* pic_instr_append_f_d   */
  /* pic_instr_append_reg_d */
  /* pic_instr_append_f     */
  /* pic_instr_append_reg   */
  value_t  *val; /* this *MUST BE* a variable (not a constant) */
  size_t    ofs;
  pic_opdst_t dst;          /* f, w or none   */
  /* pic_instr_append_f_b      */
  /* pic_instr_append_reg_bn   */
  /* pic_instr_append_reg_flag */
  /* pic_instr_append_w_k      */
  /* pic_instr_append_w_kn     */
  value_t  *literal; /* bit or literal */
  /* pic_instr_append_n        */
  label_t  *brdst;
} pic_code_t;


void     p16f877_init(pfile_t *pf);
void     pic_cmd_alloc(pfile_t *pf, cmd_t *cmd, size_t ct, variable_t **array);
void     pic_cmd_dump(pfile_t *pf, const cmd_t *cmd, boolean_t first);
void     pic_code_free(pfile_t *pf);

value_t *pic_accum_get(pfile_t *pf, unsigned n, variable_sz_t sz,
    flag_t flags);
value_t *pic_loop_get(pfile_t *pf);
value_t *pic_sign_get(pfile_t *pf);

extern variable_sz_t mul_sz;
extern variable_sz_t div_sz;

pic_code_t *pic_code_alloc(void);
pic_code_t *pic_code_get(pfile_t *pf, label_t *lbl, pic_opcode_t op);

boolean_t pic_bitstate_changed(pic_bitstate_t old, pic_bitstate_t new);
#define pic_bitstate_fixup(dst, src) \
  if (pic_bitstate_unknown == dst) {  \
    dst = src;                       \
  } else if (dst != src) {           \
    dst = pic_bitstate_indeterminate; \
  }
pic_code_t *pic_code_label_find(pfile_t *pf, const label_t *lbl);
extern pic_code_t *pic_code_lst; /* list of codes */

boolean_t pic_opcode_is_conditional(pic_opcode_t op);

void pic_branch(pfile_t *pf, cmd_branchtype_t brtype, 
    cmd_branchcond_t brcond, label_t *dst, value_t *val);

unsigned pic_code_to_pcode(pfile_t *pf, const pic_code_t *code);

typedef struct pic_bank_info_ {
  struct pic_bank_info_ *link;
  size_t                 lo;    /* first accessible location in a bank */
  size_t                 hi;    /* last accessible location in a bank  */
  size_t                 next;  /* next accessible location in a bank  */
} pic_bank_info_t;

extern pic_bank_info_t *pic_blist;

typedef unsigned pic_mem_size_t;
extern pic_mem_size_t pic_mem_size;
extern pic_mem_size_t pic_mem_size_loader;

void pic_init(pfile_t *pf);
void pic_blist_add(pfile_t *pf, pic_bank_info_t **pblist, 
    size_t lo, size_t hi);

#endif /* _pic_h__ */

