/************************************************************
 **
 ** pic_brop.c : PIC branch optimization definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <string.h>

#include "pic.h"
#include "pic_brop.h"

static int pic_code_branchbits_changed(const pic_code_t *code,
  const pic_branchbits_t *branchbits)
{
  return pic_bitstate_changed(code->branchbits.before.pclath4, 
        branchbits->pclath4)
    || pic_bitstate_changed(code->branchbits.before.pclath3, 
        branchbits->pclath3);
}

/*
 * NAME
 *   pic_code_branchbits_analyze
 *
 * DESCRIPTION
 *   branch analyzing & fixing
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   none
 *
 * NOTES
 *   I'm sure there's an easier way, but I cannot figure it out, so here
 *   goes:
 *     1st -- analyze the known state the same way its done for the
 *            data bits
 *     2nd -- insert any necessary pic_opcode_code[lo|hi]_[set|clr] instructions
 *     3rd -- if any work was done in (2), then repeat with (1)
 *   This will require a minimum of two analysis runs and doesn't yet
 *   deal with the lookup tables which I think I'll need to move to the
 *   front.
 */
static unsigned max_depth;
void pic_code_branchbits_analyze(pfile_t *pf, pic_code_t *pcode,
    const pic_branchbits_t *pbranchbits, unsigned depth)
{
  pic_branchbits_t branchbits;
  pic_code_t      *code;
  pic_code_t      *code_nxt;
  int              code_pv_cond;

  branchbits = *pbranchbits;
  if (depth > max_depth) {
    max_depth = depth;
  }
  code         = pcode;
  code_pv_cond = 0;
  /* if the branchbits have not changed, we need not continue */
  for (code = pcode;
       code
       && pic_code_branchbits_changed(code, &branchbits);
       code = code_nxt) {
    code_nxt = code->link;

    /* setup this code */
    pic_bitstate_fixup(code->branchbits.before.pclath4, branchbits.pclath4);
    pic_bitstate_fixup(code->branchbits.before.pclath3, branchbits.pclath3);

    if (!code_pv_cond
      && ((pic_opcode_ret == code->op)
        || (pic_opcode_retlw == code->op)
        || (pic_opcode_retfie == code->op))) {
      break; /* <--- we've hit the end of this path */
    }

    if ((pic_opcode_call == code->op) || (pic_opcode_goto == code->op)) {
      /* find the label */
      pic_code_t    *dcode;
      pic_bitstate_t pclath4;
      pic_bitstate_t pclath3;
      label_t       *brdst;

      brdst = code->brdst;
      if (brdst) {
        pclath3 = (label_pc_get(brdst) & (1 << 11))
          ? pic_bitstate_set : pic_bitstate_clr;
        pclath4 = (label_pc_get(brdst) & (1 << 12))
          ? pic_bitstate_set : pic_bitstate_clr;
        branchbits.pclath4 = pclath4;
        if (code->branchbits.before.pclath4 != pclath4) {
          code->branchbits.action.pclath4 = pclath4;
        }
        branchbits.pclath3 = pclath3;
        if (code->branchbits.before.pclath3 != pclath3) {
          code->branchbits.action.pclath3 = pclath3;
        }
        dcode = pic_code_label_find(pf, code->brdst);
        if (pic_opcode_goto == code->op) {
          if (code_pv_cond) {
            pic_code_branchbits_analyze(pf, dcode, &branchbits, depth + 1);
          } else {
            code_nxt = dcode;
          }
        } else {
          /* this is *really* wimpy, but for the moment I'm
             just going to say after any call the result is unknown */
          pic_code_branchbits_analyze(pf, dcode, &branchbits, depth + 1);
#if 1
          branchbits.pclath4 = pic_bitstate_indeterminate;
          branchbits.pclath3 = pic_bitstate_indeterminate;
#else
          branchbits.pclath4 = (pic_pcl < 4096)
            ? pic_bitstate_clr : pic_bitstate_indeterminate;
          branchbits.pclath3 = (pic_pcl < 2048)
            ? pic_bitstate_clr : pic_bitstate_indeterminate;
#endif
        }
      }
    } else if (pic_opcode_branchlo_set == code->op) {
      branchbits.pclath3 = pic_bitstate_set;
    } else if (pic_opcode_branchlo_clr == code->op) {
      branchbits.pclath3 = pic_bitstate_clr;
    } else if (pic_opcode_branchhi_set == code->op) {
      branchbits.pclath4 = pic_bitstate_set;
    } else if (pic_opcode_branchhi_clr == code->op) {
      branchbits.pclath4 = pic_bitstate_clr;
    }
    if (code_nxt == code->link) {
      code_pv_cond = pic_opcode_is_conditional(code->op);
    } else {
      code_pv_cond = 0;
    }
  }
}

/*
 * NAME
 *   pic_code_insert
 *
 * DESCRIPTION
 *   insert a code
 *
 * PARAMETERS
 *   pf  : pfile
 *   pos : where
 *   pop : what
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static unsigned pic_bropcode_insert(pfile_t *pf, pic_code_t **pv,
  pic_bitstate_t state, pic_opcode_t set, pic_opcode_t clr)
{
  pic_code_t *code_pv;
  unsigned    ins_ct;
  
  code_pv = *pv;

  ins_ct = 0;
  if (pic_bitstate_unknown != state) {
    pic_opcode_t op;

    op = (pic_bitstate_set == state) ? set : clr;
    if (code_pv
      && ((set == code_pv->link->op)
        || (clr == code_pv->link->op))) {
      /* simply change this instruction */
      code_pv     = code_pv->link;
      code_pv->op = op;
    } else {
      /* insert a new instruction */
      pic_code_t *ncode;

      ins_ct = 1;
      ncode = pic_code_alloc();
      if (ncode) {
        memset(ncode, 0, sizeof(*ncode));
        if (code_pv) {
          ncode->link = code_pv->link;
          code_pv->link = ncode;
        } else {
          ncode->link = pic_code_lst;
          pic_code_lst = ncode;
        }
        ncode->cmd = ncode->link->cmd;
#if 0
        /* unlike the data bits, we needn't move the label for the
         * because there are only two possible conditions:
         *    1. someone has branched to this label, in which case
         *       the branch bits are set correctly
         *    2. the code has simply run into this label, in which case
         *       the only important thing is the branch bits are set
         *       correctly. */
        ncode->lbl = ncode->link->lbl;
        ncode->link->lbl = 0;
#endif
        ncode->op = op;
      } else {
        pfile_log_syserr(pf, result_memory);
      }
      code_pv = ncode;
    }
  }
  *pv = code_pv;
  return ins_ct;
}

/*
 * NAME
 *   pic_branchbits_action_fixup
 *
 * DESCRIPTION
 *   make any necessary changes to the pclath bits for the actions
 *
 * PARAMETERS
 *   pf : pfile
 *
 * RETURN
 *   0     : no changes needed
 *   non-0 : changes needed (the whole process must be repeated)
 *
 * NOTES
 *   This probably isn't the most efficient method, but here goes:
 *   If the *previous* command is a conditional, the action
 *     happens before the previous command
 *   Otherwise, look to see if the necessary commands exist, if
 *     so, modify as necessary. The HI bits will always be set first,
 *     followed by the low bits
 *   Finally, insert any needed commands fixing up the labels as
 *     necessary.
 */
unsigned pic_branchbits_action_fixup(pfile_t *pf)
{
  pic_code_t *code_pv;
  pic_code_t *code;
  unsigned    ins_ct;  /* # of instructions inserted */

  ins_ct  = 0;
  code_pv = 0;
  code    = pic_code_lst;
  /* code_pv points to:
       * the code before a conditional in the event of
             cond [, cond ...] branch
       * the code before a ``pic_opcode_branch*''
       * the code immediatly preceding code
     code_pv->link is always valid */
  while (code) {
    ins_ct += pic_bropcode_insert(pf, &code_pv,
      code->branchbits.action.pclath4,
      pic_opcode_branchhi_set, pic_opcode_branchhi_clr);
    ins_ct += pic_bropcode_insert(pf, &code_pv,
      code->branchbits.action.pclath3,
      pic_opcode_branchlo_set, pic_opcode_branchlo_clr);
    code_pv = code;
    do {
      code = code->link;
    } while (code
      && ((pic_opcode_branchhi_set == code->op)
        || (pic_opcode_branchhi_clr == code->op)
        || (pic_opcode_branchlo_set == code->op)
        || (pic_opcode_branchlo_clr == code->op)
        || pic_opcode_is_conditional(code->op)));
  }
  return ins_ct;
}

