/************************************************************
 **
 ** pic_daop.c : PIC data optimization definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "pic.h"
#include "pic_var.h"
#include "pic_daop.h"

/*
 * NAME
 *   pic_code_databits_changed
 *
 * DESCRIPTION
 *   determine if the databits have changed
 *
 * PARAMETERS
 *   code     : code to examine
 *   databits : expected databits
 *
 * RETURN
 *   0 : the bits have changed
 *   1 : the bits have not changed
 *
 * NOTES
 *   a bit is incorrect if it is determinate and not equal to the
 *   its match
 */

static int pic_code_databits_changed(const pic_code_t *code,
  const pic_databits_t *databits)
{
  return pic_bitstate_changed(code->databits.before.irp, databits->irp)
    || pic_bitstate_changed(code->databits.before.rp1, databits->rp1)
    || pic_bitstate_changed(code->databits.before.rp0, databits->rp0);
}

/*
 * NAME
 *   pic_code_databits_analyze
 *
 * DESCRIPTION
 *   analyze the databits in the pic_codes[] array
 *
 * PARAMETERS
 *   pf       : 
 *   pcl      : pcl on entry
 *   databits : current databits
 *   depth    : depth (for statistics only)
 *
 * RETURN
 *   none
 *
 * NOTES
 *   This follows all codepaths to determine when
 *   status<irp> and status<rp1:rp0> need to be changed, thus minimizing
 *   these changes.
 *   Note about calls: the statement after a call is assumed to
 *     have lost all information about the databits. This is unfortunate,
 *     but it is very difficult to accurately detemine without building
 *     a rather large and complex tree. This is something I'll work on
 *     when I've the chance. The main problem I've run into is how to
 *     detect loops.
 */
static unsigned max_depth;
void pic_code_databits_analyze(pfile_t *pf, pic_code_t *pcode,
  const pic_databits_t *pdatabits, unsigned depth)
{
  pic_databits_t databits;
  pic_code_t    *code;
  pic_code_t    *code_nxt;
  int            code_pv_cond;

  databits = *pdatabits;
  if (depth > max_depth) {
    max_depth = depth;
  }
  code_pv_cond = 0;
  /* if the databits have not changed, we need not continue */
  for (code = pcode;
       code
       && pic_code_databits_changed(code, &databits);
       code = code_nxt) {
    code_nxt = code->link;

    /* setup this code */
    pic_bitstate_fixup(code->databits.before.irp, databits.irp);
    pic_bitstate_fixup(code->databits.before.rp1, databits.rp1);
    pic_bitstate_fixup(code->databits.before.rp0, databits.rp0);

    /* if this does not follow a conditional, and it's a return,
       then we've hit the end of this path */
    if (!code_pv_cond
      && ((pic_opcode_ret == code->op)
        || (pic_opcode_retlw == code->op)
        || (pic_opcode_retfie == code->op))) {
      break; /* <--- we've hit the end of this path */
    }

    if (code->val) {
      /* make sure the databits are correctly set */
      variable_t *var;
      size_t      n;

      if (value_subscript_get(code->val)) {
        /* setup status<irp> */
      } else {
        /* setup status<rp1:rp0> */
        pic_bitstate_t rp0;
        pic_bitstate_t rp1;

        var = value_variable_get(code->val);
        n = variable_base_get(var)       /* start at the base         */
          + pic_var_ofs_get(pf, code->val, code->ofs);

        /* first, set these as if the variable is 
           only available in one region */
        rp1 = ((n >> 8) & 0x01) ? pic_bitstate_set : pic_bitstate_clr;
        rp0 = ((n >> 7) & 0x01) ? pic_bitstate_set : pic_bitstate_clr;
        if (variable_region_test(var, 0) && variable_region_test(var, 1)) {
          /* the lower bit can be anywhere because this variable exists
           * in both conditions of the lower bit */
          rp0 = pic_bitstate_unknown;
        } else {
          databits.rp0 = rp0;
        }
        if ((pic_bitstate_unknown != rp0)
            && (code->databits.before.rp0 != rp0)) {
          code->databits.action.rp0 = rp0;
        }
        if (variable_region_test(var, 2) && variable_region_test(var, 3)) {
          /* the upper bit can be anywhere because this variable exists
           * in both conditions of the upper bit */
          rp1 = pic_bitstate_unknown;
        } else {
          databits.rp1 = rp1;
        }
        if ((pic_bitstate_unknown != rp1)
            && (code->databits.before.rp1 != rp1)) {
          code->databits.action.rp1 = rp1;
        }
      }
    }
    if ((pic_opcode_call == code->op) || (pic_opcode_goto == code->op)) {
      /* find the label */
      pic_code_t *dcode;

      dcode = pic_code_label_find(pf, code->brdst);
      if (!dcode) {
        pfile_log(pf, pfile_log_err, PIC_MSG_LABEL_NOT_FOUND, 
          label_name_get(code->brdst));
        /* abort(); */
      } else {
        if (pic_opcode_goto == code->op) {
          if (code_pv_cond) {
            pic_code_databits_analyze(pf, code, &databits, depth + 1);
          } else {
            code_nxt = dcode;
          }
        } else {
          /* this is *really* wimpy, but for the moment I'm
             just going to say after any call the result is unknown */
          pic_code_databits_analyze(pf, dcode, &databits, depth + 1);
          databits.irp = pic_bitstate_indeterminate;
          databits.rp1 = pic_bitstate_indeterminate;
          databits.rp0 = pic_bitstate_indeterminate;
        }
      }
    } else if (pic_opcode_datalo_set == code->op) {
      databits.rp0 = pic_bitstate_set;
    } else if (pic_opcode_datalo_clr == code->op) {
      databits.rp0 = pic_bitstate_clr;
    } else if (pic_opcode_datahi_set == code->op) {
      databits.rp1 = pic_bitstate_set;
    } else if (pic_opcode_datahi_clr == code->op) {
      databits.rp1 = pic_bitstate_clr;
    }
    if (code_nxt != code->link) {
      code_pv_cond = 0;
    } else {
      code_pv_cond = pic_opcode_is_conditional(code->op);
    }
  }
}

/*
 * NAME
 *   pic_code_post_analyze
 *
 * DESCRIPTION
 *   post analysis cleanup
 *
 * PARAMETERS
 *   pf : 
 *
 * RETURN
 *   none
 *
 * NOTES
 *   Once the analysis procedure has run, it's possible
 *   that the labels don't really know the correct state
 *   of the bits. Because of this, we need to find each
 *   branch & make sure everything is correct
 */
#define ISACTION(x) ((pic_bitstate_set == (x)) || (pic_bitstate_clr == (x)))
static unsigned pic_daopcode_insert(pfile_t *pf, pic_code_t **pv,
    pic_bitstate_t bit, pic_opcode_t set, pic_opcode_t clr)
{
  pic_code_t *ncode;
  pic_code_t *code_pv;
  unsigned    added;

  code_pv = *pv;

  added = 0;
  /* insert the necessary commands here */
  if (pic_bitstate_unknown != bit) {
    /* look at the code before, if its conditional we've failed! */
    if (code_pv && pic_opcode_is_conditional(code_pv->op)) {
      pfile_log(pf, pfile_log_err, PIC_MSG_CODEGEN_ERROR,
          cmd_line_get(code_pv->cmd));
    }
    ncode = pic_code_alloc();
    if (!ncode) {
      pfile_log_syserr(pf, result_memory);
    } else {
      ncode->link = code_pv->link;
      code_pv->link = ncode;
      ncode->cmd = ncode->link->cmd;
      ncode->op  = (pic_bitstate_set == bit) ? set : clr;
      ncode->lbl = ncode->link->lbl;
      ncode->link->lbl = 0;
      code_pv = ncode;
      added++;
    }
  }
  *pv = code_pv;
  return added;
}

unsigned pic_code_databits_post_analyze(pfile_t *pf)
{
  int         ismod;
  unsigned    loopct;
  unsigned    added;
  pic_code_t *code;
  pic_code_t *code_pv;

  loopct = 0;
  added  = 0;
  do {
    ++loopct;
    ismod = 0;
    for (code = pic_code_lst; code; code = code->link) {
      pic_code_t *dcode;

      dcode = pic_code_label_find(pf, code->brdst);

      if (dcode
        && pic_code_databits_changed(dcode, &code->databits.before)) {
        ismod = 1;
        pic_code_databits_analyze(pf, dcode, &code->databits.before, 1);
      }
    }
  } while (ismod);
  /* finally, insert any needed <rp1:rp0> or <irp> changes */
  for (code_pv = 0, code = pic_code_lst; 
       code; 
       code_pv = code, code = code->link) {
    added += pic_daopcode_insert(pf, &code_pv, 
        code->databits.action.rp1, pic_opcode_datahi_set,
        pic_opcode_datahi_clr);
    added += pic_daopcode_insert(pf, &code_pv,
        code->databits.action.rp0, pic_opcode_datalo_set,
        pic_opcode_datalo_clr);
  }
  return added;
}


