/************************************************************
 **
 ** pic_op.c : PIC operator generation definitions
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include <assert.h>

#include "pic.h"
#include "pic_var.h"
#include "pic_opfn.h"
#include "pic_op.h"

/*
 * NAME
 *   pic_accum_from_val
 *
 * DESCRIPTION
 *   get an accumulator that matches val
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 */
static value_t *pic_accum_from_val(pfile_t *pf, unsigned n, value_t *src)
{
  return pic_accum_get(pf, n, value_sz_get(src), value_flags_get_all(src));
}

/*
 * NAME
 *   array_setup
 *
 * DESCRIPTION
 *   setup fsr to point into an array
 *
 * PARAMETERS
 *   pf    : file
 *   val   : value
 *   atend : non-0 to position at the end of the array, 0 for start
 *
 * RETURN
 *
 * NOTES
 *   this can also be used for non-arrays
 */
typedef enum {
  array_setup_start,
  array_setup_end
} array_setup_t;

static void array_setup(pfile_t *pf, value_t *val, array_setup_t where)
{
  size_t   sz;

  /* simple assumption for the moment -- arrays must not span banks */
  assert(!value_is_lookup(val));

  if (!pic_var_spans_banks(pf, value_variable_get(val))) {
    pic_instr_append_w_k(pf, pic_opcode_movlw, val);
    pic_instr_append_reg_flag(pf, 
       (variable_base_get(value_variable_get(val)) >= 256) 
       ? pic_opcode_bsf : pic_opcode_bcf,
       "_status", "_irp");
    sz = 0;
    if (value_baseofs_get(val)) {
      sz = value_baseofs_get(val);
    }
    if (array_setup_end == where) {
      sz += value_sz_get(val) - 1;
    }
    if (sz) {
      pic_instr_append_w_kn(pf, pic_opcode_addlw, sz);
    }
    if (value_subscript_get(val)
        && (VALUE_SUBSCRIPT_DUMMY != value_subscript_get(val))) {
      value_t *sub;
      int      rc;

      rc = value_alloc(&sub, value_subscript_get(val), 0, 0, value_sz_get(val));
      if (0 != rc) {
        pfile_log_syserr(pf, rc);
      } else {
        pic_instr_append_f_d(pf, pic_opcode_addwf, sub, 0, pic_opdst_w);
        value_release(sub);
      }
    }
    val = pfile_value_find(pf, pfile_log_err, PFILE_LENGTH_DEF, "_fsr");
    if (val) {
      pic_instr_append_f(pf, pic_opcode_movwf, val, 0);
      value_release(val);
    }
  } else {
    /* ick. this needs to be calculated! basically, it's the same code
       seen in pic_var_ofs_get() only done at run-time. Accum4 & 3 are
       used for this, as well as a function call to do the dirty work.
       accum4 = is set to the dimension, and accum3 the highest value
       uses in the first bank */
    variable_t     *var;
    variable_base_t base;
    value_t        *val_ss;
    value_t        *val_tmp;
    value_t        *val_base;
    char            label_name[32];
    label_t        *lbl;

    var = value_variable_get(val);
    base = pic_var_ofs_get(pf, val, 0);
    /* accum4 = subscript */
    value_alloc(&val_ss, value_subscript_get(val), 0, 0, 2);
    val_tmp = pic_accum_get(pf, 4, 2, 0);
    pfile_constant_get(pf, variable_base_get(var), &val_base);
    pic_op(pf, operator_add, val_tmp, val_ss, val_base);
    value_release(val_tmp);
    value_release(val_ss);
    /* accum3 = highest used in the first bank */
    val_tmp = pic_accum_get(pf, 3, 2, 0);
    pfile_constant_get(pf, pic_var_firstbank_highest_get(pf, var) + 1, &val_ss);
    pic_op(pf, operator_assign, val_tmp, val_ss, 0);
    value_release(val_ss);
    value_release(val_tmp);
    /* call the normalize routine */
    pic_instr_append_reg_flag(pf, 
        (base < 256) ? pic_opcode_bcf : pic_opcode_bsf, "_status", "_irp");
    pic_array_setup_label_create(label_name, variable_sz_get(var), 
        pic_var_firstbank_get(pf, var));
    lbl = pfile_label_find(pf, pfile_log_none, label_name);
    if (!lbl) {
      pfile_label_alloc(pf, label_name, 0, &lbl);
    }
    pic_instr_append_n(pf, pic_opcode_call, lbl);
    label_release(lbl);
  }
}

/*
 * NAME
 *   pic_opcode_set_to_w
 *
 * DESCRIPTION
 *   set dst to the value in w, zero padding as necessary
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   w is expected to be 0 or 1
 */
static void pic_opcode_set_to_w(pfile_t *pf, value_t *dst)
{
  size_t ii;
  int    dst_is_array;

  if (value_subscript_get(dst)) {
    /* pic_array_setup() will destroy w, so lock it away */
    value_t *sign;

    sign = pic_sign_get(pf);
    if (sign) {
      pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
      array_setup(pf, dst, array_setup_start);
      pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
      value_release(sign);
    }
    dst_is_array = 1;
  } else {
    dst_is_array = 0;
  }

  pic_instr_append_f(pf, pic_opcode_movwf, dst, 0);
  for (ii = 1; ii < value_sz_get(dst); ii++) {
    if (dst_is_array) {
      pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
    }
    pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
  }
}

/*
 * NAME
 *   sign_extend
 *
 * DESCRIPTION
 *   pad a value with either 0, or 255 if signed and the sign bit
 *   is set
 *
 * PARAMETERS
 *   pf       :
 *   val      : 
 *   pos      : offset into val (for non-array)
 *   issigned : non-0 if sign extension is needed
 *
 * RETURN
 *
 * NOTES
 *   for an array, ``fsr'' is expected to be set to the msb
 */
static void sign_extend(pfile_t *pf, value_t *val, size_t pos, int issigned)
{
  pic_opcode_t op;

  if (issigned) {
    pic_instr_append(pf, pic_opcode_clrw);
    pic_instr_append_f_bn(pf, pic_opcode_btfsc, val, pos - 1, 7);
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
    op = pic_opcode_movwf;
  } else {
    op = pic_opcode_clrf;
  }
  while (pos < value_sz_get(val)) {
    if (ISARRAY(val)) {
      pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
    }
    pic_instr_append_f(pf, op, val, pos);
    pos++;
  }
}

static int need_sign(const value_t *val1, const value_t *val2)
{
  int rc;

  rc = 1;
  if (!ISCONST(val1)) {
    rc &= ISSIGNED(val1);
  }
  if (!ISCONST(val2)) {
    rc &= ISSIGNED(val2);
  }
  return rc;
}

/*
 * NAME
 *   sign_set
 *
 * DESCRIPTION
 *   set the sign register
 *
 * PARAMETERS
 *   pf   :
 *   val  : value
 *   atend: for arrays, non-0 if already positioned at the end
 *
 * RETURN
 *
 * NOTES
 *  sets "sign" to either 0 or 255
 */
static void sign_set(pfile_t *pf, value_t *val, int atend)
{
  value_t *sign;

  sign = pic_sign_get(pf);
  if (sign) {
    if (!ISSIGNED(val)) {
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
    } else if (ISARRAY(val)) {
      /* position to end of value */
      if (!atend) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, value_sz_get(val) - 1);
        pic_instr_append_reg_d(pf, pic_opcode_addwf, "_fsr", pic_opdst_f);
      }
      /* setup sign */
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_reg_bn(pf, pic_opcode_btfsc, "_ind", 7);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
      pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
      /* return to start of value */
      if (!atend) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, value_sz_get(val) - 1);
        pic_instr_append_reg_d(pf, pic_opcode_subwf, "_fsr", pic_opdst_f);
      }
    } else {
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_f_bn(pf, pic_opcode_btfsc, val, value_sz_get(val) - 1, 
          7);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 255);
      pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
    }
  }
  value_release(sign);
}


/************************************************************
 **
 ** unary operations
 **
 ************************************************************/

/*
 * NAME
 *   pic_opcode_assign_lookup
 *
 * DESCRIPTION
 *   assign a lookup value
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   a lookup value is a non-constant subscript to a constant
 *   array. multi byte results are done as series of single
 *   byte lookups as that's what the PIC can do
 */
static void pic_opcode_assign_lookup(pfile_t *pf, value_t *dst,
  value_t *src, operator_t op)
{
  variable_sz_t   ii;
  variable_base_t bofs;
  value_t        *subscript;
  size_t          sz;
  int             rc;

  rc = value_alloc(&subscript, value_subscript_get(src), 0, 0, 1);
  if (0 != rc) {
    pfile_log_syserr(pf, rc);
  } else {
    if (ISARRAY(dst)) {
      array_setup(pf, dst, array_setup_start);
    }
    sz = value_sz_get(src);
    if (sz > value_sz_get(dst)) {
      sz = value_sz_get(dst);
    }
    for (ii = 0, bofs = value_baseofs_get(src); ii < sz; ii++, bofs++) {
      label_t *lbl;
      char     lbl_name[256];

      if (ii && ISARRAY(dst)) {
        /* bump the array pointer */
        pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
      }
      if (ii < value_sz_get(src)) {
        /* make sure this isn't a constant. for example, if a dword
         * array is set to {1,2,3,4,5} we needn't produce a lookup table
         * for bytes 2, 3, or 4 since they're always zero */
        variable_sz_t jj;
        variable_t   *vsrc;
        unsigned char vlast;

        vsrc = value_variable_get(src);
        vlast = (variable_const_get(vsrc, ii) >> (8 * bofs)) & 0xff;

        for (jj = 1;
             (jj < value_ct_get(src))
             && (vlast == ((variable_const_get(vsrc, jj) >> (8 * bofs)) 
                 & 0xff));
             jj++)
          ;
        if ((value_ct_get(src) > 1) && (value_ct_get(src) == jj)) {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, vlast);
        } else {
          sprintf(lbl_name, "_lookup_%s_%u", value_name_get(src), bofs);
          lbl = pfile_label_find(pf, pfile_log_none, lbl_name);
          if (!lbl) {
            pfile_label_alloc(pf, lbl_name, 0, &lbl);
          }
          if (lbl) {
            /* OK, this is twisted but we actually need to set _pclath
             * here *always* because even though we twiddle bits 4,3
             * the lookup table *alsy* requires bits 0..2 to be set correctly
             * which means the AN556 following example 2 in section 1 is
             * wrong! */
            pic_instr_append_n(pf, pic_opcode_movlw, lbl);
            pic_instr_append_reg(pf, pic_opcode_movwf, "_pclath");
            pic_instr_append_f_d(pf, pic_opcode_movf, subscript, ii, 
                pic_opdst_w);
            if (value_ct_ofs_get(src)) {
              pic_instr_append_w_kn(pf, pic_opcode_addlw, 
                  value_ct_ofs_get(src));
            }
            pic_instr_append_n(pf, pic_opcode_call, lbl);
            label_release(lbl);
          }
        }
      }
      pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
      if (operator_cmpb == op) {
        pic_instr_append_f_d(pf, pic_opcode_comf, dst, ii, pic_opdst_f);
      }
    }
    if (ii < value_sz_get(dst)) {
      sign_extend(pf, dst, ii, ISSIGNED(src));
    }
    value_release(subscript);
  }
}

/*
 * NAME
 *   pic_opcode_assign_const
 *
 * DESCRIPTION
 *   assign a constant
 *
 * PARAMETERS
 *   pf :
 *   dst:
 *   src:
 *
 * RETURN
 *   none
 *
 * NOTES
 */
static void pic_opcode_assign_const(pfile_t *pf, value_t *dst, 
  variable_const_t src)
{
  variable_sz_t ii;
  unsigned char last_out;

  last_out = -1;
  if (ISARRAY(dst)) {
    array_setup(pf, dst, array_setup_start);
  }
  for (ii = 0; ii < value_sz_get(dst); ii++) {
    unsigned char n;
    pic_opcode_t      op;

    op = pic_opcode_none;
    if (ii) {
      if (ISARRAY(dst)) {
        /* bump the array pointer */
        pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
      }
      src >>= 8;
    }
    n = src & 0xff;
    if (!n) {
      op = pic_opcode_clrf;
    } else {
      if (!ii || (n != last_out)) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, n);
      }
      op = pic_opcode_movwf;
    }
    last_out = n;
    pic_instr_append_f(pf, op, dst, ii);
  }
}

/*
 * NAME
 *   pic_opcode_assign_or_cmpb
 *
 * DESCRIPTION
 *   perform either a simple assignment or a complement & assign
 *
 * PARAMETERS
 *   pf  : 
 *   dst : 
 *   src : source (can be 0)
 *   op  : operator_assign or operator_cmpb
 *
 * RETURN
 *
 * NOTES
 *   assignment : dst = src
 *   complement : dst = ~src
 */
static void pic_opcode_assign_or_cmpb(pfile_t *pf, value_t *dst, value_t *src,
  operator_t op)
{
  if (ISCONST(src)) {
    variable_const_t n;

    n = value_const_get(src);
    pic_opcode_assign_const(pf, dst, n);
    if (!value_flag_test(src, VARIABLE_FLAG_CONST)
      || value_subscript_get(src)) {
      value_flag_set(src, VARIABLE_FLAG_USED);
    }
  } else if (value_is_lookup(src)) {
    pic_opcode_assign_lookup(pf, dst, src, op);
  } else {
    variable_sz_t sz;

    sz = value_sz_get(dst);
    if (src && (sz > value_sz_get(src))) {
      sz = value_sz_get(src);
    }

    if (!src || value_same(dst, src)) {
      if (operator_cmpb == op) {
        /* complement in place */
        variable_sz_t ii;

        if (ISARRAY(dst)) {
          array_setup(pf, dst, array_setup_start);
        }
        for (ii = 0; ii < value_sz_get(dst); ii++) {
          if (ii && ISARRAY(dst)) {
            pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
          }
          pic_instr_append_f_d(pf, pic_opcode_comf, dst, ii, pic_opdst_f);
        }
        if (sz < value_sz_get(dst)) {
          sign_extend(pf, dst, ii, ISSIGNED(src));
        }
      }
    } else if (src) {
      variable_sz_t ii;
      pic_opcode_t      pop;

      pop = (operator_assign == op)
        ? pic_opcode_movf
        : pic_opcode_comf;

      if (ISARRAY(dst) && ISARRAY(src)) {
        /* array = array. not possible on a PIC, so
           change to temp = src; dst = temp         */
        value_t *tmp;

        tmp = pic_accum_from_val(pf, 1, src);
        if (tmp) {
          pic_opcode_assign_or_cmpb(pf, tmp, src, operator_assign);
        }
        src = tmp;
      } else {
        value_lock(src);
      }
      if (src) {
        if (ISARRAY(src)) {
          array_setup(pf, src, array_setup_start);
        } else if (ISARRAY(dst)) {
          array_setup(pf, dst, array_setup_start);
        }
        for (ii = 0; ii < sz; ii++) {
          if (ii && (ISARRAY(src) || ISARRAY(dst))) {
            pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
          }
          pic_instr_append_f_d(pf, pop, src, ii, pic_opdst_w);
          pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
        }
        if (sz < value_sz_get(dst)) {
          sign_extend(pf, dst, ii, ISSIGNED(src));
        }
        value_release(src);
      }
    }
  }
}

#define pic_opcode_assign(pf, dst, src) \
  pic_opcode_assign_or_cmpb(pf, dst, src, operator_assign)
#define pic_opcode_cmpb(pf, dst, src) \
  pic_opcode_assign_or_cmpb(pf, dst, src, operator_cmpb)
#define pic_opcode_logical(pf, dst, src) \
  pic_opcode_logical_not(pf, dst, src, operator_logical)


/*
 * NAME
 *   pic_opcode_incr_decr
 *
 * DESCRIPTION
 *   dst = src +/- 1
 *
 * PARAMETERS
 *   pf : 
 *   dst : 
 *   src : source (can be 0)
 *   op  : operator_incr or operator_decr
 *
 * RETURN
 *
 * NOTES
 *   if dst & src are different, first assign src to dst
 */
static void pic_opcode_incr_decr(pfile_t *pf, value_t *dst, value_t *src,
    operator_t op)
{
  variable_sz_t sz;
  variable_sz_t dst_sz;
  value_t      *tmp;
  variable_sz_t ii;

  tmp = 0;

  sz = value_sz_get(dst);
  if (src && (sz > value_sz_get(src))) {
    sz = value_sz_get(src);
  }
  dst_sz = value_sz_get(dst);

  if ((1 == sz) 
      && (value_same(dst, src) 
        || !(ISARRAY(dst) && ISARRAY(src)))) {
    pic_opcode_t pop;

    pop = (operator_incr == op) ? pic_opcode_incf : pic_opcode_decf;
    /* trivial case */
    if (ISARRAY(src)) {
      array_setup(pf, src, array_setup_start);
    } else if (ISARRAY(dst)) {
      array_setup(pf, dst, array_setup_start);
    }
    if (value_same(src, dst)) {
      pic_instr_append_f_d(pf, pop, src, 0, pic_opdst_f);
    } else {
      pic_instr_append_f_d(pf, pop, src, 0, pic_opdst_w);
      pic_instr_append_f(pf, pic_opcode_movwf, dst, 0);
    }
    if (dst_sz > sz) {
      sign_extend(pf, dst, sz, ISSIGNED(src));
    }
  } else {
    if (src) {
      if (value_flag_test(dst, VARIABLE_FLAG_NOTEMP)) {
        tmp = pic_accum_get(pf, 1, sz, value_flags_get_all(src));
        pic_opcode_assign(pf, tmp, src);
      } else {
        if (value_sz_get(src) < value_sz_get(dst)) {
          value_sz_set(dst, sz);
        }
        pic_opcode_assign(pf, dst, src);
      }
    }

    if (ISARRAY(dst)) {
      /* array increment or decrement */
      pic_opcode_t pop;

      pop = (operator_incr == op) ? pic_opcode_addwf : pic_opcode_subwf;
      array_setup(pf, dst, array_setup_start);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      for (ii = 0; ii < sz; ii++) {
        if (ii) {
          pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
        }
        pic_instr_append_reg_d(pf, pop, "_ind", pic_opdst_f);
      }
    } else if (operator_incr == op) {
      /* simple increment */
      for (ii = 0; ii < sz; ii++) {
        if (ii) {
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
        }
        pic_instr_append_f_d(pf, pic_opcode_incf, dst, ii, pic_opdst_f);
      }
    } else {
      /* simple decrement */
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      for (ii = 0; ii < sz; ii++) {
        if (ii) {
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
        }
        pic_instr_append_f_d(pf, pic_opcode_subwf, dst, ii, pic_opdst_f);
      }
    }
    if (tmp) {
      pic_opcode_assign(pf, dst, tmp);
      value_release(tmp);
    } else if (dst_sz < value_sz_get(dst)) {
      value_sz_set(dst, dst_sz);
      sign_extend(pf, dst, ii, ISSIGNED(src));
    }
  }
}

/*
 * NAME
 *   pic_opcode_logical_not
 *
 * DESCRIPTION
 *   dst = !!src or dst = !src
 *
 * PARAMETERS
 *   pf  :
 *   dst : 
 *   src : source, can be 0
 *
 * RETURN
 *
 * NOTES
 *   if dst is 0, we're just trying to set Z correctly
 */
static void pic_opcode_logical_not(pfile_t *pf, value_t *dst, value_t *src,
    operator_t op)
{
  size_t ii;

  if (!src) {
    src = dst;
  }

  if (ISARRAY(src)) {
    array_setup(pf, src, array_setup_start);
  }
  pic_instr_append_f_d(pf, pic_opcode_movf, src, 0, pic_opdst_w);

  for (ii = 1; ii < value_sz_get(src); ii++) {
    if (ISARRAY(src)) {
      pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
    } 
    pic_instr_append_f_d(pf, pic_opcode_iorwf, src, ii, pic_opdst_w);
  }
  if (dst) {
    if (operator_logical == op) {
      /* now, w is 0 or non-0, z is appropriately set. if non-0, set to 1 */
      pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
    } else {
      /* must invert w */
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
      pic_instr_append(pf, pic_opcode_clrw);
    }
    pic_opcode_set_to_w(pf, dst);
  }
}

/************************************************************
 **
 ** binary operations
 **
 ************************************************************/

/*
 * NAME
 *   pic_opcode_andl_orl
 *
 * DESCRIPTION
 *   dst = !!val1 & !!val2
 *   dst = !!val1 | !!val2
 *
 * PARAMETERS
 *   pf   : 
 *   dst  :
 *   val1 :
 *   val2 : can be 0
 *   op   : operator
 *
 * RETURN
 *
 * NOTES
 */
static void pic_opcode_andl_orl(pfile_t *pf, value_t *dst, 
    value_t *val1, value_t *val2, operator_t op)
{
  value_t *accum1;
  pic_opcode_t pop;

  pop = (operator_andl == op) ? pic_opcode_andwf : pic_opcode_iorwf;

  if (!val2) {
    val2 = val1;
    val1 = dst;
  }

  accum1 = pic_accum_get(pf, 1, 1, 0);
  pic_opcode_logical(pf, accum1, val1);             /* accum1 = !!val1 */
  if (ISCONST(val2)) {
    pic_instr_append_w_kn(pf, pic_opcode_movlw, value_const_get(val2) != 0);
  } else {
    pic_opcode_logical(pf, 0,      val2);             /* w = !!val2      */
  }
  pic_instr_append_f_d(pf, pop, accum1, 0, pic_opdst_w); /* w op= accum1    */
  pic_opcode_set_to_w(pf, dst);                     /* dst = w         */
  value_release(accum1);
}

/*
 * NAME
 *   pic_opcode_andb_orb_xorb[_2]
 *
 * DESCRIPTION
 *   perform a bitwise and, or or xor
 *   dst = val1 op val2
 *
 * PARAMETERS
 *   pf :
 *   dst :
 *   val1:
 *   val2: (can be 0)
 *   op  : operator
 *
 * RETURN
 *
 * NOTES
 *   this must be done in-place, and only one array is allowed
 */
static void pic_opcode_andb_orb_xorb_2(pfile_t *pf, value_t *dst,
    value_t *src, operator_t op)
{
  variable_sz_t    sz;
  variable_sz_t    ii;
  pic_opcode_t     pop;
  variable_const_t n;
  unsigned char    n_last;

  n      = 0;
  n_last = -1;
  if (operator_andb == op) {
    pop = pic_opcode_andwf;
  } else if (operator_orb == op) {
    pop = pic_opcode_iorwf;
  } else {
    pop = pic_opcode_xorwf;
  }

  sz = value_sz_get(dst);
  if (!ISCONST(src)) {
    if (sz > value_sz_get(src)) {
      sz = value_sz_get(src);
      if (ISSIGNED(src)) {
        sign_set(pf, src, 0);
      }
    }
  }

  /* 2 parameter version, dst op= src */
  if (ISARRAY(dst) && ISARRAY(src)) {
    /* accum1 = src */
    value_t *tmp;

    tmp = pic_accum_get(pf, 1, sz, value_flags_get_all(src));
    if (tmp) {
      pic_opcode_assign(pf, tmp, src);
    }
    src = tmp;
  } else {
    value_lock(src);
  }

  if (src) {
    if (ISARRAY(dst)) {
      array_setup(pf, dst, array_setup_start);
    } else if (ISARRAY(src)) {
      array_setup(pf, src, array_setup_start);
    }

    if (ISCONST(src)) {
      n = value_const_get(src);
    }
    for (ii = 0; ii < sz; ii++) {
      unsigned char n_cur;

      if (ii) {
        if (ISARRAY(dst) || ISARRAY(src)) {
          pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
        }
        n >>= 8;
      }
      /* w = src */
      n_cur = -1;
      if (ISCONST(src)) {
        n_cur = n & 0xff;
      } else {
        pic_instr_append_f_d(pf, pic_opcode_movf, src, ii, pic_opdst_w);
      }
      /* dst op= w */
      /* if src is a constant, we can eliminate a few things as follows:
         x & 0xff == nop
         x | 0x00 == nop
         x ^ 0x00 == nop */
      if (ISCONST(src) 
         && (((0xff == n_cur) && (operator_andb == op))
           || ((0x00 == n_cur) 
             && ((operator_orb == op) || (operator_xorb == op))))) {
      } else {
        if (ISCONST(src)) {
          if (n_cur != n_last) {
            pic_instr_append_w_kn(pf, pic_opcode_movlw, n_cur);
            n_last = n_cur;
          }
        }
        pic_instr_append_f_d(pf, pop, dst, ii, pic_opdst_f);
      }
    }
    if (sz < value_sz_get(dst)) {
      if (ISSIGNED(src)) {
        /* andb -- no change
           orb  -- dst = 0xff
           xorb -- dst ^= 0xff */
        if (operator_andb != op) {
          pic_instr_append_w_kn(pf, pic_opcode_movlw, 0xff);
          while (ii < value_sz_get(dst)) {
            if (ISARRAY(dst)) {
              pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
            }
            pic_instr_append_f(pf, op, dst, ii);
            ii++;
          }
        }
      } else {
        /* andb -- dst =0
           orb  -- no change
           xorb -- no change */
        if (operator_andb == op) {
          while (ii < value_sz_get(dst)) {
            if (ISARRAY(dst)) {
              pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
            }
            pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
            ii++;
          }
        }
      }
    }
    value_release(src);
  }
}

static void pic_opcode_andb_orb_xorb(pfile_t *pf, value_t *dst,
    value_t *val1, value_t *val2, operator_t op)
{
  if (!val2) {
    pic_opcode_andb_orb_xorb_2(pf, dst, val1, op);
  } else {
    /* dst = val1, val1 = val2 */
    variable_sz_t sz;
    
    sz = value_sz_get(val1);
    if (sz < value_sz_get(val2)) {
      sz = value_sz_get(val2);
    }
    if (sz > value_sz_get(dst)) {
      sz = value_sz_get(dst);
    }

    if (value_flag_test(dst, VARIABLE_FLAG_NOTEMP)) {
      /* need to use accum#2 since #1 might be used above */
      value_t      *tmp;

      tmp = pic_accum_get(pf, 2, sz, value_flags_get_all(val1));
      if (tmp) {
        pic_opcode_assign(pf, tmp, val1);
        pic_opcode_andb_orb_xorb_2(pf, tmp, val2, op);
        pic_opcode_assign(pf, dst, tmp);
        value_release(tmp);
      }
    } else {
      variable_sz_t dst_sz;

      dst_sz = value_sz_get(dst);

      value_sz_set(dst, sz);
      pic_opcode_assign(pf, dst, val1);
      pic_opcode_andb_orb_xorb_2(pf, dst, val2, op);
      value_sz_set(dst, dst_sz);
      if (sz < dst_sz) {
        sign_extend(pf, dst, sz, need_sign(val1, val2));
      }
    }
  }
}

/*
 * NAME
 *   pic_opcode_add_sub[_2]
 *
 * DESCRIPTION
 *   create code for add or subtract
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   the functions:
 *      accum [+/-] = accum2
 *
 *      ADD:  movf   b0,W
 *            addwf  a0
 *            -- repeat --
 *            movf   b1,W
 *            btfsc  status, c
 *            incfsz b1,W
 *            addwf  a1,F
 *      8 bit : only one array
 *              any combination c, v, a --> dst
 *
 *      SUB:  movf   b0,W
 *            subwf  a0
 *            -- repeat --
 *            movf   b1,w
 *            btfss  status, c
 *            incfsz b1,w
 *            subwf  a1,f
 *
 *    if dst sz > 2, and val sz > 2 and dst and val reside
 *    in different banks, then val must be setup as an array
 */

/* dst += val1 or dst -= val1 */

/* note: this can be used even for non-arrays */
static unsigned bank(value_t *val)
{
  return (variable_base_get(value_variable_get(val)) / 128);
}

static void pic_opcode_add_sub_2(pfile_t *pf, value_t *dst, value_t *val,
    operator_t op)
{
  size_t sz;
  int    dst_is_array;
  int    val_is_array;
  pic_opcode_t action;
  pic_opcode_t test;

  if (operator_add == op) {
    action = pic_opcode_addwf;
    test   = pic_opcode_btfsc;
  } else {
    action = pic_opcode_subwf;
    test   = pic_opcode_btfss;
  }

  if ((value_sz_get(dst) > 2) 
      && (value_sz_get(val) > 2)
      && !ISARRAY(val)
      && !ISARRAY(dst)
      && (bank(dst) != bank(val))) {
    /* val is going to be given a dummy subscript */
    value_t *v2;

    value_clone(&v2, val);
    value_subscript_set(v2, VALUE_SUBSCRIPT_DUMMY);
    val = v2;
  } else {
    value_lock(val);
  }


  dst_is_array = 0 != value_subscript_get(dst);
  val_is_array = 0 != value_subscript_get(val);

  /* how big for the result? */
  sz = value_sz_get(dst);

  if (dst_is_array && val_is_array) {
    /* this cannot work! */
    value_t *tmp;

    tmp = pic_accum_from_val(pf, 1, val);
    if (tmp) {
      pic_opcode_assign(pf, tmp, val);
    }
    val = tmp;
    val_is_array = 0;
  } else {
    value_lock(val);
  }
  if (val) {
    if (ISCONST(val)) {
      /* x += constant. this only works for 1 or 2 byte
         values, after that we'll need an accumulator */
      if (sz > 2) {
        value_t *tmp;

        tmp = pic_accum_get(pf, 1, sz, 0);
        pic_opcode_assign(pf, tmp, val);
        value_release(val);
        val = tmp;
      }
    }

    /* setup FSR */
    if (dst_is_array) {
      array_setup(pf, dst, array_setup_start);
    } 

    if (ISCONST(val)) {
      /* dst += constant */
      variable_const_t n;

      n = value_const_get(val);
      pic_instr_append_w_kn(pf, pic_opcode_movlw, n & 0xff);
      pic_instr_append_f_d(pf, action, dst, 0, pic_opdst_f);
      if (sz > 1) {
        n >>= 8;
        if (dst_is_array) {
          pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
        }
        pic_instr_append_w_kn(pf, pic_opcode_movlw, n & 0xff);
        pic_instr_append_reg_flag(pf, test, "_status", "_c");
        pic_instr_append_w_kn(pf, pic_opcode_movlw, (n + 1) & 0xff);
        pic_instr_append_f_d(pf, action, dst, 
            (dst_is_array) ? 0 : 1, pic_opdst_f);
      }
    } else {
      size_t ii;

      /* dst += var */
      if (sz > value_sz_get(val)) {
        sign_set(pf, val, 0); /* this will setup the array appropriately */
      } else if (val_is_array) {
        array_setup(pf, val, array_setup_start); /* setup the array */
      }

      pic_instr_append_f_d(pf, pic_opcode_movf, val, 0, pic_opdst_w);
      pic_instr_append_f_d(pf, action, dst, 0, pic_opdst_f);
      for (ii = 1; ii < sz; ii++) {
        /* if either is an array, bump the array pointer */
        if (dst_is_array || (val_is_array && (ii < value_sz_get(val)))) {
          pic_instr_append_reg_d(pf, pic_opcode_incf,"_fsr", pic_opdst_f);
        }
        /* w = val[ii] */
        if (ii < value_sz_get(val)) {
          pic_instr_append_f_d(pf, pic_opcode_movf, val,
              (val_is_array) ? 0 : ii, pic_opdst_w);
        } else {
          pic_instr_append_reg_d(pf, pic_opcode_movf, "_sign", pic_opdst_w);
        }
        pic_instr_append_reg_flag(pf, test, "_status", "_c");
        pic_instr_append_f_d(pf, 
            (ii + 1 == sz) ? pic_opcode_incf : pic_opcode_incfsz, val, ii, 
            pic_opdst_w);
        pic_instr_append_f_d(pf, action, dst, ii, pic_opdst_f);
      }
    }
#if 0
    if (VALUE_SUBSCRIPT_DUMMY == value_subscript_get(val)) {
      value_subscript_set(val, 0);
    }
#endif
    value_release(val);
    value_release(val);
  }
}

/*
 * NAME
 *   pic_opcode_3_2
 *
 * DESCRIPTION
 *   given a three parameter operation (x = y + z), change
 *   it to a two parameter operation (x = y, x += z)
 *
 * PARAMETERS
 *   pf  : 
 *   dst :
 *   val1:
 *   val2:
 *   op  :
 *   fn  : fn to call as appropriate
 *
 * RETURN
 *
 * NOTES
 *   if val2 is 0, this simply calls fn
 */
typedef void (*pic_opcode_2_fn_t)(pfile_t *pf, value_t *dst, value_t *val,
  operator_t op);

static void pic_opcode_3_2(pfile_t *pf, value_t *dst, value_t *val1,
  value_t *val2, operator_t op, pic_opcode_2_fn_t fn)
{
  if (!val2) {
    fn(pf, dst, val1, op);
  } else {
    variable_sz_t sz;
    variable_sz_t dst_sz;
    flag_t        dst_flags;

    dst_sz = value_sz_get(dst);
    sz = dst_sz;
    dst_flags = value_flags_get_all(dst);

    if ((dst_flags & VARIABLE_FLAG_NOTEMP) 
      || (value_sz_get(dst) != value_sz_get(val1))) {
      value_t *tmp;

      tmp = pic_accum_get(pf, 1, dst_sz, value_flags_get_all(val1));
      if (tmp) {
        pic_opcode_assign(pf, tmp, val1);
        fn(pf, tmp, val2, op);
        pic_opcode_assign(pf, dst, tmp);
        value_release(tmp);
      }
    } else {
      value_sz_set(dst, sz);
      value_flags_set_all(dst, value_flags_get_all(val1));
      pic_opcode_assign(pf, dst, val1);
      fn(pf, dst, val2, op);
      value_flags_set_all(dst, dst_flags);
      if (sz < dst_sz) {
        value_sz_set(dst, dst_sz);

        sign_extend(pf, dst, sz, need_sign(val1, val2));
      }
    }
  }
}

static void pic_opcode_add_sub(pfile_t *pf, value_t *dst,
    value_t *val1, value_t *val2, operator_t op)
{

  if (ISCONST(val2) && (value_const_get(val2) <= 2)) {
    variable_const_t ii;

    for (ii = 0; ii < value_const_get(val2); ii++) {
      pic_opcode_incr_decr(pf, dst, val1, 
          (operator_add == op) ? operator_incr : operator_decr);
    }
  } else if (!val2 && ISCONST(val1) && (value_const_get(val1) <= 2)) {
    variable_const_t ii;

    for (ii = 0; ii < value_const_get(val2); ii++) {
      pic_opcode_incr_decr(pf, dst, dst,
          (operator_add == op) ? operator_incr : operator_decr);
    }
  } else {
    pic_opcode_3_2(pf, dst, val1, val2, op, pic_opcode_add_sub_2);
  }
}

/*
 * NAME
 *   pic_opcode_shift_2
 *
 * DESCRIPTION
 *   two parameter shift : dst <<= val
 *
 * PARAMETERS
 *   dst : destination
 *   val : value
 *
 * RETURN
 *
 * NOTES
 *   on entry, if val is const it is in range
 */
static void pic_opcode_shift_2(pfile_t *pf, value_t *dst, value_t *val,
    operator_t op)
{
  value_t *tmp;
  size_t   ii;
  int      tmp_is_array;
  pic_opcode_t pop;

  pop = (operator_shift_left == op) ? pic_opcode_rlf : pic_opcode_rrf;

  if (ISCONST(val)) {
    /* trivial case */
    variable_const_t bytemove;
    variable_const_t bitmove;

    bitmove = value_const_get(val);
    bytemove = bitmove / 8;
    bitmove &= 7;

    if (bytemove) {
      value_clone(&tmp, dst);
      value_sz_set(tmp, value_sz_get(tmp) - bytemove);
      value_baseofs_set(tmp, value_baseofs_get(tmp) + bytemove);
      if (operator_shift_left == op) {
        /* I should generalize assign appropriatly, but I'm going
           to wimp out here */
        /*pic_opcode_assign(pf, tmp, dst);*/
        for (ii = value_sz_get(dst) - 1; ii >= bytemove; ii--) {
          pic_instr_append_f_d(pf, pic_opcode_movf, dst, ii - 1, 
              pic_opdst_w);
          pic_instr_append_f(pf, pic_opcode_movwf, dst, ii);
        }
        for (ii = 0; ii < bytemove; ii++) {
          pic_instr_append_f(pf, pic_opcode_clrf, dst, ii);
        }
      } else {
        pic_opcode_assign(pf, dst, tmp);
        value_baseofs_set(tmp, value_baseofs_get(dst));
      }
    } else {
      tmp = dst;
      value_lock(tmp);
    }
    if (bitmove) {
      /* bitmove */
      if ((6U + value_sz_get(tmp)) >= ((1U + value_sz_get(tmp)) * bitmove)) {
        /* inline */
        for (ii = 0; ii < bitmove; ii++) {
          size_t jj;

          pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
          for (jj = 0; jj < value_sz_get(tmp); jj++) {
            if (operator_shift_left == op) {
              pic_instr_append_f_d(pf, pop, tmp, jj, pic_opdst_f);
            } else {
              pic_instr_append_f_d(pf, pop, tmp, value_sz_get(tmp) - jj - 1, 
                  pic_opdst_f);
            }
          }
        }
        value_release(tmp);
        tmp = 0;
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, bitmove + 1);
      }
    } else {
      value_release(tmp);
      tmp = 0;
    }
  } else {
    if (value_subscript_get(val)) {
      array_setup(pf, val, array_setup_start);
    }
    /* val + 1 --> sign */
    pic_instr_append_f_d(pf, pic_opcode_incf, val, 0, pic_opdst_w);
    tmp = dst;
    value_lock(tmp);
  }
  if (tmp) {
    label_t *top;
    label_t *test;
    value_t *loop;

    /* val --> sign */
    top  = 0;
    test = 0;
    loop = 0;
    pfile_label_alloc(pf, 0, 0, &top);
    pfile_label_alloc(pf, 0, 0, &test);
    loop = pic_loop_get(pf);

    if (top && test && loop) {
      pic_instr_append_f(pf, pic_opcode_movwf, loop, 0);
      pic_instr_append_n(pf, pic_opcode_goto, test);
      pic_instr_append_label(pf, top);
      if (value_subscript_get(tmp)) {
        array_setup(pf, tmp, array_setup_start);
        tmp_is_array = 1;
      } else {
        tmp_is_array = 0;
      }
      pic_instr_append_reg_flag(pf, pic_opcode_bcf, "_status", "_c");
      for (ii = 0; ii < value_sz_get(tmp); ii++) {
        if (ii && tmp_is_array) {
          pic_instr_append_reg_d(pf, pic_opcode_incf, "_fsr", pic_opdst_f);
        }
        pic_instr_append_f_d(pf, pop, tmp, 
            (operator_shift_left == op) ? ii : value_sz_get(tmp) - ii - 1, 
            pic_opdst_f);
      }
      pic_instr_append_label(pf, test);
      pic_instr_append_f_d(pf, pic_opcode_decfsz, loop, 0, pic_opdst_f);
      pic_instr_append_n(pf, pic_opcode_goto, top);
    }
    value_release(loop);
    label_release(top);
    label_release(test);
    value_release(tmp);
  }
}


static void pic_opcode_shift(pfile_t *pf, value_t *dst,
    value_t *val1, value_t *val2, operator_t op)
{
  if (ISCONST(val2) && (value_const_get(val2) > 8)) {
    /* instead of doing the normal dst = val1, dst op= val2
       we optimize a bit if any full byte moves are required */
    size_t           dst_sz;     /* might need to restore these */
    size_t           sz;
    value_t         *tmp;
    value_t         *tmp1; /* used when byteofs > 0 and shiftleft */
    variable_const_t byteofs;
    size_t           ii;

    byteofs = value_const_get(val2) / 8; /* # of full bytes to move */

    dst_sz = value_sz_get(dst);
    /* for right shifts, sz must be the *greater* of val1.sz or dst.sz 
       because val1 could shift down into the dest area */
    sz = dst_sz;
    if ((operator_shift_right == op) && (value_sz_get(val1) != sz)) {
      sz = value_sz_get(val1);
    } else if (value_sz_get(val1) < sz) {
      sz = value_sz_get(val1);
    }

    if ((sz > dst_sz) || (value_flag_test(dst, VARIABLE_FLAG_NOTEMP))) {
      /* must use an accumulator */
      tmp = pic_accum_get(pf, 1, sz, value_flags_get_all(val1));
    } else {
      tmp = dst;
    }
    if (operator_shift_left == op) {
      /* must copy from the end */
      for (ii = value_sz_get(tmp) - 1; ii >= byteofs; ii--) {
        pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii - 1, pic_opdst_w);
        pic_instr_append_f(pf, pic_opcode_movwf, tmp, ii);
      }
      /* clear any preceding bytes */
      for (ii = 0; ii < byteofs; ii++) {
        pic_instr_append_f(pf, pic_opcode_clrf, tmp, ii);
      }
      /* adjust the base -- no need to work though the zeros */
      value_clone(&tmp1, tmp);
      value_baseofs_set(tmp1, value_baseofs_get(tmp1) + byteofs);
    } else {
      /* may copy from the beginning */
      for (ii = byteofs; ii < value_sz_get(tmp); ii++) {
        pic_instr_append_f_d(pf, pic_opcode_movf, val1, ii, pic_opdst_w);
        pic_instr_append_f(pf, pic_opcode_movwf, tmp, ii - 1);
      }
      /* clear bytes at the end */
      for (ii = value_sz_get(tmp) - byteofs; ii < value_sz_get(tmp); ii++) {
        pic_instr_append_f(pf, pic_opcode_clrf, tmp, ii);
      }
      tmp1 = tmp;
    }
    value_sz_set(tmp1, value_sz_get(tmp1) - byteofs);
    if (value_const_get(val2) & 7) {
      /* let the other routine finish this up */
      value_t *bits;

      if (result_ok == pfile_constant_get(pf, value_const_get(val2) & 7, 
            &bits)) {
        pic_opcode_shift_2(pf, tmp1, bits, op);
        value_release(bits);
      }
    }
    /* restore tmp */
    value_sz_set(tmp1, value_sz_get(tmp1) + byteofs);
    if (tmp1 != tmp) {
      value_release(tmp1);
    }
    /* do stuff... */
    if (tmp != dst) {
      pic_opcode_assign(pf, dst, tmp);
      value_release(tmp);
    }
  } else {
    pic_opcode_3_2(pf, dst, val1, val2, op, pic_opcode_shift_2);
  }
}

/*
 * NAME
 *   pic_opcode_mul_div_mod
 *
 * DESCRIPTION
 *   create code for *, /, %  
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   these are not inlined
 */
static void pic_opcode_mul_div_mod(pfile_t *pf, value_t *dst,
    value_t *val1, value_t *val2, operator_t op)
{
  int issigned;

  issigned = (ISSIGNED(val1) && (ISCONST(val2) || ISSIGNED(val2)));

  if (!issigned && value_is_const(val2) 
    && (0 == (value_const_get(val2) & (value_const_get(val2) - 1)))) {
    /* this is a mul, div, or mod by a constant power of 2 */
    variable_const_t n;
    value_t *tmp;

    n = value_const_get(val2);

    if (operator_mod == op) {
      n--;
    } else {
      variable_const_t b;

      for (b = 0; n >>= 1; b++)
        ;
      n = b;
    }

    if (result_ok == pfile_constant_get(pf, n, &tmp)) {
      value_sz_set(tmp, value_sz_get(val2));
      switch (op) {
        case operator_mul: 
          pic_opcode_shift(pf, dst, val1, tmp, operator_shift_left);  
          break;
        case operator_div: 
          pic_opcode_shift(pf, dst, val1, tmp, operator_shift_right); 
          break;
        case operator_mod: 
          pic_opcode_andb_orb_xorb(pf, dst, val1, tmp, operator_andb); 
          break;
        default: break;
      }
      value_release(tmp);
    }
  } else {
    value_t    *accum1;
    value_t    *accum2;
    value_t    *accum3;
    value_t    *accum4;
    size_t      sz;

    sz = (operator_mul == op) ? mul_sz : div_sz;

    accum1 = pic_accum_get(pf, 1, sz, 0);
    accum2 = pic_accum_get(pf, 2, sz, 0);
    accum3 = pic_accum_get(pf, 3, sz, 0);
    accum4 = pic_accum_get(pf, 4, sz, 0);

    if (accum1 && accum2 && accum3 && accum4) {
      label_t    *action;
      const char *fnptr;

      pic_opcode_assign(pf, accum1, val1);
      pic_opcode_assign(pf, (operator_mul == op) ? accum2 : accum4, val2);

      fnptr = "{_unknown}";
      action = 0;
      if (operator_mul == op) {
        fnptr = "_mul";
      } else {
        if (issigned) {
          action = pfile_label_find(pf, pfile_log_none, "_div");
          if (!action) {
            pfile_label_alloc(pf, "_div", 0, &action);
          }
          if (action) {
            label_release(action);
          }
          fnptr = "_sdiv";
        } else {
          fnptr = "_div";
        }
      }
      action = pfile_label_find(pf, pfile_log_none, fnptr);
      if (!action) {
        pfile_label_alloc(pf, fnptr, 0, &action);
      }
      if (action) {
        pic_instr_append_n(pf, pic_opcode_call, action);
        label_release(action);
      }
      switch (op) {
        case operator_mul: pic_opcode_assign(pf, dst, accum3); break;
        case operator_div: pic_opcode_assign(pf, dst, accum3); break;
        case operator_mod: pic_opcode_assign(pf, dst, accum2); break;
        default: break;
      }
    }
    value_release(accum4);
    value_release(accum3);
    value_release(accum2);
    value_release(accum1);
  }
}

/*
 * NAME
 *   pic_opcode_relational
 *
 * DESCRIPTION
 *   perform the relational operations
 *
 * PARAMETERS
 *
 * RETURN
 *
 * NOTES
 *   signed:
 *      w    = val2_msb
 *      w   ~= 0x80
 *      temp = w
 *      w    = val1_msb
 *      w   ~= 0x80
 *      w    = temp - w
 *      goto both
 *   unsigned:
 *      w    = val1_msb
 *      w    = val2_msb - w
 *   both:
 *      btfsc STATUS, Z -- not equal, done!
 *      goto done
 *      w    = val1_lsb
 *      w    = val2_lsb - w
 *   done:
 *      Z -- val1 == val2  (operator_eq, operator_ne)
 *      C -- val1 <= val2  (operator_le)
 */

/* special case eq/ne as these are much simpler than the relations */
static void pic_opcode_eq_ne(pfile_t *pf, value_t *dst, value_t *val1,
  value_t *val2, operator_t op, size_t sz)
{
  variable_const_t n;
  boolean_t        first;
  value_t         *accum2;
  size_t           ct;
  value_t         *sign;

  accum2 = pic_accum_get(pf, 2, 1, 0);
  sign   = pic_sign_get(pf);

  n = 0;
  if (ISCONST(val2)) {
    n = value_const_get(val2);
  }

  if (sz > 1) {
    pic_instr_append_f(pf, pic_opcode_clrf, accum2, 0);
  }
  first = boolean_true;
  ct = sz;
  while (ct--) {
    if (first) {
      first = boolean_false;
    } else if (ISARRAY(val2)) {
      pic_instr_append_reg_d(pf, pic_opcode_decf, "_fsr", pic_opdst_w);
    }
    /* w = val2 */
    if (ISCONST(val2)) {
      pic_instr_append_w_kn(pf, pic_opcode_movlw, (n >> (8 * ct)) & 0xff);
    } else if (ct >= value_sz_get(val2)) {
      pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
    } else {
      pic_instr_append_f_d(pf, pic_opcode_movf, val2, ct, pic_opdst_w);
    }
    /* w = val1 - w */
    if (ct >= value_sz_get(val1)) {
      pic_instr_append_f_d(pf, pic_opcode_subwf, sign, 0, pic_opdst_w);
    } else {
      pic_instr_append_f_d(pf, pic_opcode_subwf, val1, ct, pic_opdst_w);
    }
    if (sz > 1) {
      pic_instr_append_f_d(pf, pic_opcode_iorwf, accum2, 0, 
          (ct) ? pic_opdst_f : pic_opdst_w);
    }
  }
  /* now...z is set and w is zero for equal, reset for no-equal */
  if (operator_eq == op) {
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
    pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
  } else {
    pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
    pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
  }
  pic_opcode_set_to_w(pf, dst);

  value_release(sign);
  value_release(accum2);
}

static void pic_opcode_relational(pfile_t *pf, value_t *dst,
    value_t *val1, value_t *val2, operator_t op)
{
  size_t           sz;
  value_t         *tmp; /* use 1 byte of accum2 as a temp register */
  variable_const_t n;
  boolean_t        release_val2; /* release val2 at the end */

  n = 0;
  if (!val2) {
    val2 = val1;
    val1 = dst;
  }
  release_val2 = boolean_false;

  if (ISCONST(val2)) {
    n = value_const_get(val2);
  } else if (ISARRAY(val1) && ISARRAY(val2)) {
    /* cannot work with two arrays, so move one to an accumulator */
    tmp = pic_accum_from_val(pf, 1, val2);
    pic_opcode_assign(pf, tmp, val2);
    val2 = tmp;
    release_val2 = boolean_true;
  } else if (ISARRAY(val1)) {
    /* swap val1 & val2 & change the operation */
    tmp = val1;
    val1 = val2;
    val2 = tmp;

    if (operator_lt == op) {
      op = operator_gt;
    } else if (operator_le == op) {
      op = operator_ge;
    } else if (operator_ge == op) {
      op = operator_le;
    } else if (operator_gt == op) {
      op = operator_lt;
    }
  }
  sz = value_sz_get(val1);
  if (sz < value_sz_get(val2)) {
    sz = value_sz_get(val2);
  }

  if (ISARRAY(val2)) {
    array_setup(pf, val2, array_setup_end);
  }

  if (sz > value_sz_get(val1)) {
    sign_set(pf, val1, 1);
  } else if (!ISCONST(val2) && (sz > value_sz_get(val2))) {
    sign_set(pf, val2, 1);
  }

  if ((operator_eq == op) || (operator_ne == op)) {
    pic_opcode_eq_ne(pf, dst, val1, val2, op, sz);
  } else {
    label_t  *done;
    boolean_t first;
    value_t  *accum2;
    value_t  *sign;

    accum2 = pic_accum_get(pf, 2, 1, 0);
    sign   = pic_sign_get(pf);

    if (sz > 1) {
      pfile_label_alloc(pf, 0, 0, &done);
    } else {
      done = 0;
    }

    if (accum2 && sign && ((1 == sz) || done)) {
      if (need_sign(val1, val2)) {
        if (ISCONST(val2)) {
          /* w = val2_msb ^ 0x80 */
          pic_instr_append_w_kn(pf, pic_opcode_movlw, 
              (n >> (8 * (sz - 1))) ^ 0x80);
        } else {
          /* w = val2_msb */
          if (sz > value_sz_get(val2)) {
            pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
          } else {
            pic_instr_append_f_d(pf, pic_opcode_movf, val2, sz - 1, 
                pic_opdst_w);
          }
          /* w ~= 0x80 */
          pic_instr_append_w_kn(pf, pic_opcode_xorlw, 0x80);
        }
        /* accum2 = w */
        pic_instr_append_f(pf, pic_opcode_movwf, accum2, 0);
        /* w = val1_msb */
        if (sz > value_sz_get(val1)) {
          pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
        } else {
          pic_instr_append_f_d(pf, pic_opcode_movf, val1, sz - 1, pic_opdst_w);
        }
        /* w ^= 0x80 */
        pic_instr_append_w_kn(pf, pic_opcode_xorlw, 0x80);
        /* w = temp - w */
        pic_instr_append_f_d(pf, pic_opcode_subwf, accum2, 0, pic_opdst_w);
        if (--sz) {
          /* don't need this for one-byte operations */
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_n(pf, pic_opcode_goto, done);
        }
        first = boolean_false;
      } else {
        first = boolean_true;
      }
      while (sz--) {
        if (first) {
          first = boolean_false;
        } else if (ISARRAY(val2)) {
          pic_instr_append_reg_d(pf, pic_opcode_decf, "_fsr", pic_opdst_w);
        }
        if (ISCONST(val2)) {
          /* sign = current byt of w */
          pic_instr_append_w_kn(pf, pic_opcode_movlw, (n >> (sz * 8)) & 0xff);
          pic_instr_append_f(pf, pic_opcode_movwf, sign, 0);
        }
        /* w = val1_msb */
        if (sz > value_sz_get(val1)) {
          pic_instr_append_f_d(pf, pic_opcode_movf, sign, 0, pic_opdst_w);
        } else {
          pic_instr_append_f_d(pf, pic_opcode_movf, val1, sz, pic_opdst_w);
        }
        /* w = val2_msb - w */
        if (ISCONST(val2) || (sz > value_sz_get(val2))) {
          pic_instr_append_f_d(pf, pic_opcode_subwf, sign, 0, pic_opdst_w);
        } else {
          pic_instr_append_f_d(pf, pic_opcode_subwf, val2, sz, pic_opdst_w);
        }
        if (sz) {
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_n(pf, pic_opcode_goto, done);
        }
      }
      if (done) {
        pic_instr_append_label(pf, done);
      }
      /* Z set if val1 == val2
         C set if val1 <= val2 */
      if ((operator_lt == op) || (operator_gt == op)) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1); /* assume success */
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0); /* assume failure */
      }
      switch (op) {
        case operator_eq: /* Z set */
        case operator_ne: /* Z clr */
          /* these are handled elsewhere */
          break;
        case operator_lt: /* Z clr, C set */
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
          break;
        case operator_le: /* Z set or C set */
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
          break;
        case operator_ge: /* Z set or C clr */
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_c");
          break;
        case operator_gt: /* Z clr and C clr */
          pic_instr_append_reg_flag(pf, pic_opcode_btfss, "_status", "_z");
          pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_c");
          break;
        default: break;
      }
      if ((operator_lt == op) || (operator_gt == op)) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 0);
      } else {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1); /* assume failure */
      }
      pic_opcode_set_to_w(pf, dst);
    }
    label_release(done);
    value_release(sign);
    value_release(accum2);
  }
  if (release_val2) {
    value_release(val2);
  }
}

static void pic_bitset_bitclr(pfile_t *pf, value_t *dst, value_t *val1,
  operator_t op)
{
  if (ISCONST(val1)) {
    variable_const_t n;
    pic_opcode_t         pop;

    pop = (operator_bitset == op)
      ? pic_opcode_bsf
      : pic_opcode_bcf;

    n = value_const_get(val1);
    if (n >= 8U * value_sz_get(dst)) {
      pfile_log(pf, pfile_log_warn, PIC_MSG_CONSTANT_RANGE);
    } else {
      if ((n / 8) || !value_name_get(val1)) {
        pic_instr_append_f_bn(pf, pop, dst, (size_t) (n / 8), 
            n & 7);
      } else {
        pic_instr_append_f_b(pf, pop, dst, (size_t) (n / 8), 
            val1);
      }
    }
  } else {
    value_t *accum;

    /* accum = 1 */
    accum = pic_accum_get(pf, 1, value_sz_get(dst), 0);
    if (accum) {
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      pic_opcode_set_to_w(pf, accum);
      /* accum <= val1 */
      pic_opcode_shift(pf, accum, accum, val1, operator_shift_left);
      if (operator_bitclr == op) {
        /* dst = dst & ~accum */
        pic_opcode_cmpb(pf, accum, accum);
        pic_opcode_andb_orb_xorb(pf, dst, dst, accum, operator_andb);
      } else {
        /* dst = dst | accum */
        pic_opcode_andb_orb_xorb(pf, dst, dst, accum, operator_orb);
      }
      value_release(accum);
    }
  }
}

static void pic_bittest(pfile_t *pf, value_t *dst, value_t *val1,
  value_t *val2)
{
  if (ISCONST(val2)) {
    variable_const_t n;

    n = value_const_get(val2);
    if (n >= 8U * value_sz_get(val1)) {
      pfile_log(pf, pfile_log_warn, PIC_MSG_CONSTANT_RANGE);
      pic_instr_append(pf, pic_opcode_clrw);
      pic_instr_append_reg_flag(pf, pic_opcode_btfsc, "_status", "_z");
    } else {
      if (dst) {
        pic_instr_append(pf, pic_opcode_clrw);
      }
      if ((n / 8) || !value_name_get(val1)) {
        pic_instr_append_f_bn(pf, pic_opcode_btfsc, val1, (size_t) n / 8, 
             n & 7);
      } else {
        pic_instr_append_f_b(pf, pic_opcode_btfsc, val1, (size_t) n / 8, val2);
      }
      if (dst) {
        pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      }
    }
    if (dst) {
      pic_opcode_set_to_w(pf, dst);
    }
  } else {
    /* dst = (val1 & (1 << val2)) != 0 */
    value_t *accum;

    accum = pic_accum_get(pf, 1, value_sz_get(val1), 0);
    if (accum) {
      pic_instr_append_w_kn(pf, pic_opcode_movlw, 1);
      pic_opcode_set_to_w(pf, accum);
      pic_opcode_shift_2(pf, accum, val2, operator_shift_left);
      pic_opcode_andb_orb_xorb(pf, accum, accum, val1, operator_andb);
      pic_opcode_logical(pf, accum, accum);
      if (dst) {
        pic_opcode_set_to_w(pf, dst);
      }
      value_release(accum);
    }
  }
}



/*
 * NAME
 *   pic_op
 *
 * DESCRIPTION
 *   create code for an operator
 *
 * PARAMETERS
 *   pf  : file
 *   cmd : command
 *
 * RETURN
 *   none
 *
 * NOTES
 *   for the moment, I'm going to wimp out on my array handling,
 *   in that the only thing that handles them is the assignment
 *   operator. All other operators will require the array value
 *   go into an accumulator
 */
void pic_op(pfile_t *pf, operator_t op, value_t *dst, value_t *val1,
    value_t *val2)
{
  value_t *accum1;
  value_t *accum2;

  accum1 = 0;
  accum2 = 0;
  if (operator_is_binary(op)) {
    if (!val2) {
      val2 = val1;
      val1 = dst;
    }
  } else if (operator_is_unary(op)) {
    if (!val1) {
      /* op dst --> dst op= val1 */
      val1 = dst;
    }
  }
  if (!operator_is_assign(op)
      && (operator_mul != op)
      && (operator_div != op)
      && (operator_mod != op)) {
    /* if either val1 or val2 are lookups, we'll need temporaries */
    unsigned accum_n;

    accum_n = 3;
    if (value_is_lookup(val1)) {
      accum1 = pic_accum_from_val(pf, accum_n, val1);
      value_assign_ct_bump(accum1, ctr_bump_incr);
      accum_n++;
      pic_opcode_assign_lookup(pf, accum1, val1, operator_assign);
    }
    if (value_is_lookup(val2)) {
      if (value_same(val1, val2)) {
        accum2 = accum1;
        value_lock(accum2);
      } else {
        accum2 = pic_accum_from_val(pf, accum_n, val2);
        value_assign_ct_bump(accum2, ctr_bump_incr);
        pic_opcode_assign_lookup(pf, accum2, val2, operator_assign);
      }
    }
    if (accum1) {
      val1 = accum1;
    }
    if (accum2) {
      val2 = accum2;
    }
  }
  if (dst) {
    value_flag_set(dst, VARIABLE_FLAG_ASSIGNED);
  }
  if (val1) {
    if (!value_flag_test(val1, VARIABLE_FLAG_CONST)
      || value_subscript_get(val1)) {
      /* mark this as used iff it's not a constant reference into a lookup table */
      value_flag_set(val1, VARIABLE_FLAG_USED);
    }
  }
  if (val2) {
    if (!value_flag_test(val2, VARIABLE_FLAG_CONST)
      || value_subscript_get(val2)) {
      value_flag_set(val2, VARIABLE_FLAG_USED);
    }
  }

  switch (op) {
    case operator_null: /* no action */
      break;
    /* unaries */
    case operator_logical:
    case operator_notl:
      pic_opcode_logical_not(pf, dst, val1, op);
      break;
    case operator_neg:
      pic_opcode_cmpb(pf, dst, val1);
      pic_opcode_incr_decr(pf, dst, dst, operator_incr);
      break;
    /* binaries */
    case operator_add:  /* dst = val1 + val2 */
    case operator_sub:  /* dst = val1 - val2 */
      pic_opcode_add_sub(pf, dst, val1, val2, op);
      break;
    case operator_mul:  /* dst = val1 * val2 */
    case operator_div:  /* dst = val1 / val2 */
    case operator_mod:  /* dst = val1 % val2 */
      pic_opcode_mul_div_mod(pf, dst, val1, val2, op);
      break;
    case operator_lt:   /* dst = val1 < val2 */
    case operator_le:
    case operator_eq:
    case operator_ne:
    case operator_ge:
    case operator_gt:
      pic_opcode_relational(pf, dst, val1, val2, op);
      break;
    case operator_andl:
    case operator_orl:
      pic_opcode_andl_orl(pf, dst, val1, val2, op);
      break;
    case operator_andb:
    case operator_orb:
    case operator_xorb:
      pic_opcode_andb_orb_xorb(pf, dst, val1, val2, op);
      break;
    case operator_bitset:      /* dst = dst | (1 << bit)        */
    case operator_bitclr:      /* dst = (dst & ~bit)            */
      pic_bitset_bitclr(pf, dst, val1, op);
      break;
    case operator_bittst:      /* dst = (dst & (1 << bit)) != 0 */
      pic_bittest(pf, dst, val1, val2);
      break;
    case operator_assign:      /* unary */
    case operator_cmpb:
      pic_opcode_assign_or_cmpb(pf, dst, val1, op);
      break;
    case operator_shift_left:
    case operator_shift_right:
      pic_opcode_shift(pf, dst, val1, val2, op);
      break;
    case operator_incr:
    case operator_decr:
      pic_opcode_incr_decr(pf, dst, val1, op);
      break;
  }
  value_release(accum2);
  value_release(accum1);
}


