/**********************************************************
 **
 ** variable.c : manipulators for variable_t
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ***********************************************************/
#include <stdlib.h>
#include <string.h>

#include "pfile.h"
#include "slist.h"
#include "variable.h"
#include "var_decl.h"

static void variable_free(slist_el_t *el)
{
  variable_t *var;

  var = slist_el_data(el);
  FREE(var);
}

static void min_sz_get(variable_t *var)
{
  /* determine the least # of bytes that fits this constant */
  variable_ct_t sz_max;
  variable_ct_t ii;

  for (ii = 0, sz_max = 0; ii < variable_ct_get(var); ii++) {
    variable_sz_t sz;
    variable_const_t n;
    unsigned char    n0, n1, n2, n3;

    n = variable_const_get(var, 0);
    n0 =  n        & 0xff;
    n1 = (n >> 8)  & 0xff;
    n2 = (n >> 16) & 0xff;
    n3 = (n >> 24) & 0xff;

    sz = 4; /* start out at the largest */
    if (0 == n3) {
      /* leading zeros (easy case) */
      variable_flag_clr(var, VARIABLE_FLAG_SIGNED);
      sz--;
      if (0 == n2) {
        sz--;
        if (0 == n1) {
          sz--;
        }
      }
    } else if ((0xff == n3) && (n2 & 0x80)) {
      /* signed, or at least we'll pretend it is. the code generator
       * will sign extend this if necessary, so determine the real
       * size */
      variable_flag_set(var, VARIABLE_FLAG_SIGNED);
      sz--;
      if ((0xff == n2) && (n1 & 0x80)) {
        sz--;
        if ((0xff == n1) && (n0 & 0x80)) {
          sz--;
        }
      }
    }
    if (sz > sz_max) {
      sz_max = sz;
    }
  }
  var->sz = sz_max;
}

/* basic functions */
result_t variable_alloc(variable_t **dst, const char *name, 
    flag_t flags, variable_sz_t sz, variable_ct_t ct,
    const variable_const_t *data)
{
  variable_t *var;
  result_t    rc;
  size_t      name_sz;
  size_t      data_sz;

  name_sz = (name) ? 1 + strlen(name) : 0;
  data_sz = (flags & VARIABLE_FLAG_CONST) ? ct * sizeof(*data) : 0;

  var = MALLOC(sizeof(*var) + name_sz + data_sz);
  if (!var) {
    rc = result_memory;
  } else {
    char *ptr;

    rc = result_ok;
    ptr = (char *) (var + 1);

    if (data_sz) {
      var->data = (void *) ptr;
      if (data) {
        memcpy(var->data, data, data_sz);
      } else {
        memset(var->data, 0, data_sz);
      }
      ptr += data_sz;
    } else {
      var->data = 0;
    }

    if (name_sz) {
      var->name = ptr;
      memcpy(var->name, name, name_sz);
    } else {
      var->name = 0;
    } 
    if (0 == rc) {
      slist_el_init(&var->el, variable_free);
      var->flags      = flags;
      if (0 == sz) {
        var->flags |= VARIABLE_FLAG_NOSIZE;
      }
      var->sz         = sz;
      var->ct         = ct;
      var->base       = VARIABLE_BASE_UNKNOWN;
      var->expr_ct    = 0;
      var->ctr_assign = 0;
      var->ctr_use    = 0;
      var->region     = 0;
      *dst = var;
      if (variable_flag_test(var, VARIABLE_FLAG_CONST)
          && (sz == 0)
          && data) {
        min_sz_get(var);
      }
    }
  }
  return rc;
}

void variable_release(variable_t *var)
{
  if (var) {
    slist_el_release(&var->el);
  }
}

void variable_lock(variable_t *var)
{
  slist_el_lock(&var->el);
}

/* member retrieval */
boolean_t variable_flag_test(const variable_t *var, flag_t flags)
{
  return (var) ? ((var->flags & flags) == flags) : 0;
}

void variable_flag_set(variable_t *var, flag_t flags)
{
  if (var) {
    var->flags |= flags;
  }
}

void variable_flag_clr(variable_t *var, flag_t flags)
{
  if (var) {
    var->flags &= ~flags;
  }
}

flag_t variable_flags_get_all(const variable_t *var)
{
  return (var) ? var->flags : 0;
}

void variable_flags_set_all(variable_t *var, flag_t flags)
{
  var->flags = flags;
}

const char     *variable_name_get(const variable_t *var)
{
  return (var) ? var->name : 0;
}

void variable_sz_set(variable_t *var, variable_sz_t sz)
{
  var->sz = sz;
}

variable_sz_t   variable_sz_get(const variable_t *var)
{
  return (var) ? var->sz : 0;
}

variable_ct_t   variable_ct_get(const variable_t *var)
{
  return (var) ? var->ct : 0;
}

variable_base_t variable_base_get(const variable_t *var)
{
  return (var) ? var->base : 0;
}

void variable_base_set(variable_t *var, variable_base_t base)
{
  var->base = base;
}

variable_const_t variable_const_get(const variable_t *var,
  variable_ct_t pos)
{
  return (var && var->data && (pos < var->ct)) ? var->data[pos] : 0;
}

void variable_const_set(variable_t *var, variable_ct_t pos,
  variable_const_t n)
{
  if (var->data && (pos < var->ct)) {
    var->data[pos] = n;
    if (variable_flag_test(var, VARIABLE_FLAG_NOSIZE)) {
      min_sz_get(var);
    }
  }
}

slist_el_t *variable_slist_el_get(variable_t *var)
{
  return (var) ? &var->el : 0;
}

void variable_list_init(varlist_t *lst)
{
  slist_init(&lst->lst);
}

void variable_list_append(varlist_t *lst, variable_t *var)
{
  slist_append(&lst->lst, &var->el);
}

variable_t *variable_list_find(varlist_t *lst, varfind_cb_t cb, void *cb_arg,
    const void *data)
{
  variable_t *var;

  for (var = slist_el_data(slist_head(&lst->lst));
       var && !cb(cb_arg, var, data);
       var = slist_el_data(slist_el_next(&var->el)))
    ;
  if (var) {
    variable_lock(var);
  }
  return var;
}

void variable_list_reset(varlist_t *lst, pfile_t *pf)
{
  while (slist_head(&lst->lst)) {
    if (variable_refct_get(variable_list_head(lst)) > 1) {
      variable_t *var;

      var = variable_list_head(lst);
      if (variable_name_get(var)) {
        pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_VAR_LEAK, 
            variable_name_get(var));
      } else {
        pfile_log(pf, pfile_log_debug, PFILE_MSG_DBG_CONST_LEAK, 
            variable_const_get(var, 0));
      }
    }
    slist_el_remove(slist_head(&lst->lst));
  }
}

unsigned variable_list_ct_get(const varlist_t *lst)
{
  return (lst) ? slist_ct_get(&lst->lst) : 0;
}

void variable_dump(const variable_t *var, FILE *out)
{
#if 0
  fprintf(out, "{");
  if (var->name) {
    fprintf(out, "%s", variable_name_get(var));
  }
  if (var->flags & VARIABLE_FLAG_CONST) {
    if (var->name) {
      fprintf(out, ":n(%ld)sz(%u)", variable_const_get(var),
        variable_sz_get(var));
    } else {
      fprintf(out, "%ld:sz(%u)", variable_const_get(var),
        variable_sz_get(var));
    }
  } else {
    fprintf(out, ":s(%u):c(%u):b(0x%x)", variable_sz_get(var), 
      variable_ct_get(var), variable_base_get(var));
    fprintf(out, ":f(%c%c%c%c", 
        (var->flags & VARIABLE_FLAG_CONST) ? 'C' : '-',
        (var->flags & VARIABLE_FLAG_SIGNED) ? 'S' : '-',
        (var->flags & VARIABLE_FLAG_ASSIGNED) ? 'A' : '-',
        (var->flags & VARIABLE_FLAG_USED) ? 'U' : '-');
    fprintf(out, ")");
  }
  fprintf(out, "r(%u)}", slist_el_refct_get(&var->el));
#else
  fputc('{', out);
  if (var->name) {
    fprintf(out, "%s", variable_name_get(var));
  }
  if (var->flags & VARIABLE_FLAG_CONST) {
    variable_ct_t ii;
    const char   *fmt;

    fmt = (var->flags & VARIABLE_FLAG_SIGNED) ? "%s%ld" : "%s%lu";

    fprintf(out, "%s", (variable_name_get(var)) ? ":" : "");
    for (ii = 0; ii < variable_ct_get(var); ii++) {
      fprintf(out, fmt, (ii) ? ", " : "", variable_const_get(var, ii));
    }
  }
  fputc('}', out);
#endif
}

refct_t variable_refct_get(const variable_t *var)
{
  return (var) ? slist_el_refct_get(&var->el) : 0;
}

void variable_ct_set(variable_t *var, variable_ct_t ct)
{
  var->ct = ct;
}

variable_t *variable_link_get(const variable_t *var)
{
  return (var) ? slist_el_data(slist_el_next(&var->el)) : 0;
}

variable_t *variable_list_head(varlist_t *lst)
{
  return slist_el_data(slist_head(&lst->lst));
}

ctr_t variable_expr_ct_get(const variable_t *var)
{
  return (var) ? var->expr_ct : 0;
}

void variable_expr_ct_bump(variable_t *var, ctr_bump_t dir)
{
  switch (dir) {
    case ctr_bump_incr:
      var->expr_ct++; 
      break;
    case ctr_bump_decr: 
      if (var->expr_ct) {
        var->expr_ct--;
      }
      break;
  }
}

ctr_t variable_assign_ct_get(const variable_t *var)
{
  return var->ctr_assign;
}

void  variable_assign_ct_bump(variable_t *var, ctr_bump_t dir)
{
  switch (dir) {
    case ctr_bump_incr: 
      var->ctr_assign++; 
      break;
    case ctr_bump_decr: 
      if (var->ctr_assign) {
        var->ctr_assign--; 
      }
      break;
  }
}

ctr_t variable_use_ct_get(const variable_t *var)
{
  return var->ctr_use;
}

void variable_use_ct_bump(variable_t *var, ctr_bump_t dir)
{
  switch (dir) {
    case ctr_bump_incr: 
      var->ctr_use++; 
      break;
    case ctr_bump_decr: 
      if (var->ctr_use) {
        var->ctr_use--; 
      }
      break;
  }
}

ctr_t variable_ref_ct_get(const variable_t *var)
{
  return slist_el_refct_get(&var->el);
}

void variable_region_set(variable_t *var, variable_region_t r)
{
  var->region |= 1 << r;
}

boolean_t variable_region_test(const variable_t *var, 
    variable_region_t r)
{
  return (var->region & (1 << r)) != 0;
}

void variable_region_clr(variable_t *var, variable_region_t r)
{
  var->region &= ~ (1 << r);
}

variable_region_t variable_region_get_all(variable_t *var)
{
  return var->region;
}

void variable_region_set_all(variable_t *var, variable_region_t r)
{
  var->region = r;
}

