/************************************************************
 **
 ** variable.h : variable declarations
 **
 ** Copyright (c) 2004, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#ifndef variable_h__
#define variable_h__

#include <stdio.h>
#include "ptype.h"
#include "slist.h"

struct varlist_ {
  slist_t lst;
};

typedef enum {
  variable_type_null = 0,
  variable_type_var,
  variable_type_const,
  variable_type_label
} variable_type_t;

typedef unsigned char  variable_sz_t;
typedef unsigned short variable_ct_t;
typedef unsigned short variable_base_t;
typedef unsigned long  variable_const_t;
typedef unsigned char  variable_region_t;

#define VARIABLE_BASE_UNKNOWN ((variable_base_t) -1)

#define VARIABLE_FLAG_CONST    0x0001 /* named constant             */
#define VARIABLE_FLAG_SIGNED   0x0002 /* signed value               */
#define VARIABLE_FLAG_VOLATILE 0x0004 /* can change without warning */
                                      /* assigned and used assumed  */
#define VARIABLE_FLAG_NOTEMP   0x0008 /* cannot hold partial result */
#define VARIABLE_FLAG_NOSIZE   0x0010 /* size is dynamically determined
                                         (for constants only)       */
#define VARIABLE_FLAG_USED     0x0020 /* the usage counter was > 0
                                         after the first pass */
#define VARIABLE_FLAG_ASSIGNED 0x0040 /* the assign counter was >0
                                         after the first pass */
#define VARIABLE_FLAG_NO_OPT   0x0080 /* don't optimize away this var */

/* these bits are for temporary variables only
   if VARIABLE_FLAG_INTERRUPT is set, the variable is used in the
      interrupt context. if VARIABLE_FLAG_USER is also set, the
      same *code* is executed at both user and interrupt level,
   the possibilities:
     neither flag set -- this variable can be ignored
     interrupt only set -- this variable will be replaced with
       a like-named interrupt context variable
      (eg _t0 --> _tisr0)
     interrupt + user set -- this variable will be saved and restored
       during the context switch as follows:
         _tisr0 = _t0
         ... isr code
         _t0 = _tisr0
*/
#define VARIABLE_FLAG_INTERRUPT   0x0100
#define VARIABLE_FLAG_USER        0x0200
typedef boolean_t (*varfind_cb_t)(void *arg, const variable_t *var, 
    const void *data);

void variable_list_init(varlist_t *lst);
void variable_list_append(varlist_t *lst, variable_t *var);
variable_t *variable_list_find(varlist_t *lst, varfind_cb_t fn, void *arg,
    const void *data);
void variable_list_reset(varlist_t *lst, pfile_t *pf);
unsigned variable_list_ct_get(const varlist_t *lst);

/* basic functions */
result_t variable_alloc(variable_t **dst, const char *name, 
    flag_t flags, variable_sz_t sz, variable_ct_t ct,
    const variable_const_t *data);
void variable_release(variable_t *var);
void variable_lock(variable_t *var);

/* member retrieval */
boolean_t       variable_flag_test(const variable_t *var, flag_t flags);
void            variable_flag_set(variable_t *var, flag_t flags);
void            variable_flag_clr(variable_t *var, flag_t flags);
void            variable_flags_set_all(variable_t *var, flag_t flags);
flag_t          variable_flags_get_all(const variable_t *var);

const char     *variable_name_get(const variable_t *var);
variable_sz_t   variable_sz_get(const variable_t *var);
void            variable_sz_set(variable_t *var, variable_sz_t sz);
variable_ct_t   variable_ct_get(const variable_t *var);
void            variable_ct_set(variable_t *var, variable_ct_t ct);
variable_base_t variable_base_get(const variable_t *var);
void            variable_base_set(variable_t *var, variable_base_t base);
variable_const_t variable_const_get(const variable_t *var,
  variable_ct_t pos);
void             variable_const_set(variable_t *var, 
  variable_ct_t pos, variable_const_t n);
refct_t          variable_refct_get(const variable_t *var);

variable_t     *variable_link_get(const variable_t *var);
void            variable_link_set(variable_t *var, variable_t *lnk);

const char  *variable_const_buf_get(const variable_t *var);
void            variable_const_buf_set(variable_t *var, const char *buf);

void            variable_dump(const variable_t *var, FILE *dst);

variable_t     *variable_list_head(varlist_t *var);

ctr_t           variable_expr_ct_get(const variable_t *var);
void            variable_expr_ct_bump(variable_t *var, ctr_bump_t dir);

ctr_t          variable_assign_ct_get(const variable_t *var);
void           variable_assign_ct_bump(variable_t *var, ctr_bump_t dir);

ctr_t          variable_use_ct_get(const variable_t *var);
void           variable_use_ct_bump(variable_t *var, ctr_bump_t dir);

ctr_t          variable_ref_ct_get(const variable_t *var);

void           variable_region_set(variable_t *var, variable_region_t r);
boolean_t      variable_region_test(const variable_t *var, 
    variable_region_t r);
void           variable_region_clr(variable_t *var, variable_region_t r);

variable_region_t variable_region_get_all(variable_t *var);
void              variable_region_set_all(variable_t *var,
    variable_region_t r);
#ifdef MSDOS
#include "var_decl.h"
#endif

#endif /* variable_h__ */

